/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*-
 *
 * Copyright (C) 2007 William Jon McCann <mccann@jhu.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street - Suite 500, Boston, MA 02110-1335, USA.
 *
 */

#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>

#include <glib.h>
#include <glib/gi18n.h>
#include <glib-object.h>
#include <gio/gio.h>

#define GNOME_DESKTOP_USE_UNSTABLE_API
#include <libgnome-desktop/gnome-pnp-ids.h>

#include "cdos-settings-plugin-info.h"
#include "cdos-settings-manager.h"
#include "cdos-settings-profile.h"

#define CSD_MANAGER_DBUS_NAME "org.freedesktop.SettingsDaemon"
#define CSD_MANAGER_DBUS_PATH "/org/freedesktop/SettingsDaemon"
#define CSD_MANAGER_DBUS_INTERFACE CSD_MANAGER_DBUS_NAME
#define DEFAULT_SETTINGS_PREFIX "org.cdos.settings-daemon"

#define PLUGIN_EXT ".cdos-settings-plugin"

#define CDOS_SETTINGS_MANAGER_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), CDOS_TYPE_SETTINGS_MANAGER, CdosSettingsManagerPrivate))

G_LOCK_DEFINE_STATIC (thread_pool_lock);
GThreadPool *thread_pool = NULL;

static const gchar introspection_xml[] =
"<node name='/org/freedesktop/SettingsDaemon'>"
"  <interface name='org.freedesktop.SettingsDaemon'>"
"    <annotation name='org.freedesktop.DBus.GLib.CSymbol' value='found_settings_manager'/>"
"    <method name='ToggleDebug'>"
"    </method>"
"    <signal name='PluginActivated'>"
"      <arg name='name' type='s'/>"
"    </signal>"
"    <signal name='PluginDeactivated'>"
"      <arg name='name' type='s'/>"
"    </signal>"
"  </interface>"
"</node>";

struct CdosSettingsManagerPrivate
{
        guint                       owner_id;
        GDBusNodeInfo              *introspection_data;
        GDBusConnection            *connection;
        GSettings                  *settings;
        GnomePnpIds                *pnp_ids;
        GSList                     *plugins;
};

static void     cdos_settings_manager_class_init  (CdosSettingsManagerClass *klass);
static void     cdos_settings_manager_init        (CdosSettingsManager      *settings_manager);
static void     cdos_settings_manager_finalize    (GObject                   *object);

G_DEFINE_TYPE (CdosSettingsManager, cdos_settings_manager, G_TYPE_OBJECT)

static gpointer manager_object = NULL;

GQuark
cdos_settings_manager_error_quark (void)
{
        static GQuark ret = 0;
        if (ret == 0) {
                ret = g_quark_from_static_string ("found_settings_manager_error");
        }

        return ret;
}

static void
maybe_activate_plugin (CdosSettingsPluginInfo *info, gpointer user_data)
{
//	G_LOCK (thread_pool_lock);
	gchar *plugin = NULL;
	plugin = cdos_settings_plugin_info_get_location (info);
	if (g_strcmp0 (plugin, "xsettings") == 0 || g_strcmp0 (plugin, "xrandr") == 0){
		return;
	}
	g_message ("Cdos-settings-daemon: activate plugin %s", plugin);
        if (cdos_settings_plugin_info_get_enabled (info)) {
                gboolean res;
                res = cdos_settings_plugin_info_activate (info);
                if (res) {
                        g_debug ("Plugin %s: active", cdos_settings_plugin_info_get_location (info));
                } else {
                        g_warning ("Plugin %s: activation failed.", cdos_settings_plugin_info_get_location (info));
                }
        } else {
                g_debug ("Plugin %s: inactive", cdos_settings_plugin_info_get_location (info));
        }
//	G_UNLOCK (thread_pool_lock);
}

static void
maybe_activate_plugin_xsettings_xrandr (CdosSettingsPluginInfo *info)
{
        if (cdos_settings_plugin_info_get_enabled (info)) {
                gboolean res;
                res = cdos_settings_plugin_info_activate (info);
                if (res) {
                        g_debug ("Plugin %s: active", cdos_settings_plugin_info_get_location (info));
                } else {
                        g_warning ("Plugin %s: activation failed.", cdos_settings_plugin_info_get_location (info));
                }
        } else {
                g_debug ("Plugin %s: inactive", cdos_settings_plugin_info_get_location (info));
        }
}
static gint
compare_location (CdosSettingsPluginInfo *a,
                  CdosSettingsPluginInfo *b)
{
        const char *loc_a;
        const char *loc_b;

        loc_a = cdos_settings_plugin_info_get_location (a);
        loc_b = cdos_settings_plugin_info_get_location (b);

        if (loc_a == NULL || loc_b == NULL) {
                return -1;
        }

        return strcmp (loc_a, loc_b);
}

static int
compare_priority (CdosSettingsPluginInfo *a,
                  CdosSettingsPluginInfo *b)
{
        int prio_a;
        int prio_b;

        prio_a = cdos_settings_plugin_info_get_priority (a);
        prio_b = cdos_settings_plugin_info_get_priority (b);

        return prio_a - prio_b;
}

static void
emit_signal (CdosSettingsManager    *manager,
             const char              *signal,
             const char              *name)
{
        GError *error = NULL;

        /* FIXME: maybe we should queue those up until the D-Bus
         * connection is available... */
        if (manager->priv->connection == NULL)
                return;

        if (g_dbus_connection_emit_signal (manager->priv->connection,
                                           NULL,
                                           CSD_MANAGER_DBUS_PATH,
                                           CSD_MANAGER_DBUS_INTERFACE,
                                           "PluginActivated",
                                           g_variant_new ("(s)", name),
                                           &error) == FALSE) {
                g_debug ("Error emitting signal: %s", error->message);
                g_error_free (error);
        }

}

static void
on_plugin_activated (CdosSettingsPluginInfo *info,
                     CdosSettingsManager    *manager)
{
        const char *name;

        name = cdos_settings_plugin_info_get_location (info);
        g_debug ("CdosSettingsManager: emitting plugin-activated %s", name);
        emit_signal (manager, "PluginActivated", name);
}

static void
on_plugin_deactivated (CdosSettingsPluginInfo *info,
                       CdosSettingsManager    *manager)
{
        const char *name;

        name = cdos_settings_plugin_info_get_location (info);
        g_debug ("CdosSettingsManager: emitting plugin-deactivated %s", name);
        emit_signal (manager, "PluginDeactivated", name);
}

static gboolean
contained (const char * const *items,
           const char         *item)
{
        while (*items) {
                if (g_strcmp0 (*items++, item) == 0) {
                        return TRUE;
                }
        }

        return FALSE;
}

static gboolean
is_schema (const char *schema)
{
        return contained (g_settings_list_schemas (), schema);
}


static void
_load_file (CdosSettingsManager *manager,
            const char           *filename)
{
        CdosSettingsPluginInfo *info;
        char                    *key_name;
        GSList                  *l;

        g_debug ("Loading plugin: %s", filename);
        cdos_settings_profile_start ("%s", filename);

        info = cdos_settings_plugin_info_new_from_file (filename);
        if (info == NULL) {
                goto out;
        }

        l = g_slist_find_custom (manager->priv->plugins,
                                 info,
                                 (GCompareFunc) compare_location);
        if (l != NULL) {
                goto out;
        }

        key_name = g_strdup_printf ("%s.plugins.%s",
                                    DEFAULT_SETTINGS_PREFIX,
                                    cdos_settings_plugin_info_get_location (info));

        /* Ignore unknown schemas or else we'll assert */
        if (is_schema (key_name)) {
                manager->priv->plugins = g_slist_prepend (manager->priv->plugins,
                                                          g_object_ref (info));

                g_signal_connect (info, "activated",
                                  G_CALLBACK (on_plugin_activated), manager);
                g_signal_connect (info, "deactivated",
                                  G_CALLBACK (on_plugin_deactivated), manager);

                cdos_settings_plugin_info_set_settings_prefix (info, key_name);
        } else {
                g_warning ("Ignoring unknown module '%s'", key_name);
        }

        /* Priority is set in the call above */
        g_free (key_name);

 out:
        if (info != NULL) {
                g_object_unref (info);
        }

        cdos_settings_profile_end ("%s", filename);
}

static void
_load_dir (CdosSettingsManager *manager,
           const char           *path)
{
        GError     *error;
        GDir       *d;
        const char *name;

        g_debug ("Loading settings plugins from dir: %s", path);
        cdos_settings_profile_start (NULL);

        error = NULL;
        d = g_dir_open (path, 0, &error);
        if (d == NULL) {
                g_warning ("%s", error->message);
                g_error_free (error);
                return;
        }

        while ((name = g_dir_read_name (d))) {
                char *filename;

                if (!g_str_has_suffix (name, PLUGIN_EXT)) {
                        continue;
                }

                filename = g_build_filename (path, name, NULL);
                if (g_file_test (filename, G_FILE_TEST_IS_REGULAR)) {
                        _load_file (manager, filename);
                }
                g_free (filename);
        }

        g_dir_close (d);

        cdos_settings_profile_end (NULL);
}

static void
settings_plugin_add_into_pool (gpointer data, gpointer user_data){
	g_thread_pool_push (thread_pool, data, NULL);
}

static gboolean
_load_all (CdosSettingsManager *manager)
{
        cdos_settings_profile_start (NULL);

        /* load system plugins */
        _load_dir (manager, CDOS_SETTINGS_PLUGINDIR G_DIR_SEPARATOR_S);

        manager->priv->plugins = g_slist_sort (manager->priv->plugins, (GCompareFunc) compare_priority);
// 或许我可以在这里继续 neng 个线程池来试试
//	thread_pool = g_thread_pool_new ((GFunc) maybe_activate_plugin, NULL, 20, TRUE, NULL);
	
	g_message ("IN LOAD ALL\n");
//	g_slist_foreach	 (manager->priv->plugins, (GFunc) settings_plugin_add_into_pool, NULL);
        g_slist_foreach (manager->priv->plugins, (GFunc) maybe_activate_plugin, NULL);
//	g_thread_pool_free (thread_pool, 0, 1);

        cdos_settings_profile_end (NULL);
	return FALSE;
}

static void
_load_xsettings_xrandr (CdosSettingsManager *manager){
        gchar *filename = NULL;
	gchar *key_name = NULL;
        CdosSettingsPluginInfo *info;
        
	filename = g_build_filename (CDOS_SETTINGS_PLUGINDIR G_DIR_SEPARATOR_S, "xsettings.cdos-settings-plugin", NULL);
	info = cdos_settings_plugin_info_new_from_file (filename);

	key_name = g_strdup_printf ("%s.plugins.%s",
                                    DEFAULT_SETTINGS_PREFIX,
                                    cdos_settings_plugin_info_get_location (info));

	g_signal_connect (info, "activated",
                               G_CALLBACK (on_plugin_activated), manager);

        g_signal_connect (info, "deactivated",
                             	G_CALLBACK (on_plugin_deactivated), manager);

	cdos_settings_plugin_info_set_settings_prefix (info, key_name); // glib-2.0 目录下的 xml
	// 有了 info 结构之后，就可以激活这个插件了
	maybe_activate_plugin_xsettings_xrandr (info);		

        //不加载xrandr插件
        #if 0
 	filename = NULL;
	CdosSettingsPluginInfo *info2 = NULL;
	key_name = NULL;

	filename = g_build_filename (CDOS_SETTINGS_PLUGINDIR G_DIR_SEPARATOR_S, "xrandr.cdos-settings-plugin", NULL);
	info2 = cdos_settings_plugin_info_new_from_file (filename);

	key_name = g_strdup_printf ("%s.plugins.%s",
                                    DEFAULT_SETTINGS_PREFIX,
                                    cdos_settings_plugin_info_get_location (info));

	g_signal_connect (info2, "activated",
                               G_CALLBACK (on_plugin_activated), manager);

        g_signal_connect (info2, "deactivated",
                             	G_CALLBACK (on_plugin_deactivated), manager);

	cdos_settings_plugin_info_set_settings_prefix (info2, key_name); // glib-2.0 目录下的 xml
	// 有了 info 结构之后，就可以激活这个插件了
	maybe_activate_plugin_xsettings_xrandr (info2);
        #endif		
	
	g_free (filename);
	g_free (key_name);
}

static void
_unload_plugin (CdosSettingsPluginInfo *info, gpointer user_data)
{
        if (cdos_settings_plugin_info_get_enabled (info)) {
                cdos_settings_plugin_info_deactivate (info);
        }
        g_object_unref (info);
}

static void
_unload_all (CdosSettingsManager *manager)
{
         g_slist_foreach (manager->priv->plugins, (GFunc) _unload_plugin, NULL);
         g_slist_free (manager->priv->plugins);
         manager->priv->plugins = NULL;
}


static void
handle_method_call (GDBusConnection       *connection,
                    const gchar           *sender,
                    const gchar           *object_path,
                    const gchar           *interface_name,
                    const gchar           *method_name,
                    GVariant              *parameters,
                    GDBusMethodInvocation *invocation,
                    gpointer               user_data)
{
        gchar *debug = NULL;

        g_debug ("Calling method '%s.%s' for Settings Manager",
                 interface_name, method_name);

        if (g_strcmp0 (interface_name, CSD_MANAGER_DBUS_INTERFACE) == 0) {
                // 打开、关闭调试模式接口
                if (g_strcmp0 (method_name, "ToggleDebug") == 0) {
                        debug = g_getenv ("G_MESSAGES_DEBUG");
                        if (debug) {
                                g_unsetenv ("G_MESSAGES_DEBUG");
                        } else {
                                g_setenv ("G_MESSAGES_DEBUG", "all", FALSE);
                        }

                        g_dbus_method_invocation_return_value (invocation, NULL);
                }
        } else {
                g_warning ("not recognised interface: %s", interface_name);
        }
}

static const GDBusInterfaceVTable interface_vtable =
{
        handle_method_call,     // 回调方法
        NULL,   // 获取属性回调
        NULL,   // 设置属性回调
};

static void
on_bus_gotten (GObject             *source_object,
               GAsyncResult        *res,
               CdosSettingsManager *manager)
{
        GDBusConnection *connection;
        GDBusInterfaceInfo **infos;
        GError *error = NULL;
        guint i;

        connection = g_bus_get_finish (res, &error);
        if (connection == NULL) {
                g_warning ("Could not get session bus: %s", error->message);
                g_error_free (error);
                return;
        }
        manager->priv->connection = connection;


        infos = manager->priv->introspection_data->interfaces;
        for (i = 0; infos[i] != NULL; i++) {
                g_dbus_connection_register_object (connection,
                                                   CSD_MANAGER_DBUS_PATH,
                                                   infos[i],
                                                   &interface_vtable,
                                                   NULL,
                                                   NULL,
                                                   NULL);
        }
}

static void
register_manager (CdosSettingsManager *manager)
{
        manager->priv->introspection_data = g_dbus_node_info_new_for_xml (introspection_xml, NULL);
        g_assert (manager->priv->introspection_data != NULL);

        g_bus_get (G_BUS_TYPE_SESSION,
                   NULL,
                   (GAsyncReadyCallback) on_bus_gotten,
                   manager);
}

gboolean
cdos_settings_manager_start (CdosSettingsManager *manager,
                              GError              **error)
{
        gboolean ret;

        g_debug ("Starting settings manager");

        ret = FALSE;

        cdos_settings_profile_start (NULL);

        if (!g_module_supported ()) {
                g_warning ("cdos-settings-daemon is not able to initialize the plugins.");
                g_set_error (error,
                             CDOS_SETTINGS_MANAGER_ERROR,
                             CDOS_SETTINGS_MANAGER_ERROR_GENERAL,
                             "Plugins not supported");

                goto out;
        }

        g_debug ("loading PNPIDs");
        manager->priv->pnp_ids = gnome_pnp_ids_new ();

        cdos_settings_profile_start ("initializing plugins");
        manager->priv->settings = g_settings_new (DEFAULT_SETTINGS_PREFIX ".plugins");

        _load_xsettings_xrandr (manager);
	//_load_all (manager);
        g_timeout_add (3000, (GSourceFunc)_load_all , manager);
        cdos_settings_profile_end ("initializing plugins");

        ret = TRUE;
 out:
        cdos_settings_profile_end (NULL);

        return ret;
}

void
cdos_settings_manager_stop (CdosSettingsManager *manager)
{
        g_debug ("Stopping settings manager");

        _unload_all (manager);

        if (manager->priv->owner_id > 0) {
                g_bus_unown_name (manager->priv->owner_id);
                manager->priv->owner_id = 0;
        }

        if (manager->priv->settings != NULL) {
                g_object_unref (manager->priv->settings);
                manager->priv->settings = NULL;
        }

        if (manager->priv->pnp_ids != NULL) {
                g_object_unref (manager->priv->pnp_ids);
                manager->priv->pnp_ids = NULL;
        }
}

static void
cdos_settings_manager_dispose (GObject *object)
{
        CdosSettingsManager *manager;

        manager = CDOS_SETTINGS_MANAGER (object);

        cdos_settings_manager_stop (manager);

        G_OBJECT_CLASS (cdos_settings_manager_parent_class)->dispose (object);
}

static void
cdos_settings_manager_class_init (CdosSettingsManagerClass *klass)
{
        GObjectClass   *object_class = G_OBJECT_CLASS (klass);

        object_class->dispose = cdos_settings_manager_dispose;
        object_class->finalize = cdos_settings_manager_finalize;

        g_type_class_add_private (klass, sizeof (CdosSettingsManagerPrivate));
}

static void
cdos_settings_manager_init (CdosSettingsManager *manager)
{

        manager->priv = CDOS_SETTINGS_MANAGER_GET_PRIVATE (manager);
}

static void
cdos_settings_manager_finalize (GObject *object)
{
        CdosSettingsManager *manager;

        g_return_if_fail (object != NULL);
        g_return_if_fail (CDOS_IS_SETTINGS_MANAGER (object));

        manager = CDOS_SETTINGS_MANAGER (object);

        g_return_if_fail (manager->priv != NULL);

        G_OBJECT_CLASS (cdos_settings_manager_parent_class)->finalize (object);
}

CdosSettingsManager *
cdos_settings_manager_new (void)
{
        if (manager_object != NULL) {
                g_object_ref (manager_object);
        } else {
                manager_object = g_object_new (CDOS_TYPE_SETTINGS_MANAGER,
                                               NULL);
                g_object_add_weak_pointer (manager_object,
                                           (gpointer *) &manager_object);
                register_manager (manager_object);
        }

        return CDOS_SETTINGS_MANAGER (manager_object);
}
