/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*-
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 *
 */
#if HAVE_WAYLAND
#include <wayland-client.h>
#endif

#include <glib/gi18n.h>
#include "csd-enums.h"
#include "cdos-settings-utils.h"

gboolean
csd_is_wayland_session (void)
{
#if HAVE_WAYLDND
    static gboolean is_wayland = FALSE;
    static gboolean inited = FALSE;
    if (inited)
        return is_wayland;
    inited = TRUE;
    struct wl_display *display;
    display = wl_display_connect (NULL);
    if (display)
    {
        wl_display_disconnect (display);
        is_wayland = TRUE;
    }
    return is_wayland;
#else
    return FALSE;
#endif
}

GDBusProxy *
csd_get_key_grabber_proxy (void)
{
    static GDBusProxy *proxy = NULL;
    GError *error = NULL;

    if (proxy)
        return proxy;
    else
    {
        proxy = g_dbus_proxy_new_for_bus_sync (G_BUS_TYPE_SESSION,
                                                G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START,
                                                NULL,
                                                "org.freedesktop.KeyGrabber",
                                                "/org/freedesktop/KeyGrabber",
                                                "org.freedesktop.KeyGrabber",
                                                NULL,
                                                &error);
        if (error)
        {
            g_warning ("%s: %s", __func__, error->message);
            g_error_free (error);
        }
        else
            g_object_add_weak_pointer (G_OBJECT (proxy), (gpointer *)&proxy);
        return proxy;
    }
}

char *
csd_settings_get_chassis_type (void)
{
    char *ret = NULL;
    GError *error = NULL;
    GVariant *inner;
    GVariant *variant = NULL;
    GDBusConnection *connection;

    connection = g_bus_get_sync (G_BUS_TYPE_SYSTEM,
                                NULL,
                                &error);
    if (connection == NULL) {
        g_warning ("system bus not available: %s", error->message);
        g_error_free (error);
        goto out;
    }

    variant = g_dbus_connection_call_sync (connection,
                                            "org.freedesktop.hostname1",
                                            "/org/freedesktop/hostname1",
                                            "org.freedesktop.DBus.Properties",
                                            "Get",
                                            g_variant_new ("(ss)",
                                                            "org.freedesktop.hostname1",
                                                            "Chassis"),
                                            NULL,
                                            G_DBUS_CALL_FLAGS_NONE,
                                            -1,
                                            NULL,
                                            &error);
    if (variant == NULL) {
        g_debug ("Failed to get property '%s': %s", "Chassis", error->message);
        g_error_free (error);
        goto out;
    }

    g_variant_get (variant, "(v)", &inner);
    ret = g_variant_dup_string (inner, NULL);
    g_variant_unref (inner);
out:
    g_clear_object (&connection);
    g_clear_pointer (&variant, g_variant_unref);
    return ret;
}

GDBusProxy *
csd_get_power_mode_proxy (void)
{
    static GDBusProxy *proxy = NULL;
    GError *error = NULL;

    if (proxy)
        return proxy;
    else
    {
        proxy = g_dbus_proxy_new_for_bus_sync (G_BUS_TYPE_SYSTEM,
                                                G_DBUS_PROXY_FLAGS_NONE,
                                                NULL,
                                                "com.nfs.KeyboardGrabber",
                                                "/com/nfs/PowerMode",
                                                "com.nfs.PowerMode",
                                                NULL,
                                                &error);
        if (error)
        {
            g_warning ("%s: %s", __func__, error->message);
            g_error_free (error);
        }
        else
            g_object_add_weak_pointer (G_OBJECT (proxy), (gpointer *)&proxy);
        return proxy;
    }
}

GDBusProxy *
csd_get_power_brightness_proxy (void)
{
    static GDBusProxy *proxy = NULL;
    GError *error = NULL;

    if (proxy)
        return proxy;
    else
    {
        proxy = g_dbus_proxy_new_for_bus_sync (G_BUS_TYPE_SYSTEM,
                                                G_DBUS_PROXY_FLAGS_NONE,
                                                NULL,
                                                "com.nfs.KeyboardGrabber",
                                                "/com/nfs/Brightness",
                                                "com.nfs.Brightness",
                                                NULL,
                                                &error);
        if (error)
        {
            g_warning ("%s: %s", __func__, error->message);
            g_error_free (error);
        }
        else
            g_object_add_weak_pointer (G_OBJECT (proxy), (gpointer *)&proxy);
        return proxy;
    }
}

GDBusProxy *
csd_get_shortcuts_proxy (void)
{
    static GDBusProxy *proxy = NULL;
    GError *error = NULL;

    if (proxy)
        return proxy;
    else
    {
        proxy = g_dbus_proxy_new_for_bus_sync (G_BUS_TYPE_SYSTEM,
                                                G_DBUS_PROXY_FLAGS_NONE,
                                                NULL,
                                                "com.nfs.KeyboardGrabber",
                                                "/com/nfs/Shortcuts",
                                                "com.nfs.Shortcuts",
                                                NULL,
                                                &error);
        if (error)
        {
            g_warning ("%s: %s", __func__, error->message);
            g_error_free (error);
        }
        else
            g_object_add_weak_pointer (G_OBJECT (proxy), (gpointer *)&proxy);
        return proxy;
    }
}

GDBusProxy *
csd_get_keyboard_backlight_proxy (void)
{
    static GDBusProxy *proxy = NULL;
    GError *error = NULL;

    if (proxy)
        return proxy;
    else
    {
        proxy = g_dbus_proxy_new_for_bus_sync (G_BUS_TYPE_SYSTEM,
                                                G_DBUS_PROXY_FLAGS_NONE,
                                                NULL,
                                                "com.nfs.KeyboardGrabber",
                                                "/com/nfs/KbdBacklightMode",
                                                "com.nfs.KbdBacklightMode",
                                                NULL,
                                                &error);
        if (error)
        {
            g_warning ("%s: %s", __func__, error->message);
            g_error_free (error);
        }
        else
            g_object_add_weak_pointer (G_OBJECT (proxy), (gpointer *)&proxy);
        return proxy;
    }
}

GDBusProxy *
csd_get_touchpad_mode_proxy (void)
{
    static GDBusProxy *proxy = NULL;
    GError *error = NULL;

    if (proxy)
        return proxy;
    else
    {
        proxy = g_dbus_proxy_new_for_bus_sync (G_BUS_TYPE_SYSTEM,
                                                G_DBUS_PROXY_FLAGS_NONE,
                                                NULL,
                                                "com.nfs.KeyboardGrabber",
                                                "/com/nfs/TouchpadMode",
                                                "com.nfs.TouchpadMode",
                                                NULL,
                                                &error);
        if (error)
        {
            g_warning ("%s: %s", __func__, error->message);
            g_error_free (error);
        }
        else
            g_object_add_weak_pointer (G_OBJECT (proxy), (gpointer *)&proxy);
        return proxy;
    }
}

GDBusProxy *
csd_get_session_proxy (void)
{
    static GDBusProxy *proxy = NULL;
    GError *error =  NULL;

    if (proxy != NULL) {
        g_object_ref (proxy);
    } else {
        proxy = g_dbus_proxy_new_for_bus_sync (G_BUS_TYPE_SESSION,
                                            G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START,
                                            NULL,
                                            GNOME_SESSION_DBUS_NAME,
                                            GNOME_SESSION_DBUS_PATH,
                                            GNOME_SESSION_DBUS_INTERFACE,
                                            NULL,
                                            &error);
        if (error) {
            g_warning ("Failed to connect to the session manager: %s", error->message);
            g_error_free (error);
        } else {
            g_object_add_weak_pointer (G_OBJECT (proxy), (gpointer*)&proxy);
        }
    }
    return proxy;
}

gboolean
g_settings_has_key (GSettings *gsettings, const gchar *key)
{
    g_return_val_if_fail (G_IS_SETTINGS (gsettings), FALSE);
    g_return_val_if_fail (key != NULL, FALSE);
    gchar **keys;
    keys = g_settings_list_keys (gsettings);
    gint i;
    gboolean has = FALSE;
    for (i = 0; keys[i] != NULL; i++)
    {
        if (g_strcmp0 (keys[i], key) == 0)
        {
            has = TRUE;
            break;
        }
    }
    g_strfreev (keys);
    return has;
}

gboolean
is_found_desktop_env (void)
{
    const gchar *current_desktop;
    current_desktop = g_getenv ("XDG_CURRENT_DESKTOP");
    if (current_desktop)
        current_desktop = g_ascii_strdown (current_desktop, -1);

    return g_strcmp0 (current_desktop, "found") == 0 || g_strcmp0 (current_desktop, "cdos") == 0 ? TRUE : FALSE;
}

gchar*
power_action_type_to_string (gint type)
{
    gchar *action = NULL;
    switch (type) {
    case CSD_POWER_ACTION_BLANK:
        action = _("lock screen");
        break;
    case CSD_POWER_ACTION_SUSPEND:
        action = _("suspend");
        break;
    case CSD_POWER_ACTION_SHUTDOWN:
        action = _("shutdown");
        break;
    case CSD_POWER_ACTION_HIBERNATE:
        action = _("hibernate");
        break;
    case CSD_POWER_ACTION_INTERACTIVE:
        action = _("interactive");
        break;
    case CSD_POWER_ACTION_NOTHING:
        break;
    }

    return action;
}