#include <glib.h>
#include <gio/gio.h>
#include <dbus/dbus-glib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>

#define CSD_DBUS_NAME             	"org.freedesktop.SettingsDaemon.Privacy"
#define CSD_DBUS_PATH                 	"/org/freedesktop/SettingsDaemon/Privacy"
#define CSD_DBUS_INTERFACE_CAMERA       "org.freedesktop.SettingsDaemon.Privacy.Camera"
#define CSD_DBUS_INTERFACE_MICPHONE    	"org.freedesktop.SettingsDaemon.Privacy.Micphone"
#define MODPROBE_FILE 			"/etc/modprobe.d/blacklist.conf"

GMainLoop *loop;
char *script_path = "/usr/lib/found-settings-daemon/privacy.sh";
char *script_close_camera = "/usr/lib/found-settings-daemon/privacy-close-camera.sh";

typedef struct {
    guint name_id;
    GCancellable   	*bus_cancellable;
    GDBusNodeInfo       *introspection_data;
    GDBusConnection     *connection;
} PrivacyManager;

static const gchar introspection_xml[] =
"<node>"
"   <interface name='org.freedesktop.SettingsDaemon.Privacy.Camera'>"
"       <property name='Disabled' type='b' access='read'/>"
"       <method name='SetDisabled'>"
"      		<arg type='b' name='disable' direction='in'/>"
"      		<arg type='b' name='ret' direction='out'/>"
"       </method>"
"       <signal name='DisabledStatusChanged'>"
"           <arg type='b' name='disabled'/>"
"       </signal>"
"   </interface>"
/*  TODO 麦克风权限状态控制接口
"   <interface name='org.freedesktop.SettingsDaemon.Privacy.Micphone'>"
"       <property name='Disabled' type='b' access='read'/>"
"       <method name='SetDisabled'>"
"      		<arg type='b' name='disable' direction='in'/>"
"      		<arg type='b' name='ret' direction='out'/>"
"       </method>"
"       <signal name='DisabledStatusChanged'>"
"           <arg type='b' name='disabled'/>"
"       </signal>"
"  </interface>"
*/
"</node>";

static gboolean 
set_camera_status (gboolean disable)
{
	GError *error = NULL;
	gint exit_status;
	gchar *argv[3] = {0};
	gboolean retval = TRUE;

	argv[0] = script_path;
	argv[1] = disable ? "disable" : "enable";

	g_spawn_sync (NULL,
		    argv,
		    NULL,
		    G_SPAWN_DEFAULT,
		    NULL,
		    NULL,
		    NULL,
		    NULL,
		    &exit_status,
		    &error);

	if (exit_status != 0)
	{
		g_warning ("failed to disable camera");
		retval = FALSE;
	}

	if (error)
	{
		g_warning ("disable camera:%s", error->message);
		g_clear_error (&error);
		retval = FALSE;
	}

	return retval;
}

static gboolean
timed_exit_cb (void)
{
        g_debug ("Doing timed exit");
        g_main_loop_quit (loop);
        return FALSE;
}

// 属性变化信号
static void
emit_changed (PrivacyManager *manager, gboolean disable, char * interface)
{
        GError *error = NULL;

    	g_dbus_connection_emit_signal (manager->connection,
                                            CSD_DBUS_NAME,
                                            CSD_DBUS_PATH,
                                            interface,
                                            "DisabledStatusChanged",
                                            g_variant_new ("(b)", disable),
                                            &error);
        if (error) {
                g_warning ("failed to emit Changed: %s", error->message);
                g_error_free (error);
        }
}

// 设置相机状态
static gboolean
set_camera_disable (PrivacyManager *manager, gboolean disable)
{
	gboolean retval = FALSE;

	if (!access (script_path, F_OK) &&
		set_camera_status (disable))
	{
			retval = TRUE;
			emit_changed (manager, disable, CSD_DBUS_INTERFACE_CAMERA);
	}

	return retval;
}

// TODO 设置麦克风状态
static gboolean 
set_mic_disable (PrivacyManager *manager, gboolean disable)
{
	return TRUE;
}

// 调用接口
static void
handle_method_call (GDBusConnection       *connection,
                    const gchar           *sender,
                    const gchar           *object_path,
                    const gchar           *interface_name,
                    const gchar           *method_name,
                    GVariant              *parameters,
                    GDBusMethodInvocation *invocation,
                    gpointer               user_data)
{
	gboolean retval = FALSE;
	gboolean disable = FALSE;
	PrivacyManager *manager = (PrivacyManager *)user_data;
	
	g_variant_get (parameters, "(b)", &disable);
	if (g_strcmp0 (interface_name, CSD_DBUS_INTERFACE_CAMERA) == 0) {
		if (g_strcmp0 (method_name, "SetDisabled") == 0) {
			retval = set_camera_disable (manager, disable);
			if (disable == TRUE && access (script_close_camera, F_OK) == 0)
				g_spawn_command_line_async (script_close_camera, NULL);
		}
	} else if (g_strcmp0 (interface_name, CSD_DBUS_INTERFACE_MICPHONE) == 0) {
		if (g_strcmp0 (method_name, "SetDisabled") == 0) {
			retval = set_mic_disable (manager, disable);
		}
	} else {
		g_warning ("not recognised interface: %s", interface_name);
	}

	g_debug ("Calling method '%s.%s' for Privacy, result is %d", interface_name, method_name, retval);
	g_dbus_method_invocation_return_value (invocation,
						g_variant_new ("(b)", retval));
}

// 获取相机驱动状态
static gboolean
read_camera_disabled_status ()
{
	char line[100];
	gboolean retval = FALSE;
	FILE *file;
	char *camera_driver = "blacklist uvcvideo";

	if (!(file = fopen (MODPROBE_FILE, "r"))) {
		g_warning ("failed to open %s", MODPROBE_FILE);
		return retval;
	}

	while (!feof (file)) {
		fgets (line, sizeof (line) - 1, file);
		if (strncmp (line, camera_driver, strlen (camera_driver)) == 0)
		{
			retval = TRUE;
			break;
		} 
	}

	fclose (file);
	g_debug ("read_camera_disabled_status: %d", retval);
	return retval;
}

// 获取属性
static GVariant *
handle_get_property (GDBusConnection *connection,
                     const gchar *sender,
                     const gchar *object_path,
                     const gchar *interface_name,
                     const gchar *property_name,
                     GError **error, gpointer user_data)
{
	GVariant *retval = NULL;

	// 相机
	if (g_strcmp0(interface_name, CSD_DBUS_INTERFACE_CAMERA) == 0 ||
	    g_strcmp0 (property_name, "Disabled") == 0) 
	{
	    	retval = g_variant_new_boolean (read_camera_disabled_status ());
	}

	// 麦克风
	if (g_strcmp0 (interface_name, CSD_DBUS_INTERFACE_MICPHONE) == 0 ||
	    g_strcmp0 (property_name, "Disabled") == 0) 
    {
		// TODO 获取麦克风状态
	} 

    if (!retval)
	    g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_FAILED,
		             "No such interface: %s or property %s", interface_name, property_name);

    return retval;
}

static void
on_name_acquired (GDBusConnection *connection,
                  const gchar     *name,
                  gpointer         user_data)
{
    	g_debug ("Acquired the name %s\n", name);
}

static void
on_name_lost (GDBusConnection *connection,
              const gchar     *name,
              gpointer         user_data)
{
	g_debug ("Lost the name %s\n", name);
}

static const GDBusInterfaceVTable interface_vtable =
{
	handle_method_call,
	handle_get_property,
	NULL, /* SetProperty */
};

static void
on_bus_gotten (GObject *source_object,
	    GAsyncResult        *res,
	    PrivacyManager     *manager)
{
        GDBusInterfaceInfo **infos;
        GError *error = NULL;
        guint i;

        if (manager->bus_cancellable == NULL ||
                g_cancellable_is_cancelled (manager->bus_cancellable)) {
                g_warning ("Operation has been cancelled, so not retrieving system bus");
                return;
        }

        manager->connection = g_bus_get_finish (res, &error);
        if (manager->connection == NULL) {
                g_warning ("Could not get system bus: %s", error->message);
                g_error_free (error);
                return;
        }

        infos = manager->introspection_data->interfaces;
        for (i = 0; infos[i] != NULL; i++) {
                guint id = g_dbus_connection_register_object (manager->connection,
                                                   CSD_DBUS_PATH,
                                                   infos[i],
                                                   &interface_vtable,
                                                   manager,
                                                   NULL,
                                                   NULL);
                g_debug ("registe id = %d", id);
                g_assert (id > 0);
        }

        manager->name_id = g_bus_own_name_on_connection (manager->connection, 
                                                                                    CSD_DBUS_NAME,
                                                                                    G_BUS_NAME_OWNER_FLAGS_ALLOW_REPLACEMENT | G_BUS_NAME_OWNER_FLAGS_REPLACE,
                                                                                    on_name_acquired,
                                                                                    on_name_lost,
                                                                                    manager,
                                                                                    NULL);
}

// 注册服务
static void
register_manager_dbus (PrivacyManager *manager)
{
        manager->introspection_data = g_dbus_node_info_new_for_xml (introspection_xml, NULL);
        manager->bus_cancellable = g_cancellable_new ();
        g_assert (manager->introspection_data != NULL);

        g_bus_get (G_BUS_TYPE_SYSTEM,
                   manager->bus_cancellable,
                   (GAsyncReadyCallback) on_bus_gotten,
                   manager);
}

int main(int argc, char *argv[])
{
	PrivacyManager manager;
	loop = g_main_loop_new (NULL, TRUE);

	// 注册服务
	register_manager_dbus (&manager);

	// 1分钟后退出
	g_timeout_add_seconds (60, (GSourceFunc) timed_exit_cb, NULL);
	g_main_loop_run (loop);

	if (manager.name_id != 0)
		g_bus_unown_name (manager.name_id);
	if (manager.connection)
		g_object_unref (manager.connection);
	if (manager.introspection_data)
		g_dbus_node_info_unref (manager.introspection_data);
	if (manager.bus_cancellable) {
         	g_cancellable_is_cancelled (manager.bus_cancellable);
		g_object_unref (manager.bus_cancellable);
	}

	g_main_loop_unref (loop);
	return 0;
}

