/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*-
 *
 * Copyright © 2001 Ximian, Inc.
 * Copyright (C) 2007 William Jon McCann <mccann@jhu.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street - Suite 500, Boston, MA 02110-1335, USA.
 *
 */

#include "config.h"

#include <sys/types.h>
#include <sys/wait.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include <locale.h>

#include <glib.h>
#include <glib/gi18n.h>
#include <gdk/gdk.h>
#include <gtk/gtk.h>
#include <atspi/atspi.h>

#include "cdos-settings-profile.h"
#include "csd-a11y-keyboard-manager.h"

#define KEYBOARD_A11Y_SCHEMA "org.cdos.desktop.a11y.keyboard"
#define OSK_SCHEMA "org.cdos.desktop.a11y.applications"
#define OSK_KEY "screen-keyboard-enabled"
#define CDOS_KEYBOARD_SCHEMA "org.cdos.keyboard"
#define ATUO_HIDE_KEY "auto-hide"

#define CSD_A11Y_KEYBOARD_MANAGER_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), CSD_TYPE_A11Y_KEYBOARD_MANAGER, CsdA11yKeyboardManagerPrivate))

struct CsdA11yKeyboardManagerPrivate
{
        guint             start_idle_id;
        GtkWidget        *preferences_dialog;

        GSettings        *settings;
        GSettings        *osk_settings;
        GSettings        *cdos_kb_settings;
        gboolean          osk_enabled;
        gboolean          osk_shown;
        gboolean          auto_hide;
        AtspiRect        *cursor_location;

        AtspiEventListener  *listener;
        AtspiAccessible     *accessible;

        GDBusProxy          *keyboard_proxy;
        guint                cursor_location_changed_id;
};

static void     csd_a11y_keyboard_manager_class_init  (CsdA11yKeyboardManagerClass *klass);
static void     csd_a11y_keyboard_manager_init        (CsdA11yKeyboardManager      *a11y_keyboard_manager);
static void     csd_a11y_keyboard_manager_finalize    (GObject             *object);
static void     csd_a11y_keyboard_manager_dispose     (GObject *object);

G_DEFINE_TYPE_WITH_PRIVATE (CsdA11yKeyboardManager, csd_a11y_keyboard_manager, G_TYPE_OBJECT)

static gpointer manager_object = NULL;

static void
keyboard_callback (GSettings *settings,
                   gchar *key,
                   CsdA11yKeyboardManager *manager)
{

}

static gboolean
on_cursor_location_changed_timeout (CsdA11yKeyboardManager *self)
{
    self->priv->cursor_location_changed_id = 0;
    AtspiRect *rect = self->priv->cursor_location;
    if (!rect)
        return G_SOURCE_REMOVE;

    g_dbus_proxy_call (self->priv->keyboard_proxy,
                       "SetCursorLocation",
                       g_variant_new ("(iiii)",
                                      rect->x, rect->y, rect->width, rect->height),
                       G_DBUS_CALL_FLAGS_NONE,
                       -1,
                       NULL, NULL, NULL);
    return G_SOURCE_REMOVE;
}

static gboolean
is_focused (AtspiAccessible *accessible)
{
    gboolean focused;
    if (!accessible)
        return FALSE;

    AtspiStateSet *state_set = atspi_accessible_get_state_set (accessible);
    focused = atspi_state_set_contains (state_set, ATSPI_STATE_FOCUSED);
    g_object_unref (state_set);
    return focused;
}

static gboolean
get_enabled_property (CsdA11yKeyboardManager *self)
{
    gboolean enabled = FALSE;
    GVariant *v = g_dbus_proxy_get_cached_property (self->priv->keyboard_proxy,
                                                    "Enabled");
    if (v)
    {
        enabled = g_variant_get_boolean (v);
        g_variant_unref (v);
    }
    return enabled;
}

static void
on_event (AtspiEvent *event,
          CsdA11yKeyboardManager *self)
{
    AtspiAccessible *application = NULL;
    AtspiAccessible *accessible;
    gboolean focus_in = FALSE;
    gboolean editable = FALSE;
    g_autofree gchar *app_name;
    gboolean should_hide = FALSE;

    //manager->priv->osk_enabled = enabled;
    accessible = event->source;
    if (!accessible)
        return;

    application = atspi_accessible_get_application (accessible, NULL);
    if (application == NULL)
        return;

    app_name = atspi_accessible_get_name (application, NULL);
    /* cdos-desktop 事件由输入法状态判断 */
    if (g_strcmp0 (app_name, "cdos-desktop") == 0)
    {
        return;
    }

    guint time = g_get_real_time () / 1000;
    if (g_str_has_suffix (event->type, "focused"))
    {
        /*
         * detail1: focus in
         * detail2: focus out
         */
        focus_in = event->detail1;
        if (focus_in)
        {
            self->priv->accessible = accessible;
            AtspiStateSet *state_set = atspi_accessible_get_state_set (accessible);
            AtspiRole role = atspi_accessible_get_role (accessible, NULL);
            if (role == ATSPI_ROLE_TERMINAL ||
                atspi_state_set_contains (state_set, ATSPI_STATE_EDITABLE))
            {
                editable = TRUE;
            }
            else
                should_hide = TRUE;
            g_object_unref (state_set);
        }
        else
        {
            self->priv->accessible = NULL;
            should_hide = TRUE;
        }
    }
    else if (g_str_has_suffix (event->type, "insert") ||
             g_str_has_suffix (event->type, "delete") ||
             g_str_has_suffix (event->type, "text-caret-moved"))
    {
        if (self->priv->accessible == application)
        {
            if (!get_enabled_property (self))
                editable = TRUE;
        }
        else
        {
            /* 非焦点窗口 */
            return;
        }
    }
    /*
    if (self->priv->osk_shown == editable)
        return;
    g_debug ("%s: app: %s, event_type: %s, show keyboard=%d, focus_in:%d",
             __func__, app_name, event->type, editable, focus_in);
    */

    self->priv->osk_shown = editable;
    if (should_hide && self->priv->auto_hide)
    {
        if (get_enabled_property (self))
            g_dbus_proxy_call (self->priv->keyboard_proxy,
                               "Hide",
                               g_variant_new ("(u)", time),
                               G_DBUS_CALL_FLAGS_NONE,
                               -1,
                               NULL, NULL, NULL);
        if (self->priv->cursor_location_changed_id)
        {
            g_source_remove (self->priv->cursor_location_changed_id);
            self->priv->cursor_location_changed_id = 0;
        }
        if (self->priv->cursor_location)
        {
            g_free (self->priv->cursor_location);
            self->priv->cursor_location = NULL;
        }
    }
    else
    {
        if (editable)
        {
            g_dbus_proxy_call (self->priv->keyboard_proxy,
                               "Show",
                               g_variant_new ("(u)", time),
                               G_DBUS_CALL_FLAGS_NONE,
                               -1,
                               NULL, NULL, NULL);
        }
        if (editable || is_focused (accessible))
        {
            GError *error = NULL;
            AtspiRect *rect;
            AtspiText *text = atspi_accessible_get_text (accessible);
            if (text)
            {
                rect = atspi_text_get_character_extents (text, event->detail1, ATSPI_COORD_TYPE_SCREEN, &error);
                if (error)
                {
                    g_warning ("%s: %s", __func__, error->message);
                    g_error_free (error);
                }
                else
                {
                    if (!self->priv->cursor_location ||
                        rect->y != self->priv->cursor_location->y)
                    {
                        g_debug ("text coord: %d %d %d %d\n", rect->x, rect->y, rect->width, rect->height);
                        g_free (self->priv->cursor_location);
                        self->priv->cursor_location = rect;

                        if (self->priv->cursor_location_changed_id)
                        {
                            g_source_remove (self->priv->cursor_location_changed_id);
                        }
                        self->priv->cursor_location_changed_id = g_timeout_add (100,
                                                                                (GSourceFunc)on_cursor_location_changed_timeout,
                                                                                self);
                    }
                    else
                    {
                        g_free (rect);
                    }
                }
            }
        }
    }
}

static void
on_keyboard_proxy_ready (GObject *source,
                         GAsyncResult *res,
                         CsdA11yKeyboardManager *manager)
{
    GError *error = NULL;
    manager->priv->keyboard_proxy = g_dbus_proxy_new_for_bus_finish (res, &error);
    if (error)
    {
        g_warning ("%s: %s", __func__, error->message);
        return;
    }
    if (!manager->priv->listener)
    {
        int ret = atspi_init ();
        if (ret == 2)
        {
            g_warning ("%s: Create a11y bus failed", __func__);
            return;
        }
        manager->priv->listener = atspi_event_listener_new ((AtspiEventListenerCB)on_event, manager, NULL);
        atspi_event_listener_register (manager->priv->listener, "object:state-changed:focused", NULL);
        /*
           atspi_event_listener_register (self->listener, "object:state-changed:selected", NULL);
           */
        atspi_event_listener_register (manager->priv->listener, "object:text-changed:insert", NULL);
        atspi_event_listener_register (manager->priv->listener, "object:text-changed:delete", NULL);
        atspi_event_listener_register (manager->priv->listener, "object:text-caret-moved", NULL);
    }
}

static void
cleanup (CsdA11yKeyboardManager *manager)
{
    if (manager->priv->listener)
    {
        atspi_event_listener_deregister (manager->priv->listener, "object:state-changed:focused", NULL);
        /*
           atspi_event_listener_register (self->listener, "object:state-changed:selected", NULL);
           */
        atspi_event_listener_deregister (manager->priv->listener, "object:text-changed:insert", NULL);
        atspi_event_listener_deregister (manager->priv->listener, "object:text-changed:delete", NULL);
        atspi_event_listener_deregister (manager->priv->listener, "object:text-caret-moved", NULL);
        g_clear_object (&manager->priv->listener);
        //atspi_exit ();
    }
    if (manager->priv->cursor_location_changed_id)
    {
        g_source_remove (manager->priv->cursor_location_changed_id);
        manager->priv->cursor_location_changed_id = 0;
    }
    g_clear_pointer (&manager->priv->cursor_location, g_free);
    g_clear_object (&manager->priv->keyboard_proxy);
}

static void
osk_keyboard_callback (CsdA11yKeyboardManager *manager)
{
    gboolean enabled;
    enabled = g_settings_get_boolean (manager->priv->osk_settings,
                                                         "screen-keyboard-enabled");
    manager->priv->osk_enabled = enabled;
    manager->priv->osk_shown = FALSE;
    if (!enabled)
    {
        cleanup (manager);
    }
    if (enabled)
    {
        if (!manager->priv->keyboard_proxy)
        {
            g_dbus_proxy_new_for_bus (G_BUS_TYPE_SESSION,
                                      G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START,
                                      NULL,
                                      "org.Cdos",
                                      "/org/gnome/Caribou/Keyboard",
                                      "org.gnome.Caribou.Keyboard",
                                      NULL,
                                      (GAsyncReadyCallback)on_keyboard_proxy_ready,
                                      manager);
        }
    }
}

static void
on_auto_hide_changed_callback (CsdA11yKeyboardManager *manager)
{
    gboolean auto_hide;
    auto_hide = g_settings_get_boolean (manager->priv->cdos_kb_settings,
                                        "auto-hide");
    manager->priv->auto_hide = auto_hide;
}

static gboolean
start_a11y_keyboard_idle_cb (CsdA11yKeyboardManager *manager)
{
        g_debug ("Starting a11y_keyboard manager");
        cdos_settings_profile_start (NULL);

        manager->priv->settings = g_settings_new (KEYBOARD_A11Y_SCHEMA);
        g_signal_connect (manager->priv->settings, "changed",
                          G_CALLBACK (keyboard_callback), manager);

        manager->priv->osk_settings = g_settings_new (OSK_SCHEMA);
        g_signal_connect_swapped (manager->priv->osk_settings, "changed::screen-keyboard-enabled",
                                  G_CALLBACK (osk_keyboard_callback), manager);
        manager->priv->cdos_kb_settings = g_settings_new (CDOS_KEYBOARD_SCHEMA);
        g_signal_connect_swapped (manager->priv->cdos_kb_settings, "changed::auto-hide",
                                  G_CALLBACK (on_auto_hide_changed_callback), manager);
        on_auto_hide_changed_callback (manager);
        osk_keyboard_callback (manager);
        cdos_settings_profile_end (NULL);

        manager->priv->start_idle_id = 0;

        return FALSE;
}

gboolean
csd_a11y_keyboard_manager_start (CsdA11yKeyboardManager *manager,
                                 GError                **error)
{
        cdos_settings_profile_start (NULL);

        manager->priv->start_idle_id = g_idle_add ((GSourceFunc) start_a11y_keyboard_idle_cb, manager);

        cdos_settings_profile_end (NULL);

        return TRUE;
}

void
csd_a11y_keyboard_manager_stop (CsdA11yKeyboardManager *manager)
{
        CsdA11yKeyboardManagerPrivate *p = manager->priv;

        g_debug ("Stopping a11y_keyboard manager");

        if (p->start_idle_id != 0) {
                g_source_remove (p->start_idle_id);
                p->start_idle_id = 0;
        }

        cleanup (manager);
}

static void
csd_a11y_keyboard_manager_class_init (CsdA11yKeyboardManagerClass *klass)
{
        GObjectClass *object_class = G_OBJECT_CLASS (klass);

        object_class->dispose = csd_a11y_keyboard_manager_dispose;
        object_class->finalize = csd_a11y_keyboard_manager_finalize;
}

static void
csd_a11y_keyboard_manager_dispose (GObject *object)
{
        CsdA11yKeyboardManager *manager;

        manager = CSD_A11Y_KEYBOARD_MANAGER (object);
        g_clear_object (&manager->priv->settings);
        g_clear_object (&manager->priv->osk_settings);
        g_clear_object (&manager->priv->cdos_kb_settings);
        cleanup (manager);

        G_OBJECT_CLASS (csd_a11y_keyboard_manager_parent_class)->dispose (object);
}

static void
csd_a11y_keyboard_manager_init (CsdA11yKeyboardManager *manager)
{
        manager->priv = CSD_A11Y_KEYBOARD_MANAGER_GET_PRIVATE (manager);
}

static void
csd_a11y_keyboard_manager_finalize (GObject *object)
{
        CsdA11yKeyboardManager *a11y_keyboard_manager;

        g_return_if_fail (object != NULL);
        g_return_if_fail (CSD_IS_A11Y_KEYBOARD_MANAGER (object));

        a11y_keyboard_manager = CSD_A11Y_KEYBOARD_MANAGER (object);

        g_return_if_fail (a11y_keyboard_manager->priv != NULL);

        if (a11y_keyboard_manager->priv->start_idle_id != 0) {
                g_source_remove (a11y_keyboard_manager->priv->start_idle_id);
                a11y_keyboard_manager->priv->start_idle_id = 0;
        }

        G_OBJECT_CLASS (csd_a11y_keyboard_manager_parent_class)->finalize (object);
}

CsdA11yKeyboardManager *
csd_a11y_keyboard_manager_new (void)
{
        if (manager_object != NULL) {
                g_object_ref (manager_object);
        } else {
                manager_object = g_object_new (CSD_TYPE_A11Y_KEYBOARD_MANAGER, NULL);
                g_object_add_weak_pointer (manager_object,
                                           (gpointer *) &manager_object);
        }

        return CSD_A11Y_KEYBOARD_MANAGER (manager_object);
}
