/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*-
 *
 * Copyright (C) 2001-2003 Bastien Nocera <hadess@hadess.net>
 * Copyright (C) 2006-2007 William Jon McCann <mccann@jhu.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street - Suite 500, Boston, MA 02110-1335, USA.
 *
 */

#include "config.h"

#include <sys/types.h>
#include <sys/wait.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <math.h>

#include <locale.h>

#include <glib.h>
#include <glib/gi18n.h>
#include <gio/gio.h>
#include <gdk/gdk.h>
#include <gdk/gdkkeysyms.h>
#include <gdk/gdkx.h>
#include <gtk/gtk.h>
#include <gio/gdesktopappinfo.h>
#include <gio/gunixfdlist.h>

#ifdef HAVE_GUDEV
#include <gudev/gudev.h>
#endif
#include <gdesktop-enums.h>

#include "csd-enums.h"
#include "mpris-controller.h"
#include "cdos-settings-profile.h"
#include "cdos-settings-utils.h"
#include "csd-marshal.h"
#include "csd-media-keys-manager.h"

#include "shortcuts-list.h"
#include "csd-osd-window.h"
#include "csd-input-helper.h"
#include "csd-power-helper.h"

#include <canberra.h>
#include <pulse/pulseaudio.h>
#include "gvc-mixer-control.h"

/* For media keys, we need to keep using org.gnome because
   that's what apps are looking for */
#define CSD_DBUS_PATH "/org/freedesktop/SettingsDaemon"
#define CSD_DBUS_NAME "org.freedesktop.SettingsDaemon"

#define CSD_MEDIA_KEYS_DBUS_PATH CSD_DBUS_PATH "/MediaKeys"
#define CSD_MEDIA_KEYS_DBUS_NAME CSD_DBUS_NAME ".MediaKeys"

#define GNOME_SESSION_DBUS_NAME "org.gnome.SessionManager"
#define GNOME_SESSION_DBUS_PATH "/org/gnome/SessionManager"
#define GNOME_SESSION_DBUS_INTERFACE "org.gnome.SessionManager"

#define GNOME_KEYRING_DBUS_NAME "org.gnome.keyring"
#define GNOME_KEYRING_DBUS_PATH "/org/gnome/keyring/daemon"
#define GNOME_KEYRING_DBUS_INTERFACE "org.gnome.keyring.Daemon"

#define SCREENSAVER_DBUS_NAME "org.freedesktop.ScreenSaver"
#define SCREENSAVER_DBUS_PATH "/org/freedesktop/ScreenSaver"
#define SCREENSAVER_DBUS_INTERFACE "org.freedesktop.ScreenSaver"

#define CUSTOM_BINDING_SCHEMA SETTINGS_BINDING_DIR ".custom-keybinding"

#define SETTINGS_INTERFACE_DIR "org.cdos.desktop.interface"
#define SETTINGS_POWER_DIR "org.cdos.settings-daemon.plugins.power"
#define SETTINGS_XSETTINGS_DIR "org.cdos.settings-daemon.plugins.xsettings"
#define SETTINGS_TOUCHPAD_DIR "org.gnome.desktop.peripherals.touchpad"
#define TOUCHPAD_ENABLED_KEY "send-events"
#define SETTINGS_CAMERA_DIR "org.gnome.desktop.privacy"
#define CAMERA_DISABLED_KEY "disable-camera"

#define HIGH_CONTRAST "HighContrast"

#define VOLUME_STEP 5           /* percents for one volume button press */
/*
 * PA_VOLUME_UI_MAX will set volume to 152%, so change
 * PA_VOLUME_UI_MAX to PA_VOLUME_NORM
 */
#define MAX_VOLUME PA_VOLUME_NORM

#define SYSTEMD_DBUS_NAME                       "org.freedesktop.login1"
#define SYSTEMD_DBUS_PATH                       "/org/freedesktop/login1"
#define SYSTEMD_DBUS_INTERFACE                  "org.freedesktop.login1.Manager"

#define CSD_MEDIA_KEYS_MANAGER_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), CSD_TYPE_MEDIA_KEYS_MANAGER, CsdMediaKeysManagerPrivate))

typedef struct {
        char   *application;
        char   *name;
        guint32 time;
        guint   watch_id;
} MediaPlayer;

typedef struct {
        gint ref_count;
        MediaKeyType key_type;
        ShellActionMode modes;
        MetaKeyBindingFlags grab_flags;
        const char *settings_key;
        gboolean static_setting;
        const char *hard_coded;
        char *custom_path;
        char *custom_command;
        GArray *accel_ids;
} MediaKey;

 typedef struct {
        CsdMediaKeysManager *manager;
        GPtrArray *keys;

        /* NOTE: This is to implement a custom cancellation handling where
         *       we immediately emit an ungrab call if grabbing was cancelled.
         */
        gboolean cancelled;
} GrabUngrabData;

struct CsdMediaKeysManagerPrivate
{
        /* Volume bits */
        GvcMixerControl *volume;
        GvcMixerStream  *stream; /* sink */
        GvcMixerStream  *source;
        ca_context      *ca;
        GtkSettings     *gtksettings;
#ifdef HAVE_GUDEV
        GHashTable      *streams; /* key = X device ID, value = stream id */
        GUdevClient     *udev_client;
#endif /* HAVE_GUDEV */

        GtkWidget       *dialog;
        GtkWidget       *dialog_old;
        GSettings       *settings;                              // org.cdos.settings-daemon.plugins.media-keys配置
        GHashTable      *custom_settings;

        GPtrArray       *keys;

        /* HighContrast theme settings */
        GSettings       *interface_settings;
        char            *icon_theme;
        char            *gtk_theme;

        /* Power stuff */
        GSettings       *power_settings;
        GDBusProxy      *upower_proxy;        		// Upower接口代理
        GDBusProxy      *power_screen_proxy;       	// 电源管理屏幕接口代理
        GDBusProxy      *power_keyboard_proxy;  	// 电源管理键盘接口代理
        GDBusProxy      *desktop_proxy;               	// 桌面接口代理
        GDBusProxy      *rfkill_proxy;              	// rfkill接口代理
        GDBusProxy      *key_grabber;                	// KeyGrabber接口代理
        GCancellable    *grab_cancellable;
        GHashTable      *keys_to_sync;
        guint            keys_sync_source_id;
        GrabUngrabData  *keys_sync_data;
        gboolean        screensaver_lock;
        guint           airplane_action_id;
        guint           caps_lock_action_id;
        guint           num_lock_action_id;
        gboolean        airplane_mode;
        gboolean        airplane_changed;
        gboolean        key_grabbed;


        /* systemd stuff */
        GDBusProxy      *logind_proxy;              	// login1接口代理
        gint             inhibit_keys_fd;

        /* Multihead stuff */
        GdkDisplay      *display;

        GList           *media_players;

        GDBusConnection *connection;
        GCancellable    *bus_cancellable;
        GDBusNodeInfo    *introspection_data;
        GCancellable    *cancellable;

        guint            start_idle_id;

        MprisController *mpris_controller;

        GDBusProxy      *power_mode;                  	// com.nfs.PowerMode接口代理
        gboolean        *power_mode_enable;
        GDBusProxy      *power_brightness;            	// com.nfs.Brightness接口代理
        GDBusProxy      *shortcuts;			// com.nfs.Shortcuts接口代理
        GDBusProxy      *keyboard_backlight;            // com.nfs.KbdBacklightMode接口代理
        GDBusProxy      *touchpad_mode;                 // com.nfs.TouchpadMode接口代理

        guint            monitors_changed_id;
};

static void     csd_media_keys_manager_class_init  (CsdMediaKeysManagerClass *klass);
static void     csd_media_keys_manager_init        (CsdMediaKeysManager      *media_keys_manager);
static void     csd_media_keys_manager_finalize    (GObject                  *object);
static void     register_manager                   (CsdMediaKeysManager      *manager);
static void     custom_binding_changed             (GSettings           *settings,
                                                    const char          *settings_key,
                                                    CsdMediaKeysManager *manager);
static void     keys_sync_queue                    (CsdMediaKeysManager *manager,
                                                    gboolean             immediate,
                                                    gboolean             retry);
static void     keys_sync_continue                 (CsdMediaKeysManager *manager);

static gboolean do_action (CsdMediaKeysManager *manager,
                           const gchar         *device_node,
                           MediaKeyType         type,
                           gint64               timestamp);

G_DEFINE_TYPE_WITH_PRIVATE (CsdMediaKeysManager, csd_media_keys_manager, G_TYPE_OBJECT)

static gpointer manager_object = NULL;

static const char *volume_icons[] = {
        "notification-audio-volume-muted",
        "notification-audio-volume-low",
        "notification-audio-volume-medium",
        "notification-audio-volume-high",
        NULL
};

static const char *brightness_icons[] = {
        "notification-display-brightness-off",
        "notification-display-brightness-low",
        "notification-display-brightness-medium",
        "notification-display-brightness-high",
        "notification-display-brightness-full",
        NULL
};

static const char *kb_backlight_icons[] = {
        "notification-keyboard-brightness-off",
        "notification-keyboard-brightness-low",
        "notification-keyboard-brightness-medium",
        "notification-keyboard-brightness-high",
        "notification-keyboard-brightness-full",
        NULL
};


static const gchar introspection_xml[] =
"<node>"
"  <interface name='org.freedesktop.SettingsDaemon.MediaKeys'>"
"       <method name='VolumeStepDown'>"
"         <arg name='volume' type='i' direction='out' />"
"       </method>"
"       <method name='VolumeStepUp'>"
"         <arg name='volume' type='i' direction='out' />"
"       </method>"
"  </interface>"
"</node>";

static const char *
calculate_icon_name (gint value, const char **icon_names)
{
        value = CLAMP (value, 0, 100);
        gint length = g_strv_length ((gchar **)icon_names);
        gint s = (length - 1) * value / 100 + 1;
        s = CLAMP (s, 1, length - 1);

        return icon_names[s];
}

static void
media_key_unref (MediaKey *key)
{
        if (key == NULL)
                return;
        if (!g_atomic_int_dec_and_test (&key->ref_count))
                return;
        g_clear_pointer (&key->accel_ids, g_array_unref);
        g_free (key->custom_path);
        g_free (key->custom_command);
        g_free (key);
}

static MediaKey *
media_key_ref (MediaKey *key)
{
        g_atomic_int_inc (&key->ref_count);
        return key;
}

static MediaKey *
media_key_new (void)
{
        MediaKey *key = g_new0 (MediaKey, 1);

        key->accel_ids = g_array_new (FALSE, TRUE, sizeof(guint));

        return media_key_ref (key);
}

G_DEFINE_AUTOPTR_CLEANUP_FUNC (MediaKey, media_key_unref)

static void
grab_ungrab_data_free (GrabUngrabData *data)
{
        /* NOTE: The manager pointer is not owned and is invalid if the
         *       operation was cancelled.
         */

        if (!data->cancelled) {
                CsdMediaKeysManagerPrivate *priv = CSD_MEDIA_KEYS_MANAGER_GET_PRIVATE (data->manager);

                if (priv->keys_sync_data == data)
                        priv->keys_sync_data = NULL;
        }

        data->manager = NULL;
        g_clear_pointer (&data->keys, g_ptr_array_unref);
        g_free (data);
}

G_DEFINE_AUTOPTR_CLEANUP_FUNC (GrabUngrabData, grab_ungrab_data_free)

static char *
get_term_command (CsdMediaKeysManager *manager)
{
        char *cmd_term, *cmd_args;;
        char *cmd = NULL;
        GSettings *settings;

        settings = g_settings_new ("org.cdos.desktop.default-applications.terminal");
        cmd_term = g_settings_get_string (settings, "exec");
        if (cmd_term[0] == '\0')
                cmd_term = g_strdup ("gnome-terminal");

        cmd_args = g_settings_get_string (settings, "exec-arg");
        if (strcmp (cmd_term, "") != 0) {
                cmd = g_strdup_printf ("%s %s -e", cmd_term, cmd_args);
        } else {
                cmd = g_strdup_printf ("%s -e", cmd_term);
        }

        g_free (cmd_args);
        g_free (cmd_term);
        g_object_unref (settings);

        return cmd;
}

static char **
get_keyring_env (CsdMediaKeysManager *manager)
{
        GError *error = NULL;
        GVariant *variant, *item;
        GVariantIter *iter;
        char **envp;

        variant = g_dbus_connection_call_sync (manager->priv->connection,
                                               GNOME_KEYRING_DBUS_NAME,
                                               GNOME_KEYRING_DBUS_PATH,
                                               GNOME_KEYRING_DBUS_INTERFACE,
                                               "GetEnvironment",
                                               NULL,
                                               NULL,
                                               G_DBUS_CALL_FLAGS_NONE,
                                               -1,
                                               NULL,
                                               &error);
        if (variant == NULL) {
                g_warning ("Failed to call GetEnvironment on keyring daemon: %s", error->message);
                g_error_free (error);
                return NULL;
        }

        envp = g_get_environ ();

        g_variant_get (variant, "(a{ss})", &iter);

        while ((item = g_variant_iter_next_value (iter))) {
                char *key;
                char *value;

                g_variant_get (item,
                               "{ss}",
                               &key,
                               &value);

                envp = g_environ_setenv (envp, key, value, TRUE);

                g_variant_unref (item);
                g_free (key);
                g_free (value);
        }

        g_variant_iter_free (iter);
        g_variant_unref (variant);

        return envp;
}

static void
execute (CsdMediaKeysManager *manager,
         const char          *cmd,
         gint64               timestamp)
{
        gboolean retval;
        char   **argv;
        int      argc;
        GError  *error = NULL;

        retval = FALSE;

        if (g_shell_parse_argv (cmd, &argc, &argv, NULL)) {
                char   **envp;

                envp = get_keyring_env (manager);

                retval = g_spawn_async (g_get_home_dir (),
                                        argv,
                                        envp,
                                        G_SPAWN_SEARCH_PATH,
                                        NULL,
                                        NULL,
                                        NULL,
                                        &error);

                g_strfreev (argv);
                g_strfreev (envp);
        }

        if (retval == FALSE) {
                g_warning ("Couldn't execute command: %s: %s", cmd, error->message);
                g_error_free (error);
        }
}

static void
dialog_init (CsdMediaKeysManager *manager)
{
        if (manager->priv->dialog != NULL
            && !csd_osd_window_is_valid (CSD_OSD_WINDOW (manager->priv->dialog))) {
                manager->priv->dialog_old = manager->priv->dialog;
                manager->priv->dialog = NULL;
        }

        if (manager->priv->dialog == NULL) {
                manager->priv->dialog = csd_osd_window_new ();
        }
}

static char *
get_key_string (MediaKey *key)
{
    if (key->settings_key != NULL)
        return g_strdup_printf ("settings:%s", key->settings_key);
    else if (key->custom_path != NULL)
        return g_strdup_printf ("custom:%s", key->custom_path);
    else
        g_assert_not_reached ();
}

static GStrv
get_bindings (CsdMediaKeysManager *manager,
              MediaKey            *key)
{
    CsdMediaKeysManagerPrivate *priv = CSD_MEDIA_KEYS_MANAGER_GET_PRIVATE (manager);
    GPtrArray *array;
    gchar *binding;

    if (key->settings_key != NULL) {
        g_autofree GStrv keys = NULL;
        g_autofree GStrv static_keys = NULL;
        gchar **item;

        if (!key->static_setting)
            return g_settings_get_strv (priv->settings, key->settings_key);

        keys = g_settings_get_strv (priv->settings, key->settings_key);
        static_keys = g_strsplit (key->hard_coded, ",", -1);

        array = g_ptr_array_new ();
        /* Steals all strings from the settings */
        for (item = keys; *item; item++)
            g_ptr_array_add (array, *item);
        for (item = static_keys; *item; item++)
            g_ptr_array_add (array, *item);
        g_ptr_array_add (array, NULL);

        return (GStrv) g_ptr_array_free (array, FALSE);
    }
    else if (key->custom_path != NULL) {
        GSettings *settings;

        settings = g_hash_table_lookup (priv->custom_settings,
                                        key->custom_path);
        binding = g_settings_get_string (settings, "binding");
    } else
        g_assert_not_reached ();

        array = g_ptr_array_new ();
        g_ptr_array_add (array, binding);
        g_ptr_array_add (array, NULL);

        return (GStrv) g_ptr_array_free (array, FALSE);
}

static void
ungrab_accelerators_complete (GObject      *object,
                              GAsyncResult *result,
                              gpointer      user_data)
{
        g_autoptr(GrabUngrabData) data = user_data;
        gboolean success = FALSE;
        g_autoptr(GError) error = NULL;
        gint i;

        g_debug ("Ungrab call completed!");
        GVariant *_ret;
        _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (object), result, &error);
        if (_ret) {
                g_variant_get (_ret,
                               "(b)",
                               &success);
                g_variant_unref (_ret);
        }
        else {
                g_warning ("Failed to ungrab accelerators: %s", error->message);

                if (g_error_matches (error, G_DBUS_ERROR, G_DBUS_ERROR_UNKNOWN_METHOD)) {
                        keys_sync_queue (data->manager, FALSE, TRUE);
                        return;
                }

                /* We are screwed at this point; we'll still keep going assuming that we don't
                 * have the bindings registered anymore.
                 * The only alternative would be to die and force cleanup of all registered
                 * grabs that way.
                 */
        }
        if (!success) {
                g_warning ("Failed to ungrab some accelerators, they were probably not registered!");
        }
        /* Clear the accelerator IDs. */
        for (i = 0; i < data->keys->len; i++) {
                MediaKey *key;

                key = g_ptr_array_index (data->keys, i);

                /* Always clear, as it would just fail again the next time. */
                g_array_set_size (key->accel_ids, 0);
        }

        /* Nothing left to do if the operation was cancelled */
        if (data->cancelled)
                return;

        keys_sync_continue (data->manager);
}

static void
grab_accelerators_complete (GObject      *object,
                            GAsyncResult *result,
                            gpointer      user_data)
{
        g_autoptr(GrabUngrabData) data = user_data;
        g_autoptr(GVariant) actions = NULL;
        g_autoptr(GError) error = NULL;
        gint i;

        g_debug ("Grab call completed!");

        GVariant *_ret;
        _ret = g_dbus_proxy_call_finish (G_DBUS_PROXY (object), result, &error);
        if (_ret) {
                g_variant_get (_ret,
                               "(@au)",
                               &actions);
                g_debug ("%s %p", g_variant_get_type_string (_ret), _ret);
                g_debug ("%s %p", g_variant_get_type_string (actions), actions);
                g_variant_unref (_ret);
        }
        if (!_ret) {
                g_warning ("Failed to grab accelerators: %s", error->message);

                if (g_error_matches (error, G_DBUS_ERROR, G_DBUS_ERROR_UNKNOWN_METHOD)) {
                        keys_sync_queue (data->manager, FALSE, TRUE);
                        return;
                }

                /* We are screwed at this point as we can't grab the keys. Most likely
                 * this means we are not running on GNOME, or ran into some other weird
                 * error.
                 * Either way, finish the operation as there is no way we can recover
                 * from this.
                 */
                keys_sync_continue (data->manager);
                return;
        }

        /* Do an immediate ungrab if the operation was cancelled.
         * This may happen on daemon shutdown for example. */
        if (data->cancelled) {
                g_debug ("Doing an immediate ungrab on the grabbed accelerators!");

                g_dbus_proxy_call (G_DBUS_PROXY (object),
                                   "UngrabAccelerators",
                                   g_variant_ref (actions), 
                                   G_DBUS_CALL_FLAGS_NONE,
                                   -1,
                                   NULL,
                                   ungrab_accelerators_complete,
                                   g_steal_pointer (&data));
                return;
        }

        /* We need to stow away the accel_ids that have been registered successfully. */
        for (i = 0; i < data->keys->len; i++) {
                MediaKey *key;

                key = g_ptr_array_index (data->keys, i);
                g_assert (key->accel_ids->len == 0);
        }
        for (i = 0; i < data->keys->len; i++) {
                MediaKey *key;
                guint accel_id;

                key = g_ptr_array_index (data->keys, i);

                g_variant_get_child (actions, i, "u", &accel_id);
                if (accel_id == 0) {
                        g_autofree gchar *tmp = NULL;
                        tmp = get_key_string (key);
                        g_debug ("Failed to grab accelerator for keybinding %s", tmp);
                } else {
                        g_array_append_val (key->accel_ids, accel_id);
                }
        }

        keys_sync_continue (data->manager);
}

static void
keys_sync_continue (CsdMediaKeysManager *manager)
{
        CsdMediaKeysManagerPrivate *priv = CSD_MEDIA_KEYS_MANAGER_GET_PRIVATE (manager);
        g_auto(GVariantBuilder) ungrab_builder = G_VARIANT_BUILDER_INIT (G_VARIANT_TYPE ("au"));
        g_auto(GVariantBuilder) grab_builder = G_VARIANT_BUILDER_INIT (G_VARIANT_TYPE ("a(suu)"));
        GVariant        *accelerators = NULL;
        g_autoptr(GPtrArray) keys_being_ungrabbed = NULL;
        g_autoptr(GPtrArray) keys_being_grabbed = NULL;
        g_autoptr(GrabUngrabData) data = NULL;
        GHashTableIter iter;
        MediaKey *key;
        gboolean need_ungrab = FALSE;

        /* Syncing keys is a two step process in principle, i.e. we first ungrab all keys
         * and then grab the new ones.
         * To make this work, this function will be called multiple times and it will
         * either emit an ungrab or grab call or do nothing when done.
         */

        /* If the keys_to_sync hash table is empty at this point, then we are done.
         * priv->keys_sync_data will be cleared automatically when it is unref'ed.
         */
        if (g_hash_table_size (priv->keys_to_sync) == 0)
                return;

        keys_being_ungrabbed = g_ptr_array_new_with_free_func ((GDestroyNotify) media_key_unref);
        keys_being_grabbed = g_ptr_array_new_with_free_func ((GDestroyNotify) media_key_unref);

        g_hash_table_iter_init (&iter, priv->keys_to_sync);
        while (g_hash_table_iter_next (&iter, (gpointer*) &key, NULL)) {
                g_auto(GStrv) bindings = NULL;
                gchar **pos = NULL;
                gint i;

                for (i = 0; i < key->accel_ids->len; i++) {
                        g_variant_builder_add (&ungrab_builder, "u", g_array_index (key->accel_ids, guint, i));
                        g_ptr_array_add (keys_being_ungrabbed, media_key_ref (key));

                        need_ungrab = TRUE;
                }

                /* Keys that are synced but aren't in the internal list are being removed.*/
                if (!g_ptr_array_find (priv->keys, key, NULL))
                        continue;

                bindings = get_bindings (manager, key);
                pos = bindings;
                while (*pos) {
                    /* Do not try to register empty keybindings. */
                    if (strlen (*pos) > 0) {
                        g_variant_builder_add (&grab_builder, "(suu)", *pos, key->modes, key->grab_flags);
                        g_ptr_array_add (keys_being_grabbed, media_key_ref (key));
                    }
                    pos++;
                }
        }
        data = g_new0 (GrabUngrabData, 1);
        data->manager = manager;

        /* These calls intentionally do not get a cancellable. See comment in
         * GrabUngrabData.
         */
        priv->keys_sync_data = data;

        if (need_ungrab) {
                data->keys = g_steal_pointer (&keys_being_ungrabbed);
                accelerators = g_variant_new ("(@au)", g_variant_builder_end (&ungrab_builder));
                g_dbus_proxy_call (priv->key_grabber,
                                   "UngrabAccelerators",
                                   accelerators, //g_variant_builder_end (&ungrab_builder),
                                   G_DBUS_CALL_FLAGS_NONE,
                                   -1,
                                   NULL,
                                   ungrab_accelerators_complete,
                                   g_steal_pointer (&data));

        } else {
                data->keys = g_steal_pointer (&keys_being_grabbed);

                g_hash_table_remove_all (priv->keys_to_sync);

                accelerators = g_variant_new ("(@a(suu))", g_variant_builder_end (&grab_builder));
                g_dbus_proxy_call (priv->key_grabber,
                                   "GrabAccelerators",
                                   accelerators, //g_variant_builder_end (&grab_builder),
                                   G_DBUS_CALL_FLAGS_NONE,
                                   -1,
                                   NULL,
                                   grab_accelerators_complete,
                                   g_steal_pointer (&data));
        }
}

// 开始Keys同步
static gboolean
keys_sync_start (gpointer user_data)
{
        CsdMediaKeysManager *manager = CSD_MEDIA_KEYS_MANAGER (user_data);
        CsdMediaKeysManagerPrivate *priv = CSD_MEDIA_KEYS_MANAGER_GET_PRIVATE (manager);

        priv->keys_sync_source_id = 0;
        g_assert (priv->keys_sync_data == NULL);
        keys_sync_continue (manager);

        return G_SOURCE_REMOVE;
}

// 队列执行快捷键keys同步
static void
keys_sync_queue (CsdMediaKeysManager *manager, gboolean immediate, gboolean retry)
{
        CsdMediaKeysManagerPrivate *priv = CSD_MEDIA_KEYS_MANAGER_GET_PRIVATE (manager);
        guint i;

        if (priv->keys_sync_source_id)
                g_source_remove (priv->keys_sync_source_id);

        if (retry) {
                /* Abort the currently running operation, and don't retry
                 * immediately to avoid race condition if an operation was
                 * already active. */
                if (priv->keys_sync_data) {
                        priv->keys_sync_data->cancelled = TRUE;
                        priv->keys_sync_data = NULL;

                        immediate = FALSE;
                }

                /* Mark all existing keys for sync. */
                for (i = 0; i < priv->keys->len; i++) {
                        MediaKey *key = g_ptr_array_index (priv->keys, i);
                        g_hash_table_add (priv->keys_to_sync, media_key_ref (key));
                }
        } else if (priv->keys_sync_data) {
                /* We are already actively syncing, no need to do anything. */
                return;
        }

        priv->keys_sync_source_id =
                g_timeout_add (immediate ? 0 : (retry ? 1000 : 50),
                               keys_sync_start,
                               manager);
}

// 响应快捷键配置变化
static void
gsettings_changed_cb (GSettings           *settings,
                      const gchar         *settings_key,
                      CsdMediaKeysManager *manager)
{
        int      i;
        CsdMediaKeysManagerPrivate *priv = CSD_MEDIA_KEYS_MANAGER_GET_PRIVATE (manager);

        if (g_strcmp0 (settings_key, "active") == 0)
            return;
        if (g_strcmp0 (settings_key, "priority") == 0)
            return;
        /* handled in gsettings_custom_changed_cb() */
        if (g_str_equal (settings_key, "custom-keybindings"))
            return;

        /* Find the key that was modified */
        if (priv->keys == NULL)
            return;

        for (i = 0; i < priv->keys->len; i++) {
                MediaKey *key;

                key = g_ptr_array_index (priv->keys, i);

                /* Skip over hard-coded and GConf keys */
                if (key->settings_key == NULL)
                        continue;
                if (strcmp (settings_key, key->settings_key) == 0) {
                        g_hash_table_add (priv->keys_to_sync, media_key_ref (key));
                        keys_sync_queue (manager, FALSE, FALSE);
                        break;
                }
        }
}

static MediaKey *
media_key_new_for_path (CsdMediaKeysManager *manager,
                        char                *path)
{
        CsdMediaKeysManagerPrivate *priv = CSD_MEDIA_KEYS_MANAGER_GET_PRIVATE (manager);
        GSettings *settings;
        char *command, *binding;
        MediaKey *key;

        g_debug ("media_key_new_for_path: %s", path);

        settings = g_hash_table_lookup (priv->custom_settings, path);
        if (settings == NULL) {
            settings = g_settings_new_with_path (CUSTOM_BINDING_SCHEMA, path);

            g_signal_connect (settings, "changed",
                    G_CALLBACK (custom_binding_changed), manager);
            g_hash_table_insert (priv->custom_settings,
                    g_strdup (path), settings);
        }

        command = g_settings_get_string (settings, "command");
        binding = g_settings_get_string (settings, "binding");

        if (*command == '\0' && *binding == '\0') {
                g_debug ("Key binding (%s) is incomplete", path);
                g_free (command);
                g_free (binding);
                return NULL;
        }
        g_free (binding);

        key = media_key_new ();
        key->key_type = CUSTOM_KEY;
        key->modes = CSD_ACTION_MODE_LAUNCHER;
        key->custom_path = g_strdup (path);
        key->custom_command = command;
        key->grab_flags = META_KEY_BINDING_NONE;

        return key;
}

static void
update_custom_binding (CsdMediaKeysManager *manager,
                       char                *path)
{
        CsdMediaKeysManagerPrivate *priv = CSD_MEDIA_KEYS_MANAGER_GET_PRIVATE (manager);
        MediaKey *key;
        int i;

        /* Remove the existing key */
        for (i = 0; i < priv->keys->len; i++) {
                key = g_ptr_array_index (priv->keys, i);

                if (key->custom_path == NULL)
                        continue;
                if (strcmp (key->custom_path, path) == 0) {
                        g_debug ("Removing custom key binding %s", path);
                        g_hash_table_add (priv->keys_to_sync, media_key_ref (key));
                        g_ptr_array_remove_index_fast (priv->keys, i);
                        break;
                }
        }

        /* And create a new one! */
        key = media_key_new_for_path (manager, path);
        if (key) {
                g_debug ("Adding new custom key binding %s", path);
                g_ptr_array_add (priv->keys, key);

                g_hash_table_add (priv->keys_to_sync, media_key_ref (key));
        }

        keys_sync_queue (manager, FALSE, FALSE);
}

static void
update_custom_binding_command (CsdMediaKeysManager *manager,
                               GSettings           *settings,
                               char                *path)
{
        CsdMediaKeysManagerPrivate *priv = CSD_MEDIA_KEYS_MANAGER_GET_PRIVATE (manager);
        MediaKey *key;
        int i;

        for (i = 0; i < priv->keys->len; i++) {
                key = g_ptr_array_index (priv->keys, i);

                if (key->custom_path == NULL)
                        continue;
                if (strcmp (key->custom_path, path) == 0) {
                        g_free (key->custom_command);
                        key->custom_command = g_settings_get_string (settings, "command");
                        break;
                }
        }
}

static void
custom_binding_changed (GSettings           *settings,
                        const char          *settings_key,
                        CsdMediaKeysManager *manager)
{
        char *path;

        g_object_get (settings, "path", &path, NULL);

        if (strcmp (settings_key, "binding") == 0)
                update_custom_binding (manager, path);
        else if (strcmp (settings_key, "command") == 0)
                update_custom_binding_command (manager, settings, path);

        g_free (path);
}

// 响应定制快捷键custom-keybindings变化
static void
gsettings_custom_changed_cb (GSettings           *settings,
                             const char          *settings_key,
                             CsdMediaKeysManager *manager)
{
        CsdMediaKeysManagerPrivate *priv = CSD_MEDIA_KEYS_MANAGER_GET_PRIVATE (manager);
        char **bindings;
        int i, j, n_bindings;

        bindings = g_settings_get_strv (settings, settings_key);
        n_bindings = g_strv_length (bindings);

        /* Handle additions */
        for (i = 0; i < n_bindings; i++) {
                if (g_hash_table_lookup (priv->custom_settings,
                                         bindings[i]))
                        continue;
                update_custom_binding (manager, bindings[i]);
        }

        /* Handle removals */
        for (i = 0; i < priv->keys->len; i++) {
                gboolean found = FALSE;
                MediaKey *key = g_ptr_array_index (priv->keys, i);
                if (key->key_type != CUSTOM_KEY)
                        continue;

                for (j = 0; j < n_bindings && !found; j++)
                        found = strcmp (bindings[j], key->custom_path) == 0;

                if (found)
                        continue;

                g_hash_table_add (priv->keys_to_sync, media_key_ref (key));
                g_hash_table_remove (priv->custom_settings,
                                     key->custom_path);
                g_ptr_array_remove_index_fast (priv->keys, i);
                --i; /* make up for the removed key */
        }
        keys_sync_queue (manager, FALSE, FALSE);
        g_strfreev (bindings);
}

static void
add_key (CsdMediaKeysManager *manager, guint i)
{
        MediaKey *key;

        key = media_key_new ();
        key->key_type = media_keys[i].key_type;
        key->settings_key = media_keys[i].settings_key;
        key->static_setting = media_keys[i].hard_coded != NULL;
        key->hard_coded = media_keys[i].hard_coded;
        key->modes = media_keys[i].modes;
        key->grab_flags = media_keys[i].grab_flags;

        g_ptr_array_add (manager->priv->keys, key);
}

// 加载快捷键配置
static void
init_kbd (CsdMediaKeysManager *manager)
{
        CsdMediaKeysManagerPrivate *priv = CSD_MEDIA_KEYS_MANAGER_GET_PRIVATE (manager);
        char **custom_paths;
        int i;

        cdos_settings_profile_start (NULL);

        /* Media keys
         * Add hard-coded shortcuts first so that they can't be preempted */
        for (i = 0; i < G_N_ELEMENTS (media_keys); i++) {
                add_key (manager, i);
        }

        /* Custom shortcuts */
        custom_paths = g_settings_get_strv (priv->settings,
                                            "custom-keybindings");

        for (i = 0; i < g_strv_length (custom_paths); i++) {
                MediaKey *key;

                g_debug ("Setting up custom keybinding %s", custom_paths[i]);
                if (custom_paths[i][0] != '/')
                {
                    g_warning ("Invalid custom keybinding %s", custom_paths[i]);
                    continue;
                }

                key = media_key_new_for_path (manager, custom_paths[i]);
                if (!key) {
                        continue;
                }
                g_ptr_array_add (priv->keys, key);
        }
        g_strfreev (custom_paths);

        keys_sync_queue (manager, TRUE, TRUE);

        cdos_settings_profile_end (NULL);
}

static gboolean
destory_old_dialog (CsdMediaKeysManager *manager)
{
        if (manager->priv->dialog_old != NULL)
        {
                gtk_widget_destroy (manager->priv->dialog_old);
                manager->priv->dialog_old = NULL;
        }
        return FALSE;
}

static void
dialog_show (CsdMediaKeysManager *manager)
{
        int            orig_w;
        int            orig_h;
        int            screen_w;
        int            screen_h;
        int            x;
        int            y;
        GdkRectangle   geometry;
        GdkMonitor    *monitor;

        /*
         * get the window size
         * if the window hasn't been mapped, it doesn't necessarily
         * know its true size, yet, so we need to jump through hoops
         */
        gtk_window_get_default_size (GTK_WINDOW (manager->priv->dialog), &orig_w, &orig_h);

        monitor = gdk_display_get_primary_monitor (gdk_display_get_default ());
        gdk_monitor_get_geometry (monitor, &geometry);

        screen_w = geometry.width;
        screen_h = geometry.height;

        x = ((screen_w - orig_w) / 2) + geometry.x;
        y = geometry.y + (screen_h / 2) + (screen_h / 2 - orig_h) / 2;

        gtk_window_move (GTK_WINDOW (manager->priv->dialog), x, y);

        gtk_widget_show (manager->priv->dialog);
        destory_old_dialog (manager);
}

// 启动应用
static void
launch_app (GAppInfo *app_info,
            gint64    timestamp)
{
        GError *error = NULL;
        GdkAppLaunchContext *launch_context;

        /* setup the launch context so the startup notification is correct */
        launch_context = gdk_display_get_app_launch_context (gdk_display_get_default ());
        gdk_app_launch_context_set_timestamp (launch_context, timestamp);

        if (!g_app_info_launch (app_info, NULL, G_APP_LAUNCH_CONTEXT (launch_context), &error)) {
                g_warning ("Could not launch '%s': %s",
                           g_app_info_get_commandline (app_info),
                           error->message);
                g_error_free (error);
        }
        g_object_unref (launch_context);
}

static gboolean
do_url_action (CsdMediaKeysManager *manager,
               const char          *scheme,
               gint64               timestamp)
{
        GAppInfo *app_info;

        app_info = g_app_info_get_default_for_uri_scheme (scheme);
        if (app_info != NULL) {
                launch_app (app_info, timestamp);
                g_object_unref (app_info);
                return TRUE;
        } else {
                g_warning ("Could not find default application for '%s' scheme", scheme);
                return FALSE;
        }
}

// 响应多媒体操作
static void
do_media_action (CsdMediaKeysManager *manager,
                 gint64               timestamp)
{
        GAppInfo *app_info;

        app_info = g_app_info_get_default_for_type ("audio/x-vorbis+ogg", FALSE);
        if (app_info != NULL) {
                launch_app (app_info, timestamp);
                g_object_unref (app_info);
        } else {
                g_warning ("Could not find default application for '%s' mime-type", "audio/x-vorbis+ogg");
        }
}

// 响应终端操作
static void
do_terminal_action (CsdMediaKeysManager *manager)
{
        GSettings *settings;
        char *term;

        settings = g_settings_new ("org.cdos.desktop.default-applications.terminal");
        term = g_settings_get_string (settings, "exec");

        if (term)
            execute (manager, term, 0);

        g_free (term);
        g_object_unref (settings);
}

static void
choose_from_dialog (GObject *source_object,
                    GAsyncResult *res,
                    gpointer user_data)
{
        GError *error;
        GVariant *variant;

        variant = g_dbus_connection_call_finish ((GDBusConnection *)source_object, res, &error);
        if (variant == NULL) {
                g_warning ("Failed to call Shutdown on session manager: %s", error->message);
                g_error_free (error);
                return;
        }
        g_variant_unref (variant);
        return;
}

// 关机(弹窗交互)
static void
cdos_session_shutdown (CsdMediaKeysManager *manager)
{
        /* Shouldn't happen, but you never know */
        if (manager->priv->connection == NULL) {
                execute (manager, "cdos-session-quit --logout", 0);
                return;
        }

        // g_dbus_connection_call (manager->priv->connection,
        //                         GNOME_SESSION_DBUS_NAME,
        //                         GNOME_SESSION_DBUS_PATH,
        //                         GNOME_SESSION_DBUS_INTERFACE,
        //                         "Shutdown",
        //                         NULL,
        //                         NULL,
        //                         G_DBUS_CALL_FLAGS_NONE,
        //                         -1,
        //                         NULL,
        //                         (GAsyncReadyCallback) choose_from_dialog,
        //                         NULL);
        //打开电源管理窗口
        g_dbus_connection_call (manager->priv->connection,
                                "org.freedesktop.Desktop",
                                "/org/freedesktop/Desktop",
                                "org.freedesktop.Desktop.PowerDialog",
                                "Run",
                                g_variant_new ("(is)", 0, ""),
                                NULL,
                                G_DBUS_CALL_FLAGS_NONE,
                                -1,
                                NULL,
                                NULL,
                                NULL);
}


static void
show_capslock_osd_dialog (CsdMediaKeysManager *manager, gboolean state)
{

    dialog_init (manager);
    csd_osd_window_set_action_custom (CSD_OSD_WINDOW (manager->priv->dialog),
                                      state ? "caps-lock-on-symbolic" : "caps-lock-off-symbolic",
                                      FALSE);
    dialog_show (manager);

}

static gboolean
is_capslock_on (void)
{
	GdkKeymap *keymap;
	gboolean res;

	res = FALSE;

	keymap =  gdk_keymap_get_for_display (gdk_display_get_default ());

	if (keymap != NULL) {
		res = gdk_keymap_get_caps_lock_state (keymap);
	}
	return res;
}
static gboolean
do_capslock (CsdMediaKeysManager *manager)
{
        CsdMediaKeysManagerPrivate *priv = manager->priv;
        gboolean state;
        priv->caps_lock_action_id = 0;

        state = is_capslock_on();
        show_capslock_osd_dialog (manager, state);

        return FALSE;
}

static void
do_capslock_action (CsdMediaKeysManager *manager)
{
        CsdMediaKeysManagerPrivate *priv = manager->priv;
        if (priv->caps_lock_action_id > 0)
        {
                g_source_remove (priv->caps_lock_action_id);
        }
        priv->caps_lock_action_id  = g_timeout_add (200, (GSourceFunc) do_capslock, manager);
}

static void
show_numlock_osd_dialog (CsdMediaKeysManager *manager, gboolean state)
{

        dialog_init (manager);
        csd_osd_window_set_action_custom (CSD_OSD_WINDOW (manager->priv->dialog),
                                        state ? "num-lock-on-symbolic" : "num-lock-off-symbolic",
                                        FALSE);
        dialog_show (manager);

}

static gboolean
is_numlock_on (void)
{
	GdkKeymap *keymap;
	gboolean res;

	res = FALSE;
	keymap =  gdk_keymap_get_for_display (gdk_display_get_default ());

	if (keymap != NULL) {
		res = gdk_keymap_get_num_lock_state (keymap);
	}
	return res;
}

static gboolean
do_numlock (CsdMediaKeysManager *manager)
{
        CsdMediaKeysManagerPrivate *priv = manager->priv;
        priv->num_lock_action_id = 0;

        gboolean state;
        state = is_numlock_on();
        show_numlock_osd_dialog (manager, state);
        return FALSE;

}

static void
do_numlock_action (CsdMediaKeysManager *manager)
{
        CsdMediaKeysManagerPrivate *priv = manager->priv;

        if (priv->num_lock_action_id > 0)
        {
                g_source_remove (priv->num_lock_action_id);
        }
        priv->num_lock_action_id  = g_timeout_add (200, (GSourceFunc) do_numlock, manager);
}

static void
do_logout_action (CsdMediaKeysManager *manager)
{
        execute (manager, "cdos-session-quit --logout", 0);
}

static void
do_shutdown_action (CsdMediaKeysManager *manager)
{
        cdos_session_shutdown (manager);
}

static void
do_eject_action_cb (GDrive              *drive,
                    GAsyncResult        *res,
                    CsdMediaKeysManager *manager)
{
        g_drive_eject_with_operation_finish (drive, res, NULL);
}

#define NO_SCORE 0
#define SCORE_CAN_EJECT 50
#define SCORE_HAS_MEDIA 100
static void
do_eject_action (CsdMediaKeysManager *manager)
{
        GList *drives, *l;
        GDrive *fav_drive;
        guint score;
        GVolumeMonitor *volume_monitor;

        volume_monitor = g_volume_monitor_get ();


        /* Find the best drive to eject */
        fav_drive = NULL;
        score = NO_SCORE;
        drives = g_volume_monitor_get_connected_drives (volume_monitor);
        for (l = drives; l != NULL; l = l->next) {
                GDrive *drive = l->data;

                if (g_drive_can_eject (drive) == FALSE)
                        continue;
                if (g_drive_is_media_removable (drive) == FALSE)
                        continue;
                if (score < SCORE_CAN_EJECT) {
                        fav_drive = drive;
                        score = SCORE_CAN_EJECT;
                }
                if (g_drive_has_media (drive) == FALSE)
                        continue;
                if (score < SCORE_HAS_MEDIA) {
                        fav_drive = drive;
                        score = SCORE_HAS_MEDIA;
                        break;
                }
        }

        /* Show the dialogue */
        dialog_init (manager);
        csd_osd_window_set_action_custom (CSD_OSD_WINDOW (manager->priv->dialog),
                                                 "media-eject-symbolic",
                                                 FALSE);
        dialog_show (manager);

        /* Clean up the drive selection and exit if no suitable
         * drives are found */
        if (fav_drive != NULL)
                fav_drive = g_object_ref (fav_drive);

        g_list_foreach (drives, (GFunc) g_object_unref, NULL);
        if (fav_drive == NULL)
                return;

        /* Eject! */
        g_drive_eject_with_operation (fav_drive, G_MOUNT_UNMOUNT_FORCE,
                                      NULL, NULL,
                                      (GAsyncReadyCallback) do_eject_action_cb,
                                      manager);
        g_object_unref (fav_drive);
        g_object_unref (volume_monitor);
}

static void
do_home_key_action (CsdMediaKeysManager *manager,
                    gint64               timestamp)
{
        g_autofree char * cmd = NULL;
        if (manager->priv->screensaver_lock)
                return;
        if (!gtk_show_uri_on_window (NULL, g_get_home_dir (), timestamp, NULL))
                if ((cmd = g_find_program_in_path ("nemo")))
                        execute (manager, cmd, timestamp);
}

static void
do_execute_desktop (CsdMediaKeysManager *manager,
                    const char          *desktop,
                    gint64               timestamp)
{
        GDesktopAppInfo *app_info;
        if (manager->priv->screensaver_lock)
                return ;

        app_info = g_desktop_app_info_new (desktop);
        if (app_info != NULL) {
                launch_app (G_APP_INFO (app_info), timestamp);
                g_object_unref (app_info);
        } else {
                g_warning ("Could not find application '%s'", desktop);
        }
}

static void
show_touchpad_osd_dialog (CsdMediaKeysManager *manager, gboolean state)
{
    /*if(state){
        g_spawn_command_line_async("/usr/bin/dm-tool instpd",NULL);
    } else {
        g_spawn_command_line_async("/usr/bin/dm-tool rmtpd",NULL);
    } */
    dialog_init (manager);
    csd_osd_window_set_action_custom (CSD_OSD_WINDOW (manager->priv->dialog),
                                      state ? "input-touchpad-symbolic" : "touchpad-disabled-symbolic",
                                      FALSE);
    dialog_show (manager);
}

// 响应触控板操作
static void
do_touchpad_action (CsdMediaKeysManager *manager, MediaKeyType type)
{
        GSettings *settings;
        const gchar *new = "enabled";
        gchar *enabled = NULL;
        gboolean state, new_state;

        if (touchpad_is_present () == FALSE) 
        {
                show_touchpad_osd_dialog (manager, FALSE);
                return;
        }

        settings = g_settings_new (SETTINGS_TOUCHPAD_DIR);
        if (type == TOUCHPAD_KEY)
        {
                enabled = g_settings_get_string (settings, TOUCHPAD_ENABLED_KEY);
                state = g_strcmp0 (enabled, "enabled") == 0 ? TRUE : FALSE;
                new_state = !state;
                new = new_state ? "enabled" : "disabled";
        } 
        else if (type == TOUCHPAD_ON_KEY)
        {
                new_state = TRUE;
                new = "enabled";
        } 
        else if (type == TOUCHPAD_OFF_KEY)
        {
                new_state = FALSE;
                new = "disabled";
        }
        show_touchpad_osd_dialog (manager, new_state);
        g_settings_set_string (settings, TOUCHPAD_ENABLED_KEY, new);
        g_object_unref (settings);
        g_free (enabled);
}

static void
update_dialog (CsdMediaKeysManager *manager,
               GvcMixerStream      *stream,
               gint                 vol,
               gboolean             muted,
               gboolean             sound_changed,
               gboolean             quiet,
               int                  type)
{
        vol = CLAMP (vol, 0, 100);

        dialog_init (manager);
        csd_osd_window_set_volume_muted (CSD_OSD_WINDOW (manager->priv->dialog),
                                                muted);
        csd_osd_window_set_volume_level (CSD_OSD_WINDOW (manager->priv->dialog), vol);
        if(type == MIC_MUTE_KEY){
                    const gchar *icon_name = muted == TRUE? "mic-disable-symbolic":"mic-enable-symbolic";
                csd_osd_window_set_action_custom (CSD_OSD_WINDOW (manager->priv->dialog), icon_name, FALSE);
        }
        else{
                csd_osd_window_set_action (CSD_OSD_WINDOW (manager->priv->dialog),
                                           CSD_OSD_WINDOW_ACTION_VOLUME);
        }
        dialog_show (manager);

        /* if (quiet == FALSE && sound_changed != FALSE && muted == FALSE) {
                GSettings *settings = g_settings_new ("org.cdos.desktop.sound");
                gboolean enabled = g_settings_get_boolean (settings, "volume-sound-enabled");
                char *sound = g_settings_get_string (settings, "volume-sound-file");
                g_message ("Sound changed, play sound file: %s, enabled: %d", sound, enabled);
                if (enabled) {
                    ca_context_change_device (manager->priv->ca, gvc_mixer_stream_get_name (stream));
                    ca_context_play (manager->priv->ca, 1, CA_PROP_MEDIA_FILENAME, sound, NULL);
                }
                g_free(sound);
                g_object_unref (settings);
        } */
}

#ifdef HAVE_GUDEV
/* PulseAudio gives us /devices/... paths, when udev
 * expects /sys/devices/... paths. */
static GUdevDevice *
get_udev_device_for_sysfs_path (CsdMediaKeysManager *manager,
                                const char *sysfs_path)
{
        char *path;
        GUdevDevice *dev;

        path = g_strdup_printf ("/sys%s", sysfs_path);
        dev = g_udev_client_query_by_sysfs_path (manager->priv->udev_client, path);
        g_free (path);

        return dev;
}

static GvcMixerStream *
get_stream_for_device_id (CsdMediaKeysManager *manager,
                          gboolean             is_output,
                          const char          *devnode)
{
        gpointer id_ptr;
        GvcMixerStream *res;
        GUdevDevice *dev, *parent;
        GSList *streams, *l;

        id_ptr = g_hash_table_lookup (manager->priv->streams, devnode);
        if (id_ptr != NULL) {
                if (GPOINTER_TO_UINT (id_ptr) == (guint) -1)
                        return NULL;
                else
                        return gvc_mixer_control_lookup_stream_id (manager->priv->volume, GPOINTER_TO_UINT (id_ptr));
        }

        dev = g_udev_client_query_by_device_file (manager->priv->udev_client, devnode);
        if (dev == NULL) {
                g_debug ("Could not find udev device for device path '%s'", devnode);
                return NULL;
        }

        if (g_strcmp0 (g_udev_device_get_property (dev, "ID_BUS"), "usb") != 0) {
                g_debug ("Not handling XInput device %s, not USB", devnode);
                g_hash_table_insert (manager->priv->streams,
                                     g_strdup (devnode),
                                     GUINT_TO_POINTER ((guint) -1));
                g_object_unref (dev);
                return NULL;
        }

        parent = g_udev_device_get_parent_with_subsystem (dev, "usb", "usb_device");
        if (parent == NULL) {
                g_warning ("No USB device parent for XInput device %s even though it's USB", devnode);
                g_object_unref (dev);
                return NULL;
        }

        res = NULL;
        if (is_output)
                streams = gvc_mixer_control_get_sinks (manager->priv->volume);
        else
                streams = gvc_mixer_control_get_sources (manager->priv->volume);

        for (l = streams; l; l = l->next) {
                GvcMixerStream *stream = l->data;
                const char *sysfs_path;
                GUdevDevice *stream_dev, *stream_parent;

                sysfs_path = gvc_mixer_stream_get_sysfs_path (stream);
                stream_dev = get_udev_device_for_sysfs_path (manager, sysfs_path);
                if (stream_dev == NULL)
                        continue;
                stream_parent = g_udev_device_get_parent_with_subsystem (stream_dev, "usb", "usb_device");
                g_object_unref (stream_dev);
                if (stream_parent == NULL)
                        continue;

                if (g_strcmp0 (g_udev_device_get_sysfs_path (stream_parent),
                               g_udev_device_get_sysfs_path (parent)) == 0) {
                        res = stream;
                }
                g_object_unref (stream_parent);
                if (res != NULL)
                        break;
        }

        g_slist_free (streams);

        if (res)
                g_hash_table_insert (manager->priv->streams,
                                     g_strdup (devnode),
                                     GUINT_TO_POINTER (gvc_mixer_stream_get_id (res)));
        else
                g_hash_table_insert (manager->priv->streams,
                                     g_strdup (devnode),
                                     GUINT_TO_POINTER ((guint) -1));

        return res;
}
#endif /* HAVE_GUDEV */

// 响应音频操作
static gint
do_sound_action (CsdMediaKeysManager *manager,
                 const gchar         *deviceid,
                 int                  type,
                 gboolean             is_output,
                 gboolean             quiet)
{
        GvcMixerStream *stream;
        gboolean old_muted, new_muted;
        guint32 old_vol, new_vol, norm_vol_step, osd_vol;
        gboolean sound_changed;

        /* Find the stream that corresponds to the device, if any */
#ifdef HAVE_GUDEV
        stream = get_stream_for_device_id (manager, type, deviceid);
#endif /* HAVE_GUDEV */
        if (stream == NULL) {
                if (is_output)
                        stream = manager->priv->stream;
                else
                        stream = manager->priv->source;
        }

        if (stream == NULL)
                return -1;

        norm_vol_step = (guint32)(MAX_VOLUME * VOLUME_STEP / 100);

        /* FIXME: this is racy */
        new_vol = old_vol = gvc_mixer_stream_get_volume (stream);
        new_muted = old_muted = gvc_mixer_stream_get_is_muted (stream);
        sound_changed = FALSE;

        switch (type) {
        case MUTE_KEY:
        case MIC_MUTE_KEY:
                new_muted = !old_muted;
                break;
        case VOLUME_DOWN_KEY:
                if (old_vol <= norm_vol_step) {
                        new_vol = 0;
                        new_muted = TRUE;
                } else {
                        new_vol = old_vol - norm_vol_step;
                }
                break;
        case VOLUME_UP_KEY:
                new_muted = FALSE;
                /* When coming out of mute only increase the volume if it was 0 */
                if (!old_muted || old_vol == 0)
                        new_vol = MIN (old_vol + norm_vol_step, MAX_VOLUME);
                break;
        }

        if (old_muted != new_muted) {
                gvc_mixer_stream_change_is_muted (stream, new_muted);
                sound_changed = TRUE;
        }

        if (old_vol != new_vol) {
                if (gvc_mixer_stream_set_volume (stream, new_vol) != FALSE) {
                        gvc_mixer_stream_push_volume (stream);
                        sound_changed = TRUE;
                }
        }
        if (type == VOLUME_DOWN_KEY && old_vol == 0 && old_muted)
                osd_vol = -1;
        else if (type == VOLUME_UP_KEY && old_vol == MAX_VOLUME && !old_muted)
                osd_vol = 101;
        else if (!new_muted)
                osd_vol = (int)round (100 * (double) new_vol / MAX_VOLUME);
        else
                osd_vol = 0;
        update_dialog (manager, stream, osd_vol, new_muted, sound_changed, quiet, type);

        g_debug ("do_sound_action: old_vol=%d, new_vol=%d", old_vol, new_vol);
        return new_vol;
}

// static void
// sound_theme_changed (GtkSettings         *settings,
//                      GParamSpec          *pspec,
//                      CsdMediaKeysManager *manager)
// {
//         char *theme_name;

//         g_object_get (G_OBJECT (manager->priv->gtksettings), "gtk-sound-theme-name", &theme_name, NULL);
//         if (theme_name)
//                 ca_context_change_props (manager->priv->ca, CA_PROP_CANBERRA_XDG_THEME_NAME, theme_name, NULL);
//         g_free (theme_name);
// }

static void
update_default_sink (CsdMediaKeysManager *manager)
{
        GvcMixerStream *stream;

        stream = gvc_mixer_control_get_default_sink (manager->priv->volume);
        if (stream == manager->priv->stream)
                return;

        if (manager->priv->stream != NULL) {
                g_object_unref (manager->priv->stream);
                manager->priv->stream = NULL;
        }

        if (stream != NULL) {
                manager->priv->stream = g_object_ref (stream);
        } else {
                g_warning ("Unable to get default sink");
        }
}

static void
update_default_source (CsdMediaKeysManager *manager)
{
        GvcMixerStream *stream;

        stream = gvc_mixer_control_get_default_source (manager->priv->volume);
        if (stream == manager->priv->source)
                return;

        if (manager->priv->source != NULL) {
                g_object_unref (manager->priv->source);
                manager->priv->source = NULL;
        }

        if (stream != NULL) {
                manager->priv->source = g_object_ref (stream);
        } else {
                g_warning ("Unable to get default source");
        }
}

static void
on_control_state_changed (GvcMixerControl     *control,
                          GvcMixerControlState new_state,
                          CsdMediaKeysManager *manager)
{
        update_default_sink (manager);
        update_default_source (manager);
}

static void
on_control_default_sink_changed (GvcMixerControl     *control,
                                 guint                id,
                                 CsdMediaKeysManager *manager)
{
        update_default_sink (manager);
}

static void
on_control_default_source_changed (GvcMixerControl     *control,
                                   guint                id,
                                   CsdMediaKeysManager *manager)
{
        update_default_source (manager);
}

#ifdef HAVE_GUDEV
static gboolean
remove_stream (gpointer key,
               gpointer value,
               gpointer id)
{
        if (GPOINTER_TO_UINT (value) == GPOINTER_TO_UINT (id))
                return TRUE;
        return FALSE;
}
#endif /* HAVE_GUDEV */

static void
on_control_stream_removed (GvcMixerControl     *control,
                           guint                id,
                           CsdMediaKeysManager *manager)
{
        if (manager->priv->stream != NULL) {
                if (gvc_mixer_stream_get_id (manager->priv->stream) == id) {
                        g_object_unref (manager->priv->stream);
                        manager->priv->stream = NULL;
                }
        }
        if (manager->priv->source != NULL) {
            if (gvc_mixer_stream_get_id (manager->priv->source) == id) {
                        g_object_unref (manager->priv->source);
                    manager->priv->source = NULL;
            }
        }

#ifdef HAVE_GUDEV
        g_hash_table_foreach_remove (manager->priv->streams, (GHRFunc) remove_stream, GUINT_TO_POINTER (id));
#endif
}

static void
free_media_player (MediaPlayer *player)
{
        if (player->watch_id > 0) {
                g_bus_unwatch_name (player->watch_id);
                player->watch_id = 0;
        }
        g_free (player->application);
        g_free (player->name);
        g_free (player);
}

static gint
find_by_application (gconstpointer a,
                     gconstpointer b)
{
        return strcmp (((MediaPlayer *)a)->application, b);
}

static gint
find_by_name (gconstpointer a,
              gconstpointer b)
{
        return strcmp (((MediaPlayer *)a)->name, b);
}

static gint
find_by_time (gconstpointer a,
              gconstpointer b)
{
        return ((MediaPlayer *)a)->time < ((MediaPlayer *)b)->time;
}

static void
name_vanished_handler (GDBusConnection     *connection,
                       const gchar         *name,
                       CsdMediaKeysManager *manager)
{
        GList *iter;

        iter = g_list_find_custom (manager->priv->media_players,
                                   name,
                                   find_by_name);

        if (iter != NULL) {
                MediaPlayer *player;

                player = iter->data;
                g_debug ("Deregistering vanished %s (name: %s)", player->application, player->name);
                free_media_player (player);
                manager->priv->media_players = g_list_delete_link (manager->priv->media_players, iter);
        }
}

/*
 * Register a new media player. Most applications will want to call
 * this with time = GDK_CURRENT_TIME. This way, the last registered
 * player will receive media events. In some cases, applications
 * may want to register with a lower priority (usually 1), to grab
 * events only nobody is interested.
 */
static void
csd_media_keys_manager_grab_media_player_keys (CsdMediaKeysManager *manager,
                                               const char          *application,
                                               const char          *name,
                                               guint32              time)
{
        GList       *iter;
        MediaPlayer *media_player;
        guint        watch_id;

        if (time == GDK_CURRENT_TIME) {
                GTimeVal tv;

                g_get_current_time (&tv);
                time = tv.tv_sec * 1000 + tv.tv_usec / 1000;
        }

        iter = g_list_find_custom (manager->priv->media_players,
                                   application,
                                   find_by_application);

        if (iter != NULL) {
                if (((MediaPlayer *)iter->data)->time < time) {
                        MediaPlayer *player = iter->data;
                        free_media_player (player);
                        manager->priv->media_players = g_list_delete_link (manager->priv->media_players, iter);
                } else {
                        return;
                }
        }

        watch_id = g_bus_watch_name (G_BUS_TYPE_SESSION,
                                     name,
                                     G_BUS_NAME_WATCHER_FLAGS_NONE,
                                     NULL,
                                     (GBusNameVanishedCallback) name_vanished_handler,
                                     manager,
                                     NULL);

        g_debug ("Registering %s at %u", application, time);
        media_player = g_new0 (MediaPlayer, 1);
        media_player->application = g_strdup (application);
        media_player->name = g_strdup (name);
        media_player->time = time;
        media_player->watch_id = watch_id;

        manager->priv->media_players = g_list_insert_sorted (manager->priv->media_players,
                                                             media_player,
                                                             find_by_time);
}

static void
csd_media_keys_manager_release_media_player_keys (CsdMediaKeysManager *manager,
                                                  const char          *application,
                                                  const char          *name)
{
        GList *iter = NULL;

        g_return_if_fail (application != NULL || name != NULL);

        if (application != NULL) {
                iter = g_list_find_custom (manager->priv->media_players,
                                           application,
                                           find_by_application);
        }

        if (iter == NULL && name != NULL) {
                iter = g_list_find_custom (manager->priv->media_players,
                                           name,
                                           find_by_name);
        }

        if (iter != NULL) {
                MediaPlayer *player;

                player = iter->data;
                g_debug ("Deregistering %s (name: %s)", application, player->name);
                free_media_player (player);
                manager->priv->media_players = g_list_delete_link (manager->priv->media_players, iter);
        }
}

static gboolean
csd_media_player_key_pressed (CsdMediaKeysManager *manager,
                              const char          *key)
{
        const char  *application;
        gboolean     have_listeners;
        GError      *error = NULL;
        MediaPlayer *player;

        g_return_val_if_fail (key != NULL, FALSE);

        g_debug ("Media key '%s' pressed", key);

        have_listeners = (manager->priv->media_players != NULL);

        if (!have_listeners && manager->priv->mpris_controller) {
                if (!mpris_controller_key (manager->priv->mpris_controller, key)) {
                        /* Popup a dialog with an (/) icon */
                        dialog_init (manager);
                        csd_osd_window_set_action_custom (CSD_OSD_WINDOW (manager->priv->dialog),
                                                         "action-unavailable-symbolic",
                                                         FALSE);
                        dialog_show (manager);
                }
                return TRUE;
        }

        player = manager->priv->media_players->data;
        application = player->application;

        if (g_dbus_connection_emit_signal (manager->priv->connection,
                                           player->name,
                                           CSD_MEDIA_KEYS_DBUS_PATH,
                                           CSD_MEDIA_KEYS_DBUS_NAME,
                                           "MediaPlayerKeyPressed",
                                           g_variant_new ("(ss)", application ? application : "", key),
                                           &error) == FALSE) {
                g_debug ("Error emitting signal: %s", error->message);
                g_error_free (error);
        }

        return !have_listeners;
}

// 响应多媒体播放器操作
static gboolean
do_multimedia_player_action (CsdMediaKeysManager *manager,
                             const char          *icon,
                             const char          *key)
{
        return csd_media_player_key_pressed (manager, key);
}

static void
do_toggle_accessibility_key (const char *key)
{
        GSettings *settings;
        gboolean state;

        settings = g_settings_new ("org.cdos.desktop.a11y.applications");
        state = g_settings_get_boolean (settings, key);
        g_settings_set_boolean (settings, key, !state);
        g_object_unref (settings);
}

static void
do_video_rotate_action (CsdMediaKeysManager *manager, gint64 timestamp)
{
        g_warning ("%s TODO", __func__);
}

static void
do_screenreader_action (CsdMediaKeysManager *manager)
{
        do_toggle_accessibility_key ("screen-reader-enabled");
}

static void
do_screenmagnifier_action (CsdMediaKeysManager *manager)
{
        do_toggle_accessibility_key ("screen-magnifier-enabled");
}

static gboolean
check_is_magnifier_enabled ()
{
         GSettings *settings = g_settings_new ("org.cdos.desktop.a11y.applications");

         return g_settings_get_boolean (settings, "screen-magnifier-enabled");
}

static void
do_screenmagnifier_zoom_out_action (CsdMediaKeysManager *manager)
{
         //   do_toggle_accessibility_key ("screen-magnifier-enabled");
        if (!check_is_magnifier_enabled ())
        {
            return;
        }
        GSettings *settings = g_settings_new ("org.cdos.desktop.a11y.magnifier");

        gdouble factor = g_settings_get_double (settings, "mag-factor");

        if (factor <= 2.0)
        {
            factor += 0.25;
        }

        g_settings_set_double (settings, "mag-factor", factor);
}

static void
do_screenmagnifier_zoom_in_action (CsdMediaKeysManager *manager)
{
     //   do_toggle_accessibility_key ("screen-magnifier-enabled");
     g_debug ("zoom in or out key print");
     if (!check_is_magnifier_enabled ())
     {
        return;
     }
     GSettings *settings = g_settings_new ("org.cdos.desktop.a11y.magnifier");

     gdouble factor = g_settings_get_double (settings, "mag-factor");

     if (factor >= 1.0)
     {
        factor -= 0.25;
     }

     g_settings_set_double (settings, "mag-factor", factor);
}

static void
do_on_screen_keyboard_action (CsdMediaKeysManager *manager)
{
        do_toggle_accessibility_key ("screen-keyboard-enabled");
}

static void
do_text_size_action (CsdMediaKeysManager *manager,
                     MediaKeyType         type)
{
        gdouble factor, best, distance;
        guint i;

        /* Same values used in the Seeing tab of the Universal Access panel */
        static gdouble factors[] = {
                0.75,
                1.0,
                1.25,
                1.5
        };

        /* Figure out the current DPI scaling factor */
        factor = g_settings_get_double (manager->priv->interface_settings, "text-scaling-factor");
        factor += (type == INCREASE_TEXT_KEY ? 0.25 : -0.25);

        /* Try to find a matching value */
        distance = 1e6;
        best = 1.0;
        for (i = 0; i < G_N_ELEMENTS(factors); i++) {
                gdouble d;
                d = fabs (factor - factors[i]);
                if (d < distance) {
                        best = factors[i];
                        distance = d;
                }
        }

        if (best == 1.0)
                g_settings_reset (manager->priv->interface_settings, "text-scaling-factor");
        else
                g_settings_set_double (manager->priv->interface_settings, "text-scaling-factor", best);
}

static void
do_toggle_contrast_action (CsdMediaKeysManager *manager)
{
        gboolean high_contrast;
        char *theme;

        /* Are we using HighContrast now? */
        theme = g_settings_get_string (manager->priv->interface_settings, "gtk-theme");
        high_contrast = g_str_equal (theme, HIGH_CONTRAST);
        g_free (theme);

        if (high_contrast != FALSE) {
                if (manager->priv->gtk_theme == NULL)
                        g_settings_reset (manager->priv->interface_settings, "gtk-theme");
                else
                        g_settings_set_string (manager->priv->interface_settings, "gtk-theme", manager->priv->gtk_theme);
                g_settings_set_string (manager->priv->interface_settings, "icon-theme", manager->priv->icon_theme);
        } else {
                g_settings_set_string (manager->priv->interface_settings, "gtk-theme", HIGH_CONTRAST);
                g_settings_set_string (manager->priv->interface_settings, "icon-theme", HIGH_CONTRAST);
        }
}

// 锁屏
static void
do_lock_screen (CsdMediaKeysManager *manager)
{
    g_dbus_connection_call_sync (manager->priv->connection,
                                 SCREENSAVER_DBUS_NAME,
                                 SCREENSAVER_DBUS_PATH,
                                 SCREENSAVER_DBUS_INTERFACE,
                                 "Lock",
                                 NULL,
                                 NULL,
                                 G_DBUS_CALL_FLAGS_NONE,
                                 -1,
                                 NULL,
                                 NULL);
}

// 电源按键响应
static void
do_config_power_action (CsdMediaKeysManager *manager,
                        const gchar *config_key)
{
        CsdPowerActionType action_type;

        action_type = g_settings_get_enum (manager->priv->power_settings,
                                           config_key);
        g_message ("Do power action, config key = %s, type = %d(%s).", config_key, action_type, power_action_type_to_string (action_type));
        switch (action_type) {
        case CSD_POWER_ACTION_SUSPEND:
                csd_power_suspend (manager->priv->upower_proxy);
                break;
        case CSD_POWER_ACTION_INTERACTIVE:
                cdos_session_shutdown (manager);
                break;
        case CSD_POWER_ACTION_SHUTDOWN:
                csd_power_poweroff ();
                break;
        case CSD_POWER_ACTION_HIBERNATE:
                csd_power_hibernate (manager->priv->upower_proxy);
                break;
        case CSD_POWER_ACTION_BLANK:
                do_lock_screen (manager);
                break;
        case CSD_POWER_ACTION_NOTHING:
                /* these actions cannot be handled by media-keys and
                 * are not used in this context */
                break;
        }
}

// 亮度弹窗
static void
show_brightness_dialog (CsdMediaKeysManager *manager, guint percentage)
{
        dialog_init (manager);
        csd_osd_window_set_action_custom (CSD_OSD_WINDOW (manager->priv->dialog),
                                                 "display-brightness-symbolic",
                                                 TRUE);
        csd_osd_window_set_volume_level (CSD_OSD_WINDOW (manager->priv->dialog),
                                                percentage);
        dialog_show (manager);
}

static void
update_screen_cb (GObject             *source_object,
                  GAsyncResult        *res,
                  gpointer             user_data)
{
        GError *error = NULL;
        guint percentage;
        GVariant *new_percentage;
        CsdMediaKeysManager *manager = (CsdMediaKeysManager *)user_data;

        new_percentage = g_dbus_proxy_call_finish (G_DBUS_PROXY (source_object),
                                                   res, &error);
        if (new_percentage == NULL) {
                g_warning ("Failed to set new screen percentage: %s",
                           error->message);
                g_error_free (error);
                return;
        }

        /* update the dialog with the new value */
        g_variant_get (new_percentage, "(u)", &percentage);
        show_brightness_dialog (manager, percentage);
        g_variant_unref (new_percentage);
}

// 响应屏幕亮度变化
static void
do_screen_brightness_action (CsdMediaKeysManager *manager,
                             MediaKeyType type)
{
        if (manager->priv->connection == NULL ||
            manager->priv->power_screen_proxy == NULL) {
                g_warning ("No existing D-Bus connection trying to handle power keys");
                return;
        }
        if (manager->priv->power_mode_enable)
                return ;

        g_dbus_proxy_call (manager->priv->power_screen_proxy,
                           type == SCREEN_BRIGHTNESS_UP_KEY ? "StepUp" : "StepDown",
                           NULL,
                           G_DBUS_CALL_FLAGS_NONE,
                           -1,
                           NULL,
                           update_screen_cb,
                           manager);
}

static void
update_keyboard_cb (GObject             *source_object,
                    GAsyncResult        *res,
                    gpointer             user_data)
{
        GError *error = NULL;
        guint percentage;
        GVariant *new_percentage;
        CsdMediaKeysManager *manager = CSD_MEDIA_KEYS_MANAGER (user_data);

        new_percentage = g_dbus_proxy_call_finish (G_DBUS_PROXY (source_object),
                                                   res, &error);
        if (new_percentage == NULL) {
                g_warning ("Failed to set new keyboard percentage: %s",
                           error->message);
                g_error_free (error);
                return;
        }

        /* update the dialog with the new value */
        g_variant_get (new_percentage, "(u)", &percentage);

        /* FIXME: No overshoot effect for keyboard, as the power plugin has no interface
         *        to get the old brightness */
        dialog_init (manager);
        csd_osd_window_set_action_custom (CSD_OSD_WINDOW (manager->priv->dialog),
                                                 "keyboard-brightness-symbolic",
                                                 TRUE);
        csd_osd_window_set_volume_level (CSD_OSD_WINDOW (manager->priv->dialog),
                                                percentage);
        dialog_show (manager);

        g_variant_unref (new_percentage);
}

// 响应键盘背景灯光亮度改变
static void
do_keyboard_brightness_action (CsdMediaKeysManager *manager,
                               MediaKeyType type)
{
        const char *cmd;

        if (manager->priv->connection == NULL ||
            manager->priv->power_keyboard_proxy == NULL) {
                g_warning ("No existing D-Bus connection trying to handle power keys");
                return;
        }

        switch (type) {
        case KEYBOARD_BRIGHTNESS_UP_KEY:
                cmd = "StepUp";
                break;
        case KEYBOARD_BRIGHTNESS_DOWN_KEY:
                cmd = "StepDown";
                break;
        case KEYBOARD_BRIGHTNESS_TOGGLE_KEY:
                cmd = "Toggle";
                break;
        default:
                g_assert_not_reached ();
        }

        /* call into the power plugin */
        g_dbus_proxy_call (manager->priv->power_keyboard_proxy,
                           cmd,
                           NULL,
                           G_DBUS_CALL_FLAGS_NONE,
                           -1,
                           NULL,
                           update_keyboard_cb,
                           manager);
}

static void
do_camera_osd_action(CsdMediaKeysManager *manager, gboolean state)
{
        dialog_init (manager);
        csd_osd_window_set_action_custom (CSD_OSD_WINDOW (manager->priv->dialog),
                                          state ? "camera-photo-symbolic-disable" : "camera-photo-symbolic",
                                          FALSE);
        dialog_show (manager);
}

// 响应定制的快捷键
static void
do_custom_action (CsdMediaKeysManager *manager,
                  const gchar         *device_node,
                  MediaKey            *key,
                  gint64               timestamp)
{
        g_debug ("Launching custom action for key (on device node %s)", device_node);
        execute (manager, key->custom_command, timestamp);
}

static gboolean
do_camera_exist (void)
{
        GSettings *settings;
        gboolean state;

        settings = g_settings_new (SETTINGS_CAMERA_DIR);
        state = g_settings_get_boolean (settings, CAMERA_DISABLED_KEY);

        g_settings_set_boolean (settings, CAMERA_DISABLED_KEY, !state);
        g_object_unref (settings);
        return !state;
}

static gboolean
get_current_airplane_mode (CsdMediaKeysManager *manager)
{
        CsdMediaKeysManagerPrivate *priv = manager->priv;
        GVariant *v;
        gboolean mode = FALSE;
        if (!priv->rfkill_proxy)
                return mode;
        v = g_dbus_proxy_get_cached_property (priv->rfkill_proxy, "AirplaneMode");
        if (v)
        {
                mode = g_variant_get_boolean (v);
                g_variant_unref (v);
                g_debug ("%s: Current Airplane Mode:%d.", __func__, mode);
        }
        return mode;
}

static void
show_wlan_dialog (CsdMediaKeysManager *manager, gboolean mode)
{
        CsdMediaKeysManagerPrivate *priv = manager->priv;
        const gchar *icon_name;
        //没有触发功能键
        if (!priv->key_grabbed)
                return;
        icon_name = mode ? "airplane-mode-symbolic" : "airplane-mode-disable-symbolic";
        g_message ("wlan changed, mode = %d, icon_name = %s.", mode, icon_name);
        dialog_init (manager);
        csd_osd_window_set_action_custom (CSD_OSD_WINDOW (priv->dialog),
                                          icon_name,
                                          FALSE);
        dialog_show (manager);
        priv->key_grabbed = FALSE;
        priv->airplane_changed = FALSE;
}

static void
set_airplane_mode_cb (GObject *source_object,
                        GAsyncResult *res,
                        gpointer user_data)
{
        CsdMediaKeysManager *manager = CSD_MEDIA_KEYS_MANAGER (user_data);
        CsdMediaKeysManagerPrivate *priv = manager->priv;
        GVariant *v;
        v = g_dbus_proxy_call_finish (G_DBUS_PROXY (source_object),
                                                   res, NULL);
        if (v)
                g_variant_unref (v);
        g_debug ("%s:Set finished, mode: %d", __func__, priv->airplane_mode);
        show_wlan_dialog (manager, priv->airplane_mode);
}

static gboolean
airplane_action_delay (CsdMediaKeysManager *manager)
{
        CsdMediaKeysManagerPrivate *priv = manager->priv;
        gboolean mode;
        priv->airplane_action_id = 0;
        mode = get_current_airplane_mode (manager);
        g_message ("Airplane action activated, Current mode: %d, changed = %d", mode, priv->airplane_changed);
        if (priv->airplane_changed)
        {
                show_wlan_dialog (manager, priv->airplane_mode);
                return;
        }
        priv->airplane_mode = !mode;
        g_dbus_proxy_call (priv->rfkill_proxy, 
                        "SetAirplaneMode", 
                        g_variant_new ("(b)", priv->airplane_mode), 
                        G_DBUS_CALL_FLAGS_NONE, 
                        -1, 
                        NULL, 
                        set_airplane_mode_cb,
                        manager);
        return FALSE;
}

static void
do_wlan_action (CsdMediaKeysManager *manager)
{
        CsdMediaKeysManagerPrivate *priv = manager->priv;
        if (!priv->rfkill_proxy)
                return;
        /* 
         * 重要信息：
         * 采用延迟显示，主要场景如下：
         * 1、被动：飞行模式可能由bios设置，并绑定对应功能键，此时这里只需要显示对应状态的图标即可
                因此加入两个标记变量：
                airplane_changed：判断在接收到按键事件之前是否已经完成更新
                key_grabbed：判断是否接收到按键事件，也就是是否按下了功能键
         * 2、主动：飞行模式由此处进行AirplaneMode状态设置，并显示对应状态图标
         * 3、延迟的作用是等待飞行模式设置完成，防止连续触发，以及等待接收功能键事件
         */
        if (priv->airplane_action_id > 0)
        {
                g_source_remove (priv->airplane_action_id);
        }
        priv->airplane_action_id = g_timeout_add (200, airplane_action_delay, manager);
}

static void
do_desktop_overview_action (CsdMediaKeysManager *manager)
{
        CsdMediaKeysManagerPrivate *priv = manager->priv;
        GError *error = NULL;
        if (!priv->desktop_proxy || priv->screensaver_lock)
                return;
        g_dbus_proxy_call_sync (priv->desktop_proxy,
                                "OverviewToggle",
                                NULL,
                                G_DBUS_CALL_FLAGS_NONE,
                                -1,
                                NULL, 
                                &error);
        if (error)
        {
                g_warning ("Call Desktop OverviewToggle failed:%s", error->message);
                g_error_free (error);
        }
}

// 响应屏幕切换
static void
do_monitor_switch_action (CsdMediaKeysManager *manager)
{
        CsdMediaKeysManagerPrivate *priv = manager->priv;
        GError *error = NULL;
        if (!priv->desktop_proxy || priv->screensaver_lock)
                return;
        g_dbus_proxy_call_sync (priv->desktop_proxy,
                                "RunMonitorSwitchDialog",
                                NULL,
                                G_DBUS_CALL_FLAGS_NONE,
                                -1,
                                NULL, 
                                &error);
        if (error)
        {
                g_warning ("Call Desktop RunMonitorSwitchDialog failed:%s", error->message);
                g_error_free (error);
        }
}

// 执行快捷键
static gboolean
do_action (CsdMediaKeysManager *manager,
           const gchar         *device_node,
           MediaKeyType         type,
           gint64               timestamp)
{
        char *cmd;

        g_message ("Launching action for key type '%d' (on device node %s)", type, device_node);
        manager->priv->key_grabbed = TRUE;
        switch (type) {
        case TOUCHPAD_KEY:
        case TOUCHPAD_ON_KEY:
        case TOUCHPAD_OFF_KEY:
                do_touchpad_action (manager, type);
                break;
        case MUTE_KEY:
        case VOLUME_DOWN_KEY:
        case VOLUME_UP_KEY:
                do_sound_action (manager, device_node, type, TRUE, FALSE);
                break;
        case MUTE_QUIET_KEY:
                do_sound_action (manager, device_node, MUTE_KEY, TRUE, TRUE);
                break;
        case VOLUME_DOWN_QUIET_KEY:
                do_sound_action (manager, device_node, VOLUME_DOWN_KEY, TRUE, TRUE);
                break;
        case VOLUME_UP_QUIET_KEY:
                do_sound_action (manager, device_node, VOLUME_UP_KEY, TRUE, TRUE);
                break;
        case MIC_MUTE_KEY:
                if (!manager->priv->power_mode_enable)
                        do_sound_action (manager, device_node, MIC_MUTE_KEY, FALSE, TRUE);
                break;
        case LOGOUT_KEY:
                do_logout_action (manager);
                break;
        case CAPSLOCK_KEY:
                do_capslock_action (manager);
                break;
        case NUMLOCK_KEY:
                do_numlock_action (manager);
                break;
        case SHUTDOWN_KEY:
                do_shutdown_action (manager);
                break;
        case EJECT_KEY:
                do_eject_action (manager);
                break;
        case HOME_KEY:
                do_home_key_action (manager, timestamp);
                break;
        case SEARCH_KEY:
                cmd = NULL;
                if ((cmd = g_find_program_in_path ("gnome-search-tool")))
                    do_execute_desktop (manager, "gnome-search-tool.desktop", timestamp);
                g_free (cmd);
                break;
        case EMAIL_KEY:
                do_url_action (manager, "mailto", timestamp);
                break;
        case SCREENSAVER_KEY:
                do_lock_screen (manager);
                break;
        case HELP_KEY:
                do_url_action (manager, "ghelp", timestamp);
                break;
        case SCREENSHOT_KEY:
                if (g_file_test ("/usr/bin/screenshot", G_FILE_TEST_EXISTS))
                        execute (manager, "screenshot full -c", 0);
                else if (g_file_test ("/usr/bin/gnome-screenshot", G_FILE_TEST_EXISTS))
                        execute (manager, "gnome-screenshot", 0);
                else 
                        g_warning ("Can't find any screenshot tools.");
                break;
        case WINDOW_SCREENSHOT_KEY:
                if (g_file_test ("/usr/bin/screenshot", G_FILE_TEST_EXISTS))
                        execute (manager, "screenshot screen -c", 0);
                else if (g_file_test ("/usr/bin/gnome-screenshot", G_FILE_TEST_EXISTS))
                        execute (manager, "gnome-screenshot -w", 0);
                else 
                        g_warning ("Can't find any screenshot tools.");
                break;
        case AREA_SCREENSHOT_KEY:
                if (g_file_test ("/usr/bin/screenshot", G_FILE_TEST_EXISTS))
                        execute (manager, "screenshot gui", 0);
                else if (g_file_test ("/usr/bin/gnome-screenshot", G_FILE_TEST_EXISTS))
                        execute (manager, "gnome-screenshot -a", 0);
                else 
                        g_warning ("Can't find any screenshot tools.");
                break;
        case SCREENSHOT_CLIP_KEY:
                execute (manager, "gnome-screenshot -c", 0);
                break;
        case WINDOW_SCREENSHOT_CLIP_KEY:
                execute (manager, "gnome-screenshot -w -c", 0);
                break;
        case AREA_SCREENSHOT_CLIP_KEY:
                execute (manager, "gnome-screenshot -a -c", 0);
                break;
        case TERMINAL_KEY:
                do_terminal_action (manager);
                break;
        case WWW_KEY:
                do_url_action (manager, "http", timestamp);
                break;
        case MEDIA_KEY:
                do_media_action (manager, timestamp);
                break;
        case CALCULATOR_KEY:
                if (!manager->priv->screensaver_lock)
                        execute (manager, "gnome-calculator", 0);
                break;
        case SYSTEM_MONITOR_KEY:
                do_execute_desktop (manager, "gnome-system-monitor", timestamp);
                break;
        case PLAY_KEY:
                return do_multimedia_player_action (manager, NULL, "Play");
        case PAUSE_KEY:
                return do_multimedia_player_action (manager, NULL, "Pause");
        case STOP_KEY:
                return do_multimedia_player_action (manager, NULL, "Stop");
        case PREVIOUS_KEY:
                return do_multimedia_player_action (manager, NULL, "Previous");
        case NEXT_KEY:
                return do_multimedia_player_action (manager, NULL, "Next");
        case REWIND_KEY:
                return do_multimedia_player_action (manager, NULL, "Rewind");
        case FORWARD_KEY:
                return do_multimedia_player_action (manager, NULL, "FastForward");
        case REPEAT_KEY:
                return do_multimedia_player_action (manager, NULL, "Repeat");
        case RANDOM_KEY:
                return do_multimedia_player_action (manager, NULL, "Shuffle");
        case MONITOR_SWITCH_KEY:
                if (!manager->priv->power_mode_enable)
                        do_monitor_switch_action (manager);
                break;
        case WLAN_KEY:
                if (!manager->priv->power_mode_enable)
                        do_wlan_action (manager);
                break;
        case ROTATE_VIDEO_KEY:
                do_video_rotate_action (manager, timestamp);
                break;
        case SCREENREADER_KEY:
                do_screenreader_action (manager);
                break;
        case SCREENMAGNIFIER_KEY:
                do_screenmagnifier_action (manager);
                break;
        case SCREENMAGNIFIER_ZOOM_IN_KEY:
                do_screenmagnifier_zoom_out_action (manager);
                break;
        case SCREENMAGNIFIER_ZOOM_OUT_KEY:
                do_screenmagnifier_zoom_in_action (manager);
                break;
        case ON_SCREEN_KEYBOARD_KEY:
                do_on_screen_keyboard_action (manager);
                break;
        case INCREASE_TEXT_KEY:
        case DECREASE_TEXT_KEY:
                do_text_size_action (manager, type);
                break;
        case TOGGLE_CONTRAST_KEY:
                do_toggle_contrast_action (manager);
                break;
        case POWER_KEY:
                do_config_power_action (manager, "button-power");
                break;
        case SLEEP_KEY:
                do_config_power_action (manager, "button-sleep");
                break;
        case SUSPEND_KEY:
                do_config_power_action (manager, "button-suspend");
                break;
        case HIBERNATE_KEY:
                do_config_power_action (manager, "button-hibernate");
                break;
        case SCREEN_BRIGHTNESS_UP_KEY:
        case SCREEN_BRIGHTNESS_DOWN_KEY:
                do_screen_brightness_action (manager, type);
                break;
        case KEYBOARD_BRIGHTNESS_UP_KEY:
        case KEYBOARD_BRIGHTNESS_DOWN_KEY:
        case KEYBOARD_BRIGHTNESS_TOGGLE_KEY:
                do_keyboard_brightness_action (manager, type);
                break;
        case BATTERY_KEY:
                do_execute_desktop (manager, "gnome-power-statistics.desktop", timestamp);
                break;
        case CAMERA_KEY:
                do_camera_osd_action(manager, do_camera_exist());
                break;
        case SETUP_KEY:
                if (!manager->priv->power_mode_enable)
                        do_execute_desktop (manager, "found-control-center.desktop", timestamp);
                break;
        case CYCLE_WINDOWS_KEY:
                if (!manager->priv->power_mode_enable)
                        do_desktop_overview_action (manager);
                break;
        /* Note, no default so compiler catches missing keys */
        case CUSTOM_KEY:
                g_assert_not_reached ();
        }

        return FALSE;
}

/**
 * <signal name="AcceleratorActivated">
     <arg type="u" name="action"/>
     <arg type="u" name="timestamp"/>
     <arg type="s" name="deviceNode"/>
   </signal>
*/
static void
on_accelerator_activated (GDBusProxy          *grabber,
                          char                *sender,
                          char                *signal,
                          GVariant            *parameters,
                          CsdMediaKeysManager *manager)
{
        if (g_strcmp0 (signal, "AcceleratorActivated"))
                return;

        CsdMediaKeysManagerPrivate *priv = CSD_MEDIA_KEYS_MANAGER_GET_PRIVATE (manager);
        guint i;
        guint deviceid;
        guint timestamp;
        guint accel_id = 0;
        gchar *device_node = NULL;

        deviceid = 0;
        g_variant_get (parameters, "(uus)", &accel_id, &timestamp, &device_node);

        if (!device_node && !csd_is_wayland_session ())
                device_node = xdevice_get_device_node (deviceid);

        g_debug ("Received accel id %u (device-node: %s, timestamp: %u)",
                 accel_id, device_node, timestamp);

        for (i = 0; i < priv->keys->len; i++) {
                MediaKey *key;
                guint j;

                key = g_ptr_array_index (priv->keys, i);

                for (j = 0; j < key->accel_ids->len; j++) {
                        if (g_array_index (key->accel_ids, guint, j) == accel_id)
                                break;
                }
                if (j >= key->accel_ids->len)
                        continue;

                if (key->key_type == CUSTOM_KEY)
                        do_custom_action (manager, device_node, key, timestamp);
                else
                        do_action (manager, device_node, key->key_type, timestamp);

                g_free (device_node);
                return;
        }

        g_warning ("Could not find accelerator for accel id %u", accel_id);
        g_free (device_node);
}

static void
on_key_grabber_ready (CsdMediaKeysManager *manager)
{
        CsdMediaKeysManagerPrivate *priv = CSD_MEDIA_KEYS_MANAGER_GET_PRIVATE (manager);

        g_dbus_proxy_set_default_timeout (G_DBUS_PROXY (priv->key_grabber),
                                          G_MAXINT);

        // 关联输入事件信号
        g_signal_connect (priv->key_grabber, "g-signal",
                          G_CALLBACK (on_accelerator_activated), manager);

        init_kbd (manager);
}

static void
cdos_presence_changed (CsdMediaKeysManager *manager)
{
        CsdMediaKeysManagerPrivate *priv = CSD_MEDIA_KEYS_MANAGER_GET_PRIVATE (manager);
        gchar *name_owner;

        name_owner = g_dbus_proxy_get_name_owner (G_DBUS_PROXY (priv->key_grabber));
        g_ptr_array_set_size (priv->keys, 0);

        if (name_owner) {
                on_key_grabber_ready (manager);
        }
        else {
                g_signal_handlers_disconnect_by_func (priv->key_grabber, on_accelerator_activated, manager);
        }
}

// 更新主题
static void
update_theme_settings (GSettings           *settings,
                       const char          *key,
                       CsdMediaKeysManager *manager)
{
        char *theme;

        theme = g_settings_get_string (manager->priv->interface_settings, key);
        if (g_str_equal (theme, HIGH_CONTRAST)) {
                g_free (theme);
        } else {
                if (g_str_equal (key, "gtk-theme")) {
                        g_free (manager->priv->gtk_theme);
                        manager->priv->gtk_theme = theme;
                } else {
                        g_free (manager->priv->icon_theme);
                        manager->priv->icon_theme = theme;
                }
        }
}

// 初始化快捷键
static gboolean
start_media_keys_idle_cb (CsdMediaKeysManager *manager)
{
        CsdMediaKeysManagerPrivate *priv = manager->priv;
        //char *theme_name;

        g_debug ("Starting media_keys manager");
        cdos_settings_profile_start (NULL);

        priv->keys = g_ptr_array_new_with_free_func ((GDestroyNotify) media_key_unref);
        priv->keys_to_sync = g_hash_table_new_full (g_direct_hash, g_direct_equal, (GDestroyNotify) media_key_unref, NULL);

        gvc_mixer_control_open (manager->priv->volume);

        manager->priv->settings = g_settings_new (SETTINGS_BINDING_DIR);
        g_signal_connect (manager->priv->settings, "changed",
                          G_CALLBACK (gsettings_changed_cb), manager);
        g_signal_connect (manager->priv->settings, "changed::custom-keybindings",
                          G_CALLBACK (gsettings_custom_changed_cb), manager);
        manager->priv->custom_settings =
            g_hash_table_new_full (g_str_hash, g_str_equal,
                                   g_free, g_object_unref);

        priv->key_grabber = csd_get_key_grabber_proxy ();
        g_signal_connect_swapped (manager->priv->key_grabber, "notify::g-name-owner",
                                  G_CALLBACK (cdos_presence_changed), manager);
        cdos_presence_changed (manager);

        /* Sound events */
        ca_context_create (&manager->priv->ca);
        ca_context_set_driver (manager->priv->ca, "pulse");
        ca_context_change_props (manager->priv->ca, 0,
                                 CA_PROP_APPLICATION_ID, "org.gnome.VolumeControl",
                                 NULL);
        manager->priv->gtksettings = gtk_settings_get_for_screen (gdk_screen_get_default ());
        //g_object_get (G_OBJECT (manager->priv->gtksettings), "gtk-sound-theme-name", &theme_name, NULL);
        //if (theme_name)
        //        ca_context_change_props (manager->priv->ca, CA_PROP_CANBERRA_XDG_THEME_NAME, theme_name, NULL);
        //g_free (theme_name);
        //g_signal_connect (manager->priv->gtksettings, "notify::gtk-sound-theme-name",
        //                  G_CALLBACK (sound_theme_changed), manager);

        /* for the power plugin interface code */
        manager->priv->power_settings = g_settings_new (SETTINGS_POWER_DIR);

        /* Logic from http://git.gnome.org/browse/gnome-shell/tree/js/ui/status/accessibility.js#n163 */
        manager->priv->interface_settings = g_settings_new (SETTINGS_INTERFACE_DIR);
        g_signal_connect (G_OBJECT (manager->priv->interface_settings), "changed::gtk-theme",
                          G_CALLBACK (update_theme_settings), manager);
        g_signal_connect (G_OBJECT (manager->priv->interface_settings), "changed::icon-theme",
                          G_CALLBACK (update_theme_settings), manager);
        manager->priv->gtk_theme = g_settings_get_string (manager->priv->interface_settings, "gtk-theme");
        if (g_str_equal (manager->priv->gtk_theme, HIGH_CONTRAST)) {
            g_free (manager->priv->gtk_theme);
            manager->priv->gtk_theme = NULL;
        }
        manager->priv->icon_theme = g_settings_get_string (manager->priv->interface_settings, "icon-theme");

        GSettings *settings = g_settings_new ("org.cdos.sounds");
        gboolean enabled = g_settings_get_boolean(settings, "login-enabled");
        gchar *sound = g_settings_get_string (settings, "login-file");
        if (enabled) {
            ca_context_play (manager->priv->ca, 1, CA_PROP_MEDIA_FILENAME, sound, NULL);
        }
        g_free(sound);
        g_object_unref (settings);

        g_debug ("Starting mpris controller");
        manager->priv->mpris_controller = mpris_controller_new ();

        cdos_settings_profile_end (NULL);

        manager->priv->start_idle_id = 0;

        return FALSE;
}

// 启动快捷键管理功能
gboolean
csd_media_keys_manager_start (CsdMediaKeysManager *manager,
                              GError             **error)
{
        const char * const subsystems[] = { "input", "usb", "sound", NULL };

        cdos_settings_profile_start (NULL);

#ifdef HAVE_GUDEV
        manager->priv->streams = g_hash_table_new (g_direct_hash, g_direct_equal);
        manager->priv->udev_client = g_udev_client_new (subsystems);
#endif

        /* initialise Volume handler
         *
         * We do this one here to force checking gstreamer cache, etc.
         * The rest (grabbing and setting the keys) can happen in an
         * idle.
         */
        cdos_settings_profile_start ("gvc_mixer_control_new");

        manager->priv->volume = gvc_mixer_control_new ("Cdos Volume Control Media Keys");

        g_signal_connect (manager->priv->volume,
                          "state-changed",
                          G_CALLBACK (on_control_state_changed),
                          manager);
        g_signal_connect (manager->priv->volume,
                          "default-sink-changed",
                          G_CALLBACK (on_control_default_sink_changed),
                          manager);
        g_signal_connect (manager->priv->volume,
                          "default-source-changed",
                          G_CALLBACK (on_control_default_source_changed),
                          manager);
        g_signal_connect (manager->priv->volume,
                          "stream-removed",
                          G_CALLBACK (on_control_stream_removed),
                          manager);

        cdos_settings_profile_end ("gvc_mixer_control_new");

        manager->priv->start_idle_id = g_idle_add ((GSourceFunc) start_media_keys_idle_cb, manager);

        register_manager (manager_object);

        cdos_settings_profile_end (NULL);

        return TRUE;
}

void
csd_media_keys_manager_stop (CsdMediaKeysManager *manager)
{
        CsdMediaKeysManagerPrivate *priv = manager->priv;
        GList *l;

        g_debug ("Stopping media_keys manager");
        GdkScreen *screen;
        screen = gdk_screen_get_default ();
        if (manager->priv->monitors_changed_id)
        {
                g_signal_handler_disconnect (screen, manager->priv->monitors_changed_id);
                manager->priv->monitors_changed_id = 0;
        }

        if (priv->bus_cancellable != NULL) {
                g_cancellable_cancel (priv->bus_cancellable);
                g_object_unref (priv->bus_cancellable);
                priv->bus_cancellable = NULL;
        }
        if (manager->priv->introspection_data != NULL) {
                g_dbus_node_info_unref (manager->priv->introspection_data);
                manager->priv->introspection_data = NULL;
        }
        if (priv->airplane_action_id > 0)
        {
                g_source_remove (priv->airplane_action_id);
                priv->airplane_action_id = 0;
        }
        if (priv->caps_lock_action_id > 0)
        {
                g_source_remove (priv->caps_lock_action_id);
                priv->caps_lock_action_id = 0;
        }
        if (priv->num_lock_action_id > 0)
        {
                g_source_remove (priv->num_lock_action_id);
                priv->num_lock_action_id = 0;
        }
        if (manager->priv->gtksettings != NULL) {
                //g_signal_handlers_disconnect_by_func (manager->priv->gtksettings, sound_theme_changed, manager);
                manager->priv->gtksettings = NULL;
        }

        if (manager->priv->ca) {
                ca_context_destroy (manager->priv->ca);
                manager->priv->ca = NULL;
        }

#ifdef HAVE_GUDEV
        if (priv->streams) {
                g_hash_table_destroy (priv->streams);
                priv->streams = NULL;
        }
        if (priv->udev_client) {
                g_object_unref (priv->udev_client);
                priv->udev_client = NULL;
        }
#endif /* HAVE_GUDEV */

        if (priv->logind_proxy) {
                g_object_unref (priv->logind_proxy);
                priv->logind_proxy = NULL;
        }

        if (priv->power_mode) {
                g_object_unref (priv->power_mode);
                priv->power_mode = NULL;
        }

        if (priv->power_brightness) {
                g_object_unref (priv->power_brightness);
                priv->power_brightness = NULL;
        }

        if (priv->shortcuts) {
                g_object_unref (priv->shortcuts);
                priv->shortcuts = NULL;
        }

        if (priv->settings) {
                g_object_unref (priv->settings);
                priv->settings = NULL;
        }

        if (priv->power_settings) {
                g_object_unref (priv->power_settings);
                priv->power_settings = NULL;
        }

        if (priv->power_screen_proxy) {
                g_object_unref (priv->power_screen_proxy);
                priv->power_screen_proxy = NULL;
        }

        if (priv->power_keyboard_proxy) {
                g_object_unref (priv->power_keyboard_proxy);
                priv->power_keyboard_proxy = NULL;
        }

        if (priv->rfkill_proxy) {
                g_object_unref (priv->rfkill_proxy);
                priv->rfkill_proxy = NULL;
        }

        if (priv->desktop_proxy) {
                g_object_unref (priv->desktop_proxy);
                priv->desktop_proxy = NULL;
        }

        if (priv->mpris_controller) {
                g_object_unref (priv->mpris_controller);
                priv->mpris_controller = NULL;
        }

        if (priv->upower_proxy) {
                g_object_unref (priv->upower_proxy);
                priv->upower_proxy = NULL;
        }

        if (priv->cancellable != NULL) {
                g_cancellable_cancel (priv->cancellable);
                g_object_unref (priv->cancellable);
                priv->cancellable = NULL;
        }

        if (priv->connection != NULL) {
                g_object_unref (priv->connection);
                priv->connection = NULL;
        }

        if (priv->keys != NULL) {
                g_ptr_array_free (priv->keys, TRUE);
                priv->keys = NULL;
        }

        if (priv->stream) {
                g_object_unref (priv->stream);
                priv->stream = NULL;
        }

        if (priv->volume) {
                g_object_unref (priv->volume);
                priv->volume = NULL;
        }

        if (priv->dialog != NULL) {
                gtk_widget_destroy (priv->dialog);
                priv->dialog = NULL;
        }

        if (priv->dialog_old != NULL) {
                gtk_widget_destroy (priv->dialog_old);
                priv->dialog_old = NULL;
        }

        if (priv->media_players != NULL) {
                for (l = priv->media_players; l; l = l->next) {
                        MediaPlayer *mp = l->data;
                        g_free (mp->application);
                        g_free (mp);
                }
                g_list_free (priv->media_players);
                priv->media_players = NULL;
        }
}

static GObject *
csd_media_keys_manager_constructor (GType                  type,
                                    guint                  n_construct_properties,
                                    GObjectConstructParam *construct_properties)
{
        CsdMediaKeysManager      *media_keys_manager;

        media_keys_manager = CSD_MEDIA_KEYS_MANAGER (G_OBJECT_CLASS (csd_media_keys_manager_parent_class)->constructor (type,
                                                                                                                         n_construct_properties,
                                                                                                                         construct_properties));

        return G_OBJECT (media_keys_manager);
}

static void
csd_media_keys_manager_class_init (CsdMediaKeysManagerClass *klass)
{
        GObjectClass   *object_class = G_OBJECT_CLASS (klass);

        object_class->constructor = csd_media_keys_manager_constructor;
        object_class->finalize = csd_media_keys_manager_finalize;
}

static void
inhibit_done (GObject      *source,
              GAsyncResult *result,
              gpointer      user_data)
{
        GDBusProxy *proxy = G_DBUS_PROXY (source);
        CsdMediaKeysManager *manager = CSD_MEDIA_KEYS_MANAGER (user_data);
        GError *error = NULL;
        GVariant *res;
        GUnixFDList *fd_list = NULL;
        gint idx;

        res = g_dbus_proxy_call_with_unix_fd_list_finish (proxy, &fd_list, result, &error);
        if (res == NULL) {
                g_warning ("Unable to inhibit keypresses: %s", error->message);
                g_error_free (error);
        } else {
                g_variant_get (res, "(h)", &idx);
                manager->priv->inhibit_keys_fd = g_unix_fd_list_get (fd_list, idx, &error);
                if (manager->priv->inhibit_keys_fd == -1) {
                        g_warning ("Failed to receive system inhibitor fd: %s", error->message);
                        g_error_free (error);
                }
                g_debug ("System inhibitor fd is %d", manager->priv->inhibit_keys_fd);
                g_object_unref (fd_list);
                g_variant_unref (res);
        }
}

static MediaKeyType
get_media_key_type_from_settings_key (const gchar *settings_key)
{
        gint i;
        for (i = 0; i < G_N_ELEMENTS (media_keys); i++) {
                if (g_strcmp0 (media_keys[i].settings_key, settings_key) == 0)
                {
                        return media_keys[i].key_type;
                }
        }
        return UNKNOWN_KEY;
}

// 响应com.nfs.PowerMode接口信号变化
static void
proxy_on_signal_changed (GDBusProxy *proxy,
                        char       *sender_name,
                        char       *signal_name,
                        GVariant   *parameters,
                        gpointer    user_data)
{
        if (g_strcmp0 (signal_name, "Changed"))
                return;
        //show dialog window.
        CsdMediaKeysManager *manager = CSD_MEDIA_KEYS_MANAGER (user_data);
        gint mode;
        g_variant_get (parameters, "(i)", &mode);
        const gchar *icon_name = NULL;
        if (mode == 1)
        {
                icon_name = "power-mode-saving";
        }
        else if (mode == 2) 
        {
                icon_name = "power-mode-performance";
        } else {
                icon_name = "power-mode-balance";
        }
        g_debug ("Current power mode: %d, icon:%s", mode, icon_name);
        dialog_init (manager);
	csd_osd_window_set_action_custom (CSD_OSD_WINDOW (manager->priv->dialog),
                                          icon_name,
                                          FALSE);
	dialog_show (manager);
}

// 响应亮度变化
static void
brightness_on_signal_changed (GDBusProxy *proxy,
                        char       *sender_name,
                        char       *signal_name,
                        GVariant   *parameters,
                        gpointer    user_data)
{
        CsdMediaKeysManager *manager = CSD_MEDIA_KEYS_MANAGER (user_data);
        if (g_strcmp0 (signal_name, "Changed"))
                return;
        gint brightness = 0;
        g_variant_get (parameters, "(i)", &brightness);
        g_debug ("Current brightness: %d", brightness);
        show_brightness_dialog (manager, brightness);
}

// 快捷键信号响应
static void
shortcuts_on_signal_changed (GDBusProxy *proxy,
                        char       *sender_name,
                        char       *signal_name,
                        GVariant   *parameters,
                        gpointer    user_data)
{

        if (g_strcmp0 (signal_name, "Activate"))
                return;
        gchar *device_node = NULL;
        CsdMediaKeysManager *manager = CSD_MEDIA_KEYS_MANAGER (user_data);
        const gchar *settings_key = NULL;
        g_variant_get (parameters, "(s)", &settings_key);
        MediaKeyType key_type = get_media_key_type_from_settings_key (settings_key);

        g_message ("Settings key activate: %s, key_type: %d", settings_key, key_type);
        if (key_type != UNKNOWN_KEY)
        {
                manager->priv->key_grabbed = TRUE;
                device_node = xdevice_get_device_node (0);
                switch (key_type)
                {
                        case MIC_MUTE_KEY:
                                do_sound_action (manager, device_node, MIC_MUTE_KEY, FALSE, TRUE);
                                break;
                        case MONITOR_SWITCH_KEY:
                                do_monitor_switch_action (manager);
                                break;
                        case WLAN_KEY:
                                do_wlan_action (manager);
                                break;
                        case SETUP_KEY:
                                do_execute_desktop (manager, "found-control-center.desktop", 0);
                                break;
                        case CYCLE_WINDOWS_KEY:
                                do_desktop_overview_action (manager);
                                break;
                        default:
                                break;
                }
                //do_action (manager, device_node, key_type, 0);
                g_free (device_node);
        }
}

static void
power_mode_get_enabled (GObject *source_object,
                        GAsyncResult *res,
                        gpointer user_data)
{
        GDBusProxy *proxy = G_DBUS_PROXY (source_object);
        CsdMediaKeysManager *manager = CSD_MEDIA_KEYS_MANAGER (user_data);
        GError *error = NULL;
        GVariant *v;
        gboolean enabled = FALSE;
        v = g_dbus_proxy_call_finish (proxy, res, &error);
        if (error)
        {
                g_debug ("Get power mode enabled failed: %s", error->message);
                g_error_free (error);
                return;
        }
        g_variant_get (v, "(b)", &enabled);
        g_debug ("Change power mode enabled:%d", enabled);
        manager->priv->power_mode_enable = enabled;
        if (enabled)
        {
                /* Fix: Fn+Q功能键事件，会导致其它功能键失效（无法接收事件）;
                 * 因此将失效功能键也改为由内核上报事件，并响应相关处理
                 */
                manager->priv->power_brightness = csd_get_power_brightness_proxy ();
                manager->priv->shortcuts = csd_get_shortcuts_proxy ();
                g_signal_connect (proxy, 
                                  "g-signal", 
                                  G_CALLBACK (proxy_on_signal_changed), 
                                  manager);
                if (manager->priv->power_brightness)
                {
                        g_signal_connect (manager->priv->power_brightness, 
                                  "g-signal", 
                                  G_CALLBACK (brightness_on_signal_changed), 
                                  manager);
                        
                }
                if (manager->priv->shortcuts)
                {
                        g_signal_connect (manager->priv->shortcuts, 
                                  "g-signal", 
                                  G_CALLBACK (shortcuts_on_signal_changed), 
                                  manager);
                }
        }
        g_variant_unref (v);
}

static void
monitors_changed_cb (GdkScreen    *screen,
		     CsdMediaKeysManager *manager)
{
        if (manager->priv->dialog)
        {
                gtk_widget_destroy (manager->priv->dialog);
                manager->priv->dialog = NULL;
        }
}

static void
keyboard_backlight_on_signal_changed (GDBusProxy *proxy,
                                      char       *sender_name,
                                      char       *signal_name,
                                      GVariant   *parameters,
                                      gpointer    user_data)
{
        if (g_strcmp0 (signal_name, "Changed") != 0)
                return;

        CsdMediaKeysManager *manager = CSD_MEDIA_KEYS_MANAGER (user_data);
        gint mode;
        g_variant_get (parameters, "(i)", &mode);
        const gchar *icon_name = NULL;
        if (mode == 0)
        {
                icon_name = "keyboard-backlight-off";
        }
        else if (mode == 1) 
        {
                icon_name = "keyboard-backlight-auto";
        }
        else if (mode == 2)
        {
                icon_name = "keyboard-backlight-50-percent";
        }
        else if (mode == 3)
        {
                icon_name = "keyboard-backlight-100-percent";
        }
        else
        {
                icon_name = "keyboard-backlight-off";
        }

        g_debug ("Current keyboard backlight mode: %d, icon:%s", mode, icon_name);
        dialog_init (manager);
	csd_osd_window_set_action_custom (CSD_OSD_WINDOW (manager->priv->dialog),
                                          icon_name,
                                          FALSE);
	dialog_show (manager);
}

static void
touchpad_mode_on_signal_changed (GDBusProxy *proxy,
                                 char       *sender_name,
                                 char       *signal_name,
                                 GVariant   *parameters,
                                 gpointer    user_data)
{
        if (g_strcmp0 (signal_name, "Changed") != 0)
                return;

        CsdMediaKeysManager *manager = CSD_MEDIA_KEYS_MANAGER (user_data);
        gint mode;
        g_variant_get (parameters, "(i)", &mode);
        const gchar *icon_name = NULL;
        if (mode == 0)
        {
                icon_name = "touchpad-mode-switch";
        }
        else if (mode == 1) 
        {
                icon_name = "touchpad-mode-switch";
        }
        else
        {
                icon_name = "touchpad-mode-switch";
        }

        g_debug ("Current touchpad mode: %d, icon:%s", mode, icon_name);
        dialog_init (manager);
	csd_osd_window_set_action_custom (CSD_OSD_WINDOW (manager->priv->dialog),
                                          icon_name,
                                          FALSE);
	dialog_show (manager);
}

// 初始化快捷键管理
static void
csd_media_keys_manager_init (CsdMediaKeysManager *manager)
{
        GError *error;
        GDBusConnection *bus;
        GdkScreen *screen;
        error = NULL;
        manager->priv = CSD_MEDIA_KEYS_MANAGER_GET_PRIVATE (manager);
        manager->priv->screensaver_lock = FALSE;
        bus = g_bus_get_sync (G_BUS_TYPE_SYSTEM, NULL, &error);
        if (bus == NULL) {
                g_warning ("Failed to connect to system bus: %s",
                           error->message);
                g_error_free (error);
                return;
        }

        manager->priv->logind_proxy =
                g_dbus_proxy_new_sync (bus,
                                       0,
                                       NULL,
                                       SYSTEMD_DBUS_NAME,
                                       SYSTEMD_DBUS_PATH,
                                       SYSTEMD_DBUS_INTERFACE,
                                       NULL,
                                       &error);

        if (manager->priv->logind_proxy == NULL) {
                g_warning ("Failed to connect to systemd: %s",
                           error->message);
                g_error_free (error);
        }

        g_object_unref (bus);

        g_debug ("Adding system inhibitors for power keys");
        manager->priv->inhibit_keys_fd = -1;
        g_dbus_proxy_call_with_unix_fd_list (manager->priv->logind_proxy,
                                             "Inhibit",
                                             g_variant_new ("(ssss)",
                                                            "handle-power-key:handle-suspend-key:handle-hibernate-key",
                                                            g_get_user_name (),
                                                            "Cinnamon handling keypresses",
                                                            "block"),
                                             0,
                                             G_MAXINT,
                                             NULL,
                                             NULL,
                                             inhibit_done,
                                             manager);

        /* Fix: 针对Lenovo N7笔记本，添加Fn+Q切换电源模式的功能键;
         * 该功能键事件，由内核通过DBus上报;
         */
        manager->priv->power_mode = csd_get_power_mode_proxy ();
        manager->priv->power_mode_enable = FALSE;
        if (manager->priv->power_mode)
        {
                g_debug ("Support change power mode.");
                g_dbus_proxy_call (manager->priv->power_mode, 
                                   "GetEnabled", 
                                   NULL, 
                                   G_DBUS_CALL_FLAGS_NONE, 
                                   -1, 
                                   NULL, 
                                   power_mode_get_enabled, 
                                   manager);
        }

        /*
         * 联想笔记本，Fn+空格切换键盘背光模式
         */
        manager->priv->keyboard_backlight = csd_get_keyboard_backlight_proxy ();
        if (manager->priv->keyboard_backlight)
        {
                g_debug ("Support change keyboard backlight mode.");

                g_signal_connect (manager->priv->keyboard_backlight,
                                  "g-signal",
                                  G_CALLBACK (keyboard_backlight_on_signal_changed),
                                  manager);
        }

        /*
         * 联想笔记本，Fn+T切换触摸板模式
         */
        manager->priv->touchpad_mode = csd_get_touchpad_mode_proxy ();
        if (manager->priv->touchpad_mode)
        {
                g_debug ("Support change touchpad mode.");

                g_signal_connect (manager->priv->touchpad_mode,
                                  "g-signal",
                                  G_CALLBACK (touchpad_mode_on_signal_changed),
                                  manager);
        }

        //if primary monitor scaled, the window shown will  be abnormal.
        screen = gdk_screen_get_default ();
        manager->priv->monitors_changed_id = \
                g_signal_connect (G_OBJECT (screen), "monitors-changed",
                                  G_CALLBACK (monitors_changed_cb), 
                                  manager);
}

static void
csd_media_keys_manager_finalize (GObject *object)
{
        CsdMediaKeysManager *media_keys_manager;

        g_return_if_fail (object != NULL);
        g_return_if_fail (CSD_IS_MEDIA_KEYS_MANAGER (object));

        media_keys_manager = CSD_MEDIA_KEYS_MANAGER (object);

        g_return_if_fail (media_keys_manager->priv != NULL);

        if (media_keys_manager->priv->start_idle_id != 0) {
            g_source_remove (media_keys_manager->priv->start_idle_id);
            media_keys_manager->priv->start_idle_id = 0;
        }
        if (media_keys_manager->priv->inhibit_keys_fd != -1)
            close (media_keys_manager->priv->inhibit_keys_fd);

        G_OBJECT_CLASS (csd_media_keys_manager_parent_class)->finalize (object);
}

static void
upower_ready_cb (GObject             *source_object,
                 GAsyncResult        *res,
                 CsdMediaKeysManager *manager)
{
        GError *error = NULL;

        manager->priv->upower_proxy = g_dbus_proxy_new_finish (res, &error);
        if (manager->priv->upower_proxy == NULL) {
                g_warning ("Failed to get proxy for upower: %s",
                           error->message);
                g_error_free (error);
        }
}

static void
power_screen_ready_cb (GObject             *source_object,
                       GAsyncResult        *res,
                       CsdMediaKeysManager *manager)
{
        GError *error = NULL;

        manager->priv->power_screen_proxy = g_dbus_proxy_new_finish (res, &error);
        if (manager->priv->power_screen_proxy == NULL) {
                g_warning ("Failed to get proxy for power (screen): %s",
                           error->message);
                g_error_free (error);
        }
}

static void
power_keyboard_ready_cb (GObject             *source_object,
                         GAsyncResult        *res,
                         CsdMediaKeysManager *manager)
{
        GError *error = NULL;

        manager->priv->power_keyboard_proxy = g_dbus_proxy_new_finish (res, &error);
        if (manager->priv->power_keyboard_proxy == NULL) {
                g_warning ("Failed to get proxy for power (keyboard): %s",
                           error->message);
                g_error_free (error);
        }
}

static void
settings_rfkill_ready_cb (GObject             *source_object,
                         GAsyncResult        *res,
                         CsdMediaKeysManager *manager)
{
        GError *error = NULL;

        manager->priv->rfkill_proxy = g_dbus_proxy_new_finish (res, &error);
        if (manager->priv->rfkill_proxy == NULL) {
                g_warning ("Failed to get proxy for rfkill: %s",
                           error->message);
                g_error_free (error);
                return ;
        }
        manager->priv->airplane_changed = FALSE;
        manager->priv->airplane_mode = get_current_airplane_mode (manager);
}

static void
desktop_ready_cb (GObject             *source_object,
                  GAsyncResult        *res,
                  CsdMediaKeysManager *manager)
{
        GError *error = NULL;

        manager->priv->desktop_proxy = g_dbus_proxy_new_finish (res, &error);
        if (manager->priv->desktop_proxy == NULL) {
                g_warning ("Failed to get proxy for desktop: %s",
                           error->message);
                g_error_free (error);
        }
}

// 关联Rfkill接口属性变化和ScreenSaver接口信号
static void
dbus_connection_on_signal_cb (GDBusConnection *connection,
                        const gchar *sender_name,
                        const gchar *object_path,
                        const gchar *interface_name,
                        const gchar *signal_name,
                        GVariant *parameters,
                        gpointer user_data)
{
        CsdMediaKeysManager *manager = CSD_MEDIA_KEYS_MANAGER (user_data);
        g_debug ("interface: %s, singal:%s.", interface_name, signal_name);
        if (g_strcmp0 (object_path, CSD_DBUS_PATH "/Rfkill") == 0)
        {
                if (g_strcmp0 (signal_name, "PropertiesChanged") == 0)
                {
                        //当AirplaneMode变化时，更新值
                        gboolean new_mode = get_current_airplane_mode (manager);
                        g_debug ("%s:Airplane mode changed: new = %d, old = %d, grabbed = %d;", __func__, 
                                new_mode, 
                                manager->priv->airplane_mode, 
                                manager->priv->key_grabbed);
                        if (new_mode != manager->priv->airplane_mode)
                        {
                                manager->priv->airplane_mode = new_mode;
                                manager->priv->airplane_changed = TRUE;
                        }
                }
        } 
        else if (g_strcmp0 (object_path, SCREENSAVER_DBUS_PATH) == 0)
        {
                if (g_strcmp0 (signal_name, "ActiveChanged") == 0)
                {
                        gboolean lock;
                        g_variant_get (parameters, "(b)", &lock);
                        manager->priv->screensaver_lock = lock;
                        g_debug ("screensaver lock:%d.", lock);
                }
        }
}

static void
handle_method_call (GDBusConnection       *connection,
                    const gchar           *sender,
                    const gchar           *object_path,
                    const gchar           *interface_name,
                    const gchar           *method_name,
                    GVariant              *parameters,
                    GDBusMethodInvocation *invocation,
                    gpointer               user_data)
{
        gchar *device_node;
        gint vol;
        CsdMediaKeysManager *manager = CSD_MEDIA_KEYS_MANAGER (user_data);

        g_debug ("Calling method '%s.%s' for Power",
                 interface_name, method_name);

        if (g_str_equal(interface_name, CSD_MEDIA_KEYS_DBUS_NAME)) {
                device_node = xdevice_get_device_node (0);
                if (g_str_equal (method_name, "VolumeStepDown")) {
                        vol = do_sound_action (manager, device_node, VOLUME_DOWN_KEY, TRUE, FALSE);
                } else if (g_str_equal (method_name, "VolumeStepUp")) {
                        vol = do_sound_action (manager, device_node, VOLUME_UP_KEY, TRUE, FALSE);
                }

                g_dbus_method_invocation_return_value (invocation,
                                                        g_variant_new ("(i)", vol));
        } else {
                g_warning ("not recognised interface: %s", interface_name);
        }
}

static const GDBusInterfaceVTable interface_vtable =
{
        handle_method_call,
        NULL,
        NULL,
};

static void
on_bus_gotten (GObject             *source_object,
               GAsyncResult        *res,
               CsdMediaKeysManager *manager)
{
        GDBusConnection *connection;
        GDBusInterfaceInfo **infos;
        GError *error = NULL;
        guint i;

        if (manager->priv->bus_cancellable == NULL ||
            g_cancellable_is_cancelled (manager->priv->bus_cancellable)) {
                g_warning ("Operation has been cancelled, so not retrieving session bus");
                return;
        }

        connection = g_bus_get_finish (res, &error);
        if (connection == NULL) {
                g_warning ("Could not get session bus: %s", error->message);
                g_error_free (error);
                return;
        }
        manager->priv->connection = connection;

        infos = manager->priv->introspection_data->interfaces;
        for (i = 0; infos[i] != NULL; i++) {
                g_dbus_connection_register_object (connection,
                                                   CSD_MEDIA_KEYS_DBUS_PATH,
                                                   infos[i],
                                                   &interface_vtable,
                                                   manager,
                                                   NULL,
                                                   NULL);
        }

        g_dbus_proxy_new (manager->priv->connection,
                          G_DBUS_PROXY_FLAGS_NONE,
                          NULL,
                          CSD_DBUS_NAME,
                          CSD_DBUS_PATH "/Power",
                          CSD_DBUS_NAME ".Power.Screen",
                          NULL,
                          (GAsyncReadyCallback) power_screen_ready_cb,
                          manager);

        g_dbus_proxy_new (manager->priv->connection,
                          G_DBUS_PROXY_FLAGS_NONE,
                          NULL,
                          CSD_DBUS_NAME,
                          CSD_DBUS_PATH "/Power",
                          CSD_DBUS_NAME ".Power.Keyboard",
                          NULL,
                          (GAsyncReadyCallback) power_keyboard_ready_cb,
                          manager);
        g_dbus_proxy_new (manager->priv->connection,
                          G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START,
                          NULL,
                          CSD_DBUS_NAME,
                          CSD_DBUS_PATH "/Rfkill",
                          CSD_DBUS_NAME ".Rfkill",
                          NULL,
                          (GAsyncReadyCallback) settings_rfkill_ready_cb,
                          manager);
        g_dbus_proxy_new (manager->priv->connection,
                          G_DBUS_PROXY_FLAGS_NONE,
                          NULL,
                          "org.freedesktop.Desktop",
                          "/org/freedesktop/Desktop",
                          "org.freedesktop.Desktop",
                          NULL,
                          (GAsyncReadyCallback) desktop_ready_cb,
                          manager);

        //connect screensaver signal
        g_dbus_connection_signal_subscribe (manager->priv->connection,
                                        NULL,
                                        SCREENSAVER_DBUS_INTERFACE,
                                        "ActiveChanged",
                                        SCREENSAVER_DBUS_PATH,
                                        NULL,
                                        G_DBUS_SIGNAL_FLAGS_NONE,
                                        dbus_connection_on_signal_cb,
                                        manager, NULL);
        g_dbus_connection_signal_subscribe (manager->priv->connection, 
                                        NULL,
                                        "org.freedesktop.DBus.Properties",
                                        "PropertiesChanged",
                                        CSD_DBUS_PATH "/Rfkill",
                                        NULL, 
                                        G_DBUS_SIGNAL_FLAGS_NONE,
                                        dbus_connection_on_signal_cb, 
                                        manager, NULL);
}

static void
register_manager (CsdMediaKeysManager *manager)
{
        manager->priv->introspection_data = g_dbus_node_info_new_for_xml (introspection_xml, NULL);
        manager->priv->bus_cancellable = g_cancellable_new ();

        g_bus_get (G_BUS_TYPE_SESSION,
                   manager->priv->bus_cancellable,
                   (GAsyncReadyCallback) on_bus_gotten,
                   manager);

        g_dbus_proxy_new_for_bus (G_BUS_TYPE_SYSTEM,
                                  G_DBUS_PROXY_FLAGS_NONE,
                                  NULL,
                                  "org.freedesktop.UPower",
                                  "/org/freedesktop/UPower",
                                  "org.freedesktop.UPower",
                                  NULL,
                                  (GAsyncReadyCallback) upower_ready_cb,
                                  manager);
}

CsdMediaKeysManager *
csd_media_keys_manager_new (void)
{
        if (manager_object != NULL) {
                g_object_ref (manager_object);
        } else {
                manager_object = g_object_new (CSD_TYPE_MEDIA_KEYS_MANAGER, NULL);
                g_object_add_weak_pointer (manager_object,
                                           (gpointer *) &manager_object);
        }

        return CSD_MEDIA_KEYS_MANAGER (manager_object);
}
