#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <sys/wait.h>
#include <math.h>
#include <glib.h>
#include <glib/gi18n.h>
#include <gdk/gdkx.h>
#include <X11/extensions/XTest.h>
#include <X11/extensions/dpms.h>
#include <canberra-gtk.h>

#define GNOME_DESKTOP_USE_UNSTABLE_API
#include <libgnome-desktop/gnome-rr.h>

#include "csd-power-manager.h"
#include "gpm-common.h"
#include "utils.h"

enum BacklightHelperCommand {
        BACKLIGHT_HELPER_GET,
        BACKLIGHT_HELPER_GET_MAX,
        BACKLIGHT_HELPER_SET
};

int
gsd_power_backlight_abs_to_percentage (int min, int max, int value)
{
        g_return_val_if_fail (max > min, -1);
        g_return_val_if_fail (value >= min, -1);
        g_return_val_if_fail (value <= max, -1);
        return (((value - min) * 100) / (max - min));
}

/**
 * 所有功能都是调用 backlight-helper 实现
 * settings daemon 需要对外提供 dbus 接口，给应用程序使用
 * */
static gboolean
run_backlight_helper (enum BacklightHelperCommand   command,
                      gchar                        *value,
                      gchar                       **stdout_data,
                      gint                         *exit_status,
                      GError                      **error)
{
        static gchar *helper_args[] = {
                "--get-brightness",
                "--get-max-brightness",
                "--set-brightness",
        };
        gchar *argv[5] = { 0 };

        argv[0] = "pkexec";
        argv[1] = LIBEXECDIR "/csd-backlight-helper";
        argv[2] = helper_args[command];
        argv[3] = value;

        return g_spawn_sync (NULL,
                             command == BACKLIGHT_HELPER_SET ? argv : &argv[1],
                             get_helper_environ (),
                             G_SPAWN_SEARCH_PATH,
                             NULL,
                             NULL,
                             stdout_data,
                             NULL,
                             exit_status,
                             error);
}

static gint64
backlight_helper_get_value (enum BacklightHelperCommand command, GError **error)
{
        gboolean ret;
        gchar *stdout_data = NULL;
        gint exit_status = 0;
        gint64 value = -1;
        gchar *endptr = NULL;

        ret = run_backlight_helper (command, NULL,
                                    &stdout_data, &exit_status, error);
        if (!ret)
                goto out;

        if (WEXITSTATUS (exit_status) != 0) {
                 g_set_error (error,
                             CSD_POWER_MANAGER_ERROR,
                             CSD_POWER_MANAGER_ERROR_FAILED,
                             "gsd-backlight-helper failed: %s",
                             stdout_data ? stdout_data : "No reason");
                goto out;
        }

        value = g_ascii_strtoll (stdout_data, &endptr, 10);

        if (endptr == stdout_data) {
                value = -1;
                g_set_error (error,
                             CSD_POWER_MANAGER_ERROR,
                             CSD_POWER_MANAGER_ERROR_FAILED,
                             "failed to parse value: %s",
                             stdout_data);
                goto out;
        }

        /* out of range */
        if (value > G_MAXINT) {
                value = -1;
                g_set_error (error,
                             CSD_POWER_MANAGER_ERROR,
                             CSD_POWER_MANAGER_ERROR_FAILED,
                             "value out of range: %s",
                             stdout_data);
                goto out;
        }

        if (value < 0) {
                g_set_error (error,
                             CSD_POWER_MANAGER_ERROR,
                             CSD_POWER_MANAGER_ERROR_FAILED,
                             "value negative, but helper did not fail: %s",
                             stdout_data);
                goto out;
        }

out:
        g_free (stdout_data);
        return value;
}

static gboolean
backlight_helper_set_value (gint value,
                            GError **error)
{
        gboolean ret = FALSE;
        gint exit_status = 0;
        gchar *vstr = NULL;

        vstr = g_strdup_printf ("%i", value);
        ret = run_backlight_helper (BACKLIGHT_HELPER_SET, vstr,
                                    NULL, &exit_status, error);
        g_free (vstr);
        return ret;
}

int
backlight_get_abs (GError **error)
{
        return backlight_helper_get_value (BACKLIGHT_HELPER_GET, error);
}

int
backlight_get_percentage (GError **error)
{
        gint now;
        gint value = -1;
        gint max;

        max = backlight_helper_get_value (BACKLIGHT_HELPER_GET_MAX, error);
        if (max < 0)
                return value;
        now = backlight_helper_get_value (BACKLIGHT_HELPER_GET, error);
        if (now < 0)
                return value;
        value = ABS_TO_PERCENTAGE (0, max, now);
	g_print ("%s %d, value is %d\n", __func__, __LINE__, value);
        return value;
}

int
backlight_get_min ()
{
        return 0;
}

int
backlight_get_max (GError **error)
{
        return  backlight_helper_get_value (BACKLIGHT_HELPER_GET_MAX, error);
}

gboolean
backlight_set_percentage ( gint *value,
                          GError **error)
{
        gboolean ret = FALSE;
        gint max;
        guint discrete;

        max = backlight_helper_get_value (BACKLIGHT_HELPER_GET_MAX, error);
        if (max < 0)
                return ret;
        discrete = PERCENTAGE_TO_ABS (0, max, *value);
        ret = backlight_helper_set_value (discrete, error);
        if (ret)
                *value = ABS_TO_PERCENTAGE (0, max, discrete);

        return ret;
}

int
backlight_step_up (GError **error)
{
        gboolean ret = FALSE;
        gint percentage_value = -1;
        gint max;
        gint now;
        gint step;
        guint discrete;

        now = backlight_helper_get_value (BACKLIGHT_HELPER_GET, error);
        if (now < 0)
                return percentage_value;
        max = backlight_helper_get_value (BACKLIGHT_HELPER_GET_MAX, error);
        if (max < 0)
                return percentage_value;
        step = BRIGHTNESS_STEP_AMOUNT (max + 1);
        discrete = MIN (now + step, max);
        ret = backlight_helper_set_value (discrete, error);
        if (ret)
                percentage_value = ABS_TO_PERCENTAGE (0, max, discrete);

        return percentage_value;
}

int
backlight_step_down (GError **error)
{
        gboolean ret = FALSE;
        gint percentage_value = -1;
        gint max;
        gint now;
        gint step;
        guint discrete;

        now = backlight_helper_get_value (BACKLIGHT_HELPER_GET, error);
        if (now < 0)
                return percentage_value;
        max = backlight_helper_get_value (BACKLIGHT_HELPER_GET_MAX, error);
        if (max < 0)
                return percentage_value;
        step = BRIGHTNESS_STEP_AMOUNT (max + 1);
        discrete = MAX (now - step, 0);
        ret = backlight_helper_set_value (discrete, error);
        if (ret)
                percentage_value = ABS_TO_PERCENTAGE (0, max, discrete);

        return percentage_value;
}

int
backlight_set_abs ( guint value,
                   GError **error)
{
        gboolean ret = FALSE;
        ret = backlight_helper_set_value (value, error);

        return ret;
}


