#include "utils.h"

#include <glib.h>
#include <sys/types.h>
#include <dirent.h>
#include<fcntl.h>

typedef struct {
	gint brightness;
	gint max_brightness;
	gint fd;
} Kbd_Backlight;

// 获取亮度值
static gint
read_brightness (Kbd_Backlight *info)
{
	gchar buf[16];
	gchar *end = NULL;
	ssize_t len;
	info->brightness = -1;

	g_return_val_if_fail (info->fd >= 0, info->brightness);

	lseek (info->fd, 0, SEEK_SET);
	len = read (info->fd, buf, G_N_ELEMENTS (buf) - 1);

	if (len > 0) {
		buf[len] = '\0';
		info->brightness = g_ascii_strtoll (buf, &end, 10);

		if (info->brightness < 0 ||
		    info->brightness > info->max_brightness ||
		    end == buf) 
		{
			info->brightness = -1;
			g_warning ("failed to convert brightness: %s", buf);
		}
	}

	return info->brightness;
}

// 设置亮度值
static gint
write_brightness (Kbd_Backlight *info, gint brightness)
{
	gchar *brightness_text = NULL;
	int len = 0;
	int retval = 0;

	g_return_val_if_fail (info->fd >= 0, info->max_brightness);

	brightness_text = g_strdup_printf ("%i", brightness);
	len = strlen (brightness_text);

	lseek (info->fd, 0, SEEK_SET);
	if (write (info->fd, brightness_text, len) != len) {
		g_warning ("failed to write brightness %s to kbd backlight device", brightness_text);
		retval = -1;
	}

	g_free (brightness_text);
	return retval;
}

static gboolean
find_kbd_backlight (Kbd_Backlight *info)
{
	gboolean ret;
	gboolean found = FALSE;
	
	const gchar *filename;
	gchar *dir_path = NULL;
	gchar *end = NULL;
	gchar *path_max = NULL;
	gchar *path_now = NULL;
	gchar *path_hw_changed = NULL;
	gchar *buf_max = NULL;
	gchar *buf_now = NULL;
	GError *error = NULL;

	info->fd = -1;

	dir_path = get_kbd_backlight_dir_path ();
	if (dir_path == NULL) {
		g_warning("faild to find kbd_backlight file");
		goto out;
	}

	// 查询键盘背光亮度最大值
	path_max = g_build_filename (dir_path, "max_brightness", NULL);
	ret = g_file_get_contents (path_max, &buf_max, NULL, &error);
	if (!ret) {
		g_warning ("failed to get max brightness: %s", error->message);
		g_error_free (error);
		goto out;
	}
	
	info->max_brightness = g_ascii_strtoull (buf_max, &end, 10);
	if (info->max_brightness == 0 && end == buf_max) {
		g_warning ("failed to convert max brightness: %s", buf_max);
		goto out;
	}

	// 查看当前键盘背光亮度值
	path_now = g_build_filename (dir_path, "brightness", NULL);
	info->fd = open (path_now, O_RDWR);

	// 检查键盘背光亮度值
	if (read_brightness (info) < 0)
		goto out;

	found = TRUE;
out:
	g_free (dir_path);
	g_free (path_max);
	g_free (path_now);
	g_free (path_hw_changed);
	g_free (buf_max);
	g_free (buf_now);
	return found;
}

int
main (int argc, char *argv[])
{
	guint retval = -1;
	GError *error = NULL;
	gchar *cmd = NULL;
	Kbd_Backlight info;

    	if (argc != 2) {
            	g_warning("Please input right parameters");
        	return retval;
    	}

    	if (!find_kbd_backlight (&info)) {
		g_warning ("cannot find a keyboard backlight");
		return retval;
	}

	g_debug("keyboard backlight brightness is %d  max_brightness is %d", info.brightness, info.max_brightness);
	
	if (g_strcmp0 (argv[1], "on") == 0) {
		if (info.brightness == 0) {
			retval = write_brightness (&info, info.max_brightness);
		}
	} else if (g_strcmp0 (argv[1], "off") == 0) {
		if (info.brightness > 0) {
			retval = write_brightness (&info, 0);
		}
	} else {
		g_warning ("get error parameter: %s", argv[1]);
	}

	if (info.fd > 0)
		close (info.fd);
	if (cmd)
		g_free (cmd);

	return retval;
}

