/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*-
 *
 * Copyright (C) 2007 William Jon McCann <mccann@jhu.edu>
 * Copyright (C) 2011 Richard Hughes <richard@hughsie.com>
 * Copyright (C) 2011 Ritesh Khadgaray <khadgaray@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street - Suite 500, Boston, MA 02110-1335, USA.
 *
 */

#include "config.h"

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/wait.h>
#include <glib/gi18n.h>
#include <gdk/gdkx.h>
#include <gtk/gtk.h>
#include <libupower-glib/upower.h>
#include <libnotify/notify.h>
#include <canberra-gtk.h>
#include <gio/gunixfdlist.h>

#include <sys/time.h>
#include <sys/types.h>

#include <X11/extensions/dpms.h>

#define GNOME_DESKTOP_USE_UNSTABLE_API
#include <libgnome-desktop/gnome-rr.h>
#include <libgnome-desktop/gnome-idle-monitor.h>

#include "gpm-common.h"
#include "gpm-phone.h"
#include "gpm-idletime.h"
#include "cdos-settings-profile.h"
#include "cdos-settings-session.h"
#include "csd-enums.h"
#include "csd-power-manager.h"
#include "csd-power-helper.h"
#include "cdos-settings-utils.h"
#include "utils.h"

#define GNOME_SESSION_DBUS_NAME                 "org.gnome.SessionManager"
#define GNOME_SESSION_DBUS_PATH                 "/org/gnome/SessionManager"
#define GNOME_SESSION_DBUS_PATH_PRESENCE        "/org/gnome/SessionManager/Presence"
#define GNOME_SESSION_DBUS_INTERFACE            "org.gnome.SessionManager"
#define GNOME_SESSION_DBUS_INTERFACE_PRESENCE   "org.gnome.SessionManager.Presence"

#define UPOWER_DBUS_NAME                        "org.freedesktop.UPower"
#define UPOWER_DBUS_PATH                        "/org/freedesktop/UPower"
#define UPOWER_DBUS_DEVICE_PATH                 "/org/freedesktop/UPower/devices/DisplayDevice"
#define UPOWER_DBUS_DEVICE_INTERFACE            "org.freedesktop.UPower.Device"
#define UPOWER_DBUS_PATH_KBDBACKLIGHT           "/org/freedesktop/UPower/KbdBacklight"
#define UPOWER_DBUS_INTERFACE                   "org.freedesktop.UPower"
#define UPOWER_DBUS_INTERFACE_KBDBACKLIGHT      "org.freedesktop.UPower.KbdBacklight"

#define CSD_POWER_SETTINGS_SCHEMA               "org.cdos.settings-daemon.plugins.power"

#define CSD_DBUS_SERVICE                        "org.freedesktop.SettingsDaemon"
#define CSD_DBUS_PATH                           "/org/freedesktop/SettingsDaemon"
#define CSD_POWER_DBUS_PATH                     CSD_DBUS_PATH "/Power"
#define CSD_POWER_DBUS_INTERFACE                "org.freedesktop.SettingsDaemon.Power"
#define CSD_POWER_DBUS_INTERFACE_SCREEN         "org.freedesktop.SettingsDaemon.Power.Screen"
#define CSD_POWER_DBUS_INTERFACE_KEYBOARD       "org.freedesktop.SettingsDaemon.Power.Keyboard"

#define GS_DBUS_NAME                            "org.freedesktop.ScreenSaver"
#define GS_DBUS_PATH                            "/org/freedesktop/ScreenSaver"
#define GS_DBUS_INTERFACE                       "org.freedesktop.ScreenSaver"

#define CSD_POWER_MANAGER_NOTIFY_TIMEOUT_NEVER          0 /* ms */
#define CSD_POWER_MANAGER_NOTIFY_TIMEOUT_SHORT          10 * 1000 /* ms */
#define CSD_POWER_MANAGER_NOTIFY_TIMEOUT_LONG           30 * 1000 /* ms */

#define CSD_POWER_MANAGER_CRITICAL_ALERT_TIMEOUT        5 /* seconds */
#define CSD_POWER_MANAGER_LID_CLOSE_SAFETY_TIMEOUT      30 /* seconds */

#define SYSTEMD_DBUS_NAME                       "org.freedesktop.login1"
#define SYSTEMD_DBUS_PATH                       "/org/freedesktop/login1"
#define SYSTEMD_DBUS_INTERFACE                  "org.freedesktop.login1.Manager"

/* Keep this in sync with gnome-shell */
#define SCREENSAVER_FADE_TIME                           10 /* seconds */

#define XSCREENSAVER_WATCHDOG_TIMEOUT                   120 /* seconds */

#define X11_SCREEN_DELAY_TIMEOUT                        2 /* seconds */

/* add by niyl, pause media player before suspend */
#define MEDIA_PLAYER_RHYTHMBOX "org.mpris.MediaPlayer2.rhythmbox"
#define MEDIA_PLAYER_SPOTIFY "org.mpris.MediaPlayer2.spotify"
#define MEDIA_PLAYER_OBJECT_PATH "/org/mpris/MediaPlayer2"
#define MEDIA_PLAYER_INTERFACE "org.mpris.MediaPlayer2.Player"
/* add end, pause media player before suspend */

// 空闲超时定时器类型
enum {
        CSD_POWER_IDLETIME_NULL_ID,
        CSD_POWER_IDLETIME_DIM_ID,                       // 超时光线变暗(屏幕背光和键盘背光)
        CSD_POWER_IDLETIME_BLANK_ID,                   // 超时息屏
        CSD_POWER_IDLETIME_SLEEP_ID,                    // 超时睡眠
        CSD_POWER_IDLETIME_LOCK_ID,                     // 超时锁屏
        CSD_POWER_IDLETIME_SCREENSAVER_ID       // 超时启动屏保
};

static const gchar introspection_xml[] =
"<node>"
"  <interface name='org.freedesktop.SettingsDaemon.Power'>"
"       <property name='Icon' type='s' access='read'/>"
"       <property name='Tooltip' type='s' access='read'/>"
"       <property name='LowPowerThreshold' type='u' access='readwrite'/>"
"       <property name='LowPowerNotify' type='b' access='readwrite'/>"
"       <property name='LowPowerSaveMode' type='b' access='readwrite'/>"
"       <property name='SaveModeReduceBrightness' type='u' access='readwrite'/>"
"       <method name='GetPrimaryDevice'>"
"         <arg name='device' type='(susdut)' direction='out' />"
"       </method>"
"       <method name='GetDevices'>"
"         <arg name='devices' type='a(susdut)' direction='out' />"
"       </method>"
"       <method name='GetMode'>"
"         <arg name='mode' type='i' direction='out' />"
"       </method>"
"       <method name='SetMode'>"
"         <arg name='mode' type='i' direction='in' />"
"       </method>"
"       <signal name='ModeChanged'>"
"       </signal>"
"       </interface>"
" <interface name='org.freedesktop.SettingsDaemon.Power.Screen'>"
"    <property name='Brightness' type='u' access='readwrite'/>"
"    <property name='Enabled' type='b' access='read'/>"
"    <method name='StepUp'>"
"      <arg type='u' name='new_percentage' direction='out'/>"
"    </method>"
"    <method name='SetValue'>"
"      <arg type='u' name='new_percentage' direction='in'/>"
"      <arg type='u' name='result' direction='out'/>"
"    </method>"
"    <method name='StepDown'>"
"      <arg type='u' name='new_percentage' direction='out'/>"
"    </method>"
"    <method name='GetPercentage'>"
"      <arg type='u' name='new_percentage' direction='out'/>"
"    </method>"
"    <signal name='Changed'>"
"    </signal>"
"  </interface>"
"  <interface name='org.freedesktop.SettingsDaemon.Power.Keyboard'>"
"    <method name='StepUp'>"
"      <arg type='u' name='new_percentage' direction='out'/>"
"    </method>"
"    <method name='StepDown'>"
"      <arg type='u' name='new_percentage' direction='out'/>"
"    </method>"
"    <method name='Toggle'>"
"      <arg type='u' name='new_percentage' direction='out'/>"
"    </method>"
"    <method name='TurnOn'>"
"      <arg type='u' name='result' direction='out'/>"
"    </method>"
"    <method name='TurnOff'>"
"      <arg type='u' name='result' direction='out'/>"
"    </method>"
"  </interface>"
"</node>";

/* on ACPI machines we have 4-16 levels, on others it's ~150 */
#define BRIGHTNESS_STEP_AMOUNT(max) ((max) < 20 ? 1 : (max) / 20)

static void
power_devices_changed (GDBusProxy *proxy,
                        GVariant    *changed_properties,
                        GStrv       invalidated_properties,
                        gpointer    data);

static void
power_status_changed (GDBusProxy *proxy,
                        GVariant    *changed_properties,
                        GStrv       invalidated_properties,
                        gpointer    data);
static gboolean
is_session_active (CsdPowerManager *manager);

/* take a discrete value with offset and convert to percentage */
static int
abs_to_percentage (int min, int max, int value)
{
        g_return_val_if_fail (max > min, -1);
        g_return_val_if_fail (value >= min, -1);
        g_return_val_if_fail (value <= max, -1);
        return (((value - min) * 100) / (max - min));
}
#define ABS_TO_PERCENTAGE(min, max, value) abs_to_percentage(min, max, value)
#define PERCENTAGE_TO_ABS(min, max, value) (min + (((max - min) * value) / 100))

#define CSD_POWER_MANAGER_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), CSD_TYPE_POWER_MANAGER, CsdPowerManagerPrivate))

typedef enum {
        CSD_POWER_IDLE_MODE_NORMAL,
        CSD_POWER_IDLE_MODE_DIM,
        CSD_POWER_IDLE_MODE_BLANK,
        CSD_POWER_IDLE_MODE_SLEEP,
        CSD_POWER_IDLE_MODE_SCREENSAVER
} CsdPowerIdleMode;

struct CsdPowerManagerPrivate
{
        CdosSettingsSession    *session;
        gboolean                 lid_is_closed;
        GSettings               *settings;
        GSettings               *settings_screensaver;
        UpClient                *up_client;
        GDBusNodeInfo           *introspection_data;
        GDBusConnection         *connection;
        GCancellable            *bus_cancellable;
        GDBusProxy              *upower_proxy;
        GDBusProxy              *upower_device_proxy;   // upower代理
        GDBusProxy              *upower_kdb_proxy;      // upower上键盘服务代理， 目前版本upower不支持该功能
        gboolean		backlight_helper_force;
        gchar*                  backlight_helper_preference_args;
        gint                     kbd_brightness_now;
        gint                     kbd_brightness_max;
        gint                     kbd_brightness_old;
        gint                     kbd_brightness_pre_dim;
        GnomeRRScreen           *x11_screen;
        gboolean                 use_time_primary;
        gchar                   *previous_summary;
        GIcon                   *previous_icon;
        GpmPhone                *phone;
        GPtrArray               *devices_array;
        guint                    action_percentage;
        guint                    action_time;
        guint                    critical_percentage;
        guint                    critical_time;
        guint                    low_percentage;        // 低电量百分比
        guint                    low_time;
        gint                    notify_low_num[3];      // 记录低电量通知
        CsdPowerRunMode   run_mode;       //记录电源运行模式
        CsdPowerRunMode   run_mode_bp;   // 记录自动进入节能模式之前的模式
        gboolean                is_settings_run_mode;   // 当前是否正在设置运行模式
        guint                    hd_mode_timer_id;
        gboolean                 hd_helper_spawned;
        gchar			 *hd_mode;
        gint                     pre_dim_brightness; /* level, not percentage */
        UpDevice                *device_composite;
        NotifyNotification      *notification_discharging;
        NotifyNotification      *notification_low;
        ca_context              *canberra_context;
        ca_proplist             *critical_alert_loop_props;
        guint32                  critical_alert_timeout_id;
        GDBusProxy              *screensaver_proxy;
        GDBusProxy              *session_proxy;
        GDBusProxy              *session_presence_proxy;
        GpmIdletime             *idletime;
        GnomeIdleMonitor        *idle_monitor;
        CsdPowerIdleMode         current_idle_mode;
        CsdPowerIdleMode         previous_idle_mode;
        guint                    temporary_unidle_on_ac_id;
        guint                    x11_screen_timer_id;
        guint                    lid_close_safety_timer_id;
        guint                    xscreensaver_watchdog_timer_id;
        gboolean                 is_virtual_machine;

        /* systemd stuff */
        GDBusProxy              *logind_proxy;
        gint                     inhibit_lid_switch_fd;
        gboolean                 inhibit_lid_switch_taken;
        gint                     inhibit_suspend_fd;
        gboolean                 inhibit_suspend_taken;

        /* Fix: 针对Lenovo N7笔记本，添加Fn+Q切换电源模式的功能键;
         * 该功能键事件,由内核通过DBus上报,亮度调节也是如此;
         */
        GDBusProxy      *power_mode;               // 电源模式服务代理
        gboolean        *power_mode_enable;     // TRUE 表示 N7笔记本
        GDBusProxy      *power_brightness;      // 亮度设置服务代理
};

enum {
        PROP_0,
};

static void     csd_power_manager_class_init  (CsdPowerManagerClass *klass);
static void     csd_power_manager_init        (CsdPowerManager      *power_manager);
static void     csd_power_manager_finalize    (GObject              *object);

static UpDevice *engine_get_composite_device (CsdPowerManager *manager, UpDevice *original_device);
static UpDevice *engine_update_composite_device (CsdPowerManager *manager, UpDevice *original_device);
static GIcon    *engine_get_icon (CsdPowerManager *manager);
static gchar    *engine_get_summary (CsdPowerManager *manager);
static void      do_power_action_type (CsdPowerManager *manager, CsdPowerActionType action_type);
static void      do_lid_closed_action (CsdPowerManager *manager);
static void      lock_screensaver (CsdPowerManager *manager);
static void      kill_lid_close_safety_timer (CsdPowerManager *manager);
static void      idle_became_active_cb (GnomeIdleMonitor *idle_monitor, guint watch_id, gpointer user_data);
static void      set_temporary_unidle_on_ac (CsdPowerManager *manager, gboolean enable);
static void
backlight_iface_emit_changed (CsdPowerManager *manager,
		              const char      *interface_name,
			      gint32           value);
static guint turn_on_kbd_backlight();
static guint turn_off_kbd_backlight();
static int set_run_mode (CsdPowerManager *manager, CsdPowerRunMode mode, gboolean manual);
static int get_brightness_percentage (CsdPowerManager *manager);
static int set_brightness_percentage (CsdPowerManager *manager, gint *percent);

G_DEFINE_TYPE_WITH_PRIVATE (CsdPowerManager, csd_power_manager, G_TYPE_OBJECT)

static gpointer manager_object = NULL;

GQuark
csd_power_manager_error_quark (void)
{
        static GQuark quark = 0;
        if (!quark)
                quark = g_quark_from_static_string ("csd_power_manager_error");
        return quark;
}

static gboolean
play_loop_timeout_cb (CsdPowerManager *manager)
{
        ca_context *context;
        context = ca_gtk_context_get_for_screen (gdk_screen_get_default ());
        ca_context_play_full (context, 0,
                              manager->priv->critical_alert_loop_props,
                              NULL,
                              NULL);
        return TRUE;
}

static gboolean
play_loop_stop (CsdPowerManager *manager)
{
        if (manager->priv->critical_alert_timeout_id == 0) {
                g_warning ("no sound loop present to stop");
                return FALSE;
        }

        if (manager->priv->critical_alert_timeout_id) {
            g_source_remove (manager->priv->critical_alert_timeout_id);
            manager->priv->critical_alert_timeout_id = 0;
        }
        
        ca_proplist_destroy (manager->priv->critical_alert_loop_props);

        manager->priv->critical_alert_loop_props = NULL;
        manager->priv->critical_alert_timeout_id = 0;

        return TRUE;
}

static gboolean
play_loop_start (CsdPowerManager *manager,
                 const gchar *id,
                 const gchar *desc,
                 gboolean force,
                 guint timeout)
{
        ca_context *context;

        if (timeout == 0) {
                g_warning ("received invalid timeout");
                return FALSE;
        }

        /* if a sound loop is already running, stop the existing loop */
        if (manager->priv->critical_alert_timeout_id != 0) {
                g_warning ("was instructed to play a sound loop with one already playing");
                play_loop_stop (manager);
        }

        ca_proplist_create (&(manager->priv->critical_alert_loop_props));
        ca_proplist_sets (manager->priv->critical_alert_loop_props,
                          CA_PROP_EVENT_ID, id);
        ca_proplist_sets (manager->priv->critical_alert_loop_props,
                          CA_PROP_EVENT_DESCRIPTION, desc);

        manager->priv->critical_alert_timeout_id = g_timeout_add_seconds (timeout,
                                                                          (GSourceFunc) play_loop_timeout_cb,
                                                                          manager);
        g_source_set_name_by_id (manager->priv->critical_alert_timeout_id,
                                 "[CsdPowerManager] play-loop");

        /* play the sound, using sounds from the naming spec */
        context = ca_gtk_context_get_for_screen (gdk_screen_get_default ());
        ca_context_play (context, 0,
                         CA_PROP_EVENT_ID, id,
                         CA_PROP_EVENT_DESCRIPTION, desc, NULL);
        return TRUE;
}

static void
notify_close_if_showing (NotifyNotification *notification)
{
        gboolean ret;
        GError *error = NULL;

        if (notification == NULL)
                return;
        ret = notify_notification_close (notification, &error);
        if (!ret) {
                g_warning ("failed to close notification: %s",
                           error->message);
                g_error_free (error);
        }
}
\
static const gchar *
get_first_themed_icon_name (GIcon *icon)
{
        const gchar* const *icon_names;
        const gchar *icon_name = NULL;

        /* no icon */
        if (icon == NULL)
                goto out;

        /* just use the first icon */
        icon_names = g_themed_icon_get_names (G_THEMED_ICON (icon));
        if (icon_names != NULL)
                icon_name = icon_names[0];
out:
        return icon_name;
}

typedef enum {
        WARNING_NONE            = 0,
        WARNING_DISCHARGING     = 1,
        WARNING_LOW             = 2,
        WARNING_CRITICAL        = 3,
        WARNING_ACTION          = 4
} CsdPowerManagerWarning;

static GVariant *
get_icon_property_variant (CsdPowerManager  *manager)
{
        GIcon *icon;
        GVariant *retval;

        icon = engine_get_icon (manager);
        if (icon != NULL) {
                char *str;
                str = g_icon_to_string (icon);
                g_object_unref (icon);
                retval = g_variant_new_string (str);
                g_free (str);
        } else {
                retval = g_variant_new_string ("");
        }
        return retval;
}

static GVariant *
get_tooltip_property_variant (CsdPowerManager  *manager)
{
        char *tooltip;
        GVariant *retval;

        tooltip = engine_get_summary (manager);
        retval = g_variant_new_string (tooltip != NULL ? tooltip : "");
        g_free (tooltip);

        return retval;
}

// 释放属性改变信号
static void
engine_emit_changed (CsdPowerManager *manager, const gchar *prop)
{
        GVariantBuilder props_builder;
        GVariant *props_changed = NULL;
        GError *error = NULL;

        /* not yet connected to the bus */
        if (manager->priv->connection == NULL)
                return;

        g_variant_builder_init (&props_builder, G_VARIANT_TYPE ("a{sv}"));

        if (g_strcmp0(prop, "Icon") == 0) {
                g_variant_builder_add (&props_builder, "{sv}", "Icon",
                                       get_icon_property_variant (manager));
        } else if (g_strcmp0(prop, "Tooltip") == 0) {
                g_variant_builder_add (&props_builder, "{sv}", "Tooltip",
                                       get_tooltip_property_variant (manager));
        }

        props_changed = g_variant_new ("(s@a{sv}@as)", CSD_POWER_DBUS_INTERFACE,
                                       g_variant_builder_end (&props_builder),
                                       g_variant_new_strv (NULL, 0));
        g_variant_ref_sink (props_changed);

        if (!g_dbus_connection_emit_signal (manager->priv->connection,
                                            NULL,
                                            CSD_POWER_DBUS_PATH,
                                            "org.freedesktop.DBus.Properties",
                                            "PropertiesChanged",
                                            props_changed,
                                            &error))
                goto out;

 out:
        if (error) {
                g_warning ("%s", error->message);
                g_clear_error (&error);
        }
        if (props_changed)
                g_variant_unref (props_changed);
}

static CsdPowerManagerWarning
engine_get_warning_csr (CsdPowerManager *manager, UpDevice *device)
{
        gdouble percentage;

        /* get device properties */
        g_object_get (device, "percentage", &percentage, NULL);

        if (percentage < 26.0f)
                return WARNING_LOW;
        else if (percentage < 13.0f)
                return WARNING_CRITICAL;
        return WARNING_NONE;
}

static CsdPowerManagerWarning
engine_get_warning_percentage (CsdPowerManager *manager, UpDevice *device)
{
        gdouble percentage;

        /* get device properties */
        g_object_get (device, "percentage", &percentage, NULL);

        if (percentage <= manager->priv->action_percentage)
                return WARNING_ACTION;
        if (percentage <= manager->priv->critical_percentage)
                return WARNING_CRITICAL;
        if (percentage <= manager->priv->low_percentage)
                return WARNING_LOW;
        return WARNING_NONE;
}

static CsdPowerManagerWarning
engine_get_warning_time (CsdPowerManager *manager, UpDevice *device)
{
        UpDeviceKind kind;
        gint64 time_to_empty;

        /* get device properties */
        g_object_get (device,
                      "kind", &kind,
                      "time-to-empty", &time_to_empty,
                      NULL);

        /* this is probably an error condition */
        if (time_to_empty == 0) {
                g_debug ("time zero, falling back to percentage for %s",
                         up_device_kind_to_string (kind));
                return engine_get_warning_percentage (manager, device);
        }

        if (time_to_empty <= manager->priv->action_time)
                return WARNING_ACTION;
        if (time_to_empty <= manager->priv->critical_time)
                return WARNING_CRITICAL;
        if (time_to_empty <= manager->priv->low_time)
                return WARNING_LOW;
        return WARNING_NONE;
}

/**
 * This gets the possible engine state for the device according to the
 * policy, which could be per-percent, or per-time.
 **/
static CsdPowerManagerWarning
engine_get_warning (CsdPowerManager *manager, UpDevice *device)
{
        UpDeviceKind kind;
        UpDeviceState state;
        CsdPowerManagerWarning warning_type;

        /* get device properties */
        g_object_get (device,
                      "kind", &kind,
                      "state", &state,
                      NULL);

        /* default to no engine */
        warning_type = WARNING_NONE;

        /* if the device in question is on ac, don't give a warning */
        if (state == UP_DEVICE_STATE_CHARGING)
                goto out;

        if (kind == UP_DEVICE_KIND_MOUSE ||
            kind == UP_DEVICE_KIND_KEYBOARD) {

                warning_type = engine_get_warning_csr (manager, device);

        } else if (kind == UP_DEVICE_KIND_UPS ||
#if UP_CHECK_VERSION(0,9,5)
                   kind == UP_DEVICE_KIND_MEDIA_PLAYER ||
                   kind == UP_DEVICE_KIND_TABLET ||
                   kind == UP_DEVICE_KIND_COMPUTER ||
#endif
                   kind == UP_DEVICE_KIND_PDA) {

                warning_type = engine_get_warning_percentage (manager, device);

        } else if (kind == UP_DEVICE_KIND_PHONE) {

                warning_type = engine_get_warning_percentage (manager, device);

        } else if (kind == UP_DEVICE_KIND_BATTERY) {
                /* only use the time when it is accurate, and settings is not disabled */
                if (manager->priv->use_time_primary)
                        warning_type = engine_get_warning_time (manager, device);
                else
                        warning_type = engine_get_warning_percentage (manager, device);
        }

        /* If we have no important engines, we should test for discharging */
        if (warning_type == WARNING_NONE) {
                if (state == UP_DEVICE_STATE_DISCHARGING)
                        warning_type = WARNING_DISCHARGING;
        }

 out:
        return warning_type;
}

static gchar *
engine_get_summary (CsdPowerManager *manager)
{
        guint i;
        GPtrArray *array;
        UpDevice *device;
        UpDeviceState state;
        GString *tooltip = NULL;
        gchar *part;
        gboolean is_present;


        /* need to get AC state */
        tooltip = g_string_new ("");

        /* do we have specific device types? */
        array = manager->priv->devices_array;
        for (i=0;i<array->len;i++) {
                device = g_ptr_array_index (array, i);
                g_object_get (device,
                              "is-present", &is_present,
                              "state", &state,
                              NULL);
                if (!is_present)
                        continue;
                if (state == UP_DEVICE_STATE_EMPTY)
                        continue;
                part = gpm_upower_get_device_summary (device);
                if (part != NULL)
                        g_string_append_printf (tooltip, "%s\n", part);
                g_free (part);
        }

        /* remove the last \n */
        g_string_truncate (tooltip, tooltip->len-1);

        g_debug ("tooltip: %s", tooltip->str);

        return g_string_free (tooltip, FALSE);
}

static GIcon *
engine_get_icon_priv (CsdPowerManager *manager,
                      UpDeviceKind device_kind,
                      CsdPowerManagerWarning warning,
                      gboolean use_state)
{
        guint i;
        GPtrArray *array;
        UpDevice *device;
        CsdPowerManagerWarning warning_temp;
        UpDeviceKind kind;
        UpDeviceState state;
        gboolean is_present;

        /* do we have specific device types? */
        array = manager->priv->devices_array;
        for (i=0;i<array->len;i++) {
                device = g_ptr_array_index (array, i);

                /* get device properties */
                g_object_get (device,
                              "kind", &kind,
                              "state", &state,
                              "is-present", &is_present,
                              NULL);

                /* if battery then use composite device to cope with multiple batteries */
                if (kind == UP_DEVICE_KIND_BATTERY)
                        device = engine_get_composite_device (manager, device);

                warning_temp = GPOINTER_TO_INT(g_object_get_data (G_OBJECT(device),
                                                                  "engine-warning-old"));
                if (kind == device_kind && is_present) {
                        if (warning != WARNING_NONE) {
                                if (warning_temp == warning)
                                        return gpm_upower_get_device_icon (device, TRUE);
                                continue;
                        }
                        if (use_state) {
                                if (state == UP_DEVICE_STATE_CHARGING ||
                                    state == UP_DEVICE_STATE_DISCHARGING)
                                        return gpm_upower_get_device_icon (device, TRUE);
                                continue;
                        }
                        return gpm_upower_get_device_icon (device, TRUE);
                }
        }
        return NULL;
}

static GIcon *
engine_get_icon (CsdPowerManager *manager)
{
        GIcon *icon = NULL;


        /* we try CRITICAL: BATTERY, UPS, MOUSE, KEYBOARD */
        icon = engine_get_icon_priv (manager, UP_DEVICE_KIND_BATTERY, WARNING_CRITICAL, FALSE);
        if (icon != NULL)
                return icon;
        icon = engine_get_icon_priv (manager, UP_DEVICE_KIND_UPS, WARNING_CRITICAL, FALSE);
        if (icon != NULL)
                return icon;
        icon = engine_get_icon_priv (manager, UP_DEVICE_KIND_MOUSE, WARNING_CRITICAL, FALSE);
        if (icon != NULL)
                return icon;
        icon = engine_get_icon_priv (manager, UP_DEVICE_KIND_KEYBOARD, WARNING_CRITICAL, FALSE);
        if (icon != NULL)
                return icon;

        /* we try CRITICAL: BATTERY, UPS, MOUSE, KEYBOARD */
        icon = engine_get_icon_priv (manager, UP_DEVICE_KIND_BATTERY, WARNING_LOW, FALSE);
        if (icon != NULL)
                return icon;
        icon = engine_get_icon_priv (manager, UP_DEVICE_KIND_UPS, WARNING_LOW, FALSE);
        if (icon != NULL)
                return icon;
        icon = engine_get_icon_priv (manager, UP_DEVICE_KIND_MOUSE, WARNING_LOW, FALSE);
        if (icon != NULL)
                return icon;
        icon = engine_get_icon_priv (manager, UP_DEVICE_KIND_KEYBOARD, WARNING_LOW, FALSE);
        if (icon != NULL)
                return icon;

        /* we try (DIS)CHARGING: BATTERY, UPS */
        icon = engine_get_icon_priv (manager, UP_DEVICE_KIND_BATTERY, WARNING_NONE, TRUE);
        if (icon != NULL)
                return icon;
        icon = engine_get_icon_priv (manager, UP_DEVICE_KIND_UPS, WARNING_NONE, TRUE);
        if (icon != NULL)
                return icon;

        /* we try PRESENT: BATTERY, UPS */
        icon = engine_get_icon_priv (manager, UP_DEVICE_KIND_BATTERY, WARNING_NONE, FALSE);
        if (icon != NULL)
                return icon;
        icon = engine_get_icon_priv (manager, UP_DEVICE_KIND_UPS, WARNING_NONE, FALSE);
        if (icon != NULL)
                return icon;

        /* do not show an icon */
        return NULL;
}

static UpDevice *
engine_get_composite_device (CsdPowerManager *manager,
                             UpDevice *original_device)
{
        guint battery_devices = 0;
        GPtrArray *array;
        UpDevice *device;
        UpDeviceKind kind;
        UpDeviceKind original_kind;
        guint i;

        /* get the type of the original device */
        g_object_get (original_device,
                      "kind", &original_kind,
                      NULL);

        /* find out how many batteries in the system */
        array = manager->priv->devices_array;
        for (i=0;i<array->len;i++) {
                device = g_ptr_array_index (array, i);
                g_object_get (device,
                              "kind", &kind,
                              NULL);
                if (kind == original_kind)
                        battery_devices++;
        }

        /* just use the original device if only one primary battery */
        if (battery_devices <= 1) {
                g_debug ("using original device as only one primary battery");
                device = original_device;
                goto out;
        }

        /* use the composite device */
        device = manager->priv->device_composite;
out:
        /* return composite device or original device */
        return device;
}

static UpDevice *
engine_update_composite_device (CsdPowerManager *manager,
                                UpDevice *original_device)
{
        guint i;
        gdouble percentage = 0.0;
        gdouble energy = 0.0;
        gdouble energy_full = 0.0;
        gdouble energy_rate = 0.0;
        gdouble energy_total = 0.0;
        gdouble energy_full_total = 0.0;
        gdouble energy_rate_total = 0.0;
        gint64 time_to_empty = 0;
        gint64 time_to_full = 0;
        guint battery_devices = 0;
        gboolean is_charging = FALSE;
        gboolean is_discharging = FALSE;
        gboolean is_fully_charged = TRUE;
        GPtrArray *array;
        UpDevice *device;
        UpDeviceState state;
        UpDeviceKind kind;
        UpDeviceKind original_kind;

        /* get the type of the original device */
        g_object_get (original_device,
                      "kind", &original_kind,
                      NULL);

        /* update the composite device */
        array = manager->priv->devices_array;
        for (i=0;i<array->len;i++) {
                device = g_ptr_array_index (array, i);
                g_object_get (device,
                              "kind", &kind,
                              "state", &state,
                              "energy", &energy,
                              "energy-full", &energy_full,
                              "energy-rate", &energy_rate,
                              NULL);
                if (kind != original_kind)
                        continue;

                /* one of these will be charging or discharging */
                if (state == UP_DEVICE_STATE_CHARGING)
                        is_charging = TRUE;
                if (state == UP_DEVICE_STATE_DISCHARGING)
                        is_discharging = TRUE;
                if (state != UP_DEVICE_STATE_FULLY_CHARGED)
                        is_fully_charged = FALSE;

                /* sum up composite */
                energy_total += energy;
                energy_full_total += energy_full;
                energy_rate_total += energy_rate;
                battery_devices++;
        }

        /* just use the original device if only one primary battery */
        if (battery_devices == 1) {
                g_debug ("using original device as only one primary battery");
                device = original_device;
                goto out;
        }

        /* use percentage weighted for each battery capacity */
        if (energy_full_total > 0.0)
                percentage = 100.0 * energy_total / energy_full_total;

        /* set composite state */
        if (is_charging)
                state = UP_DEVICE_STATE_CHARGING;
        else if (is_discharging)
                state = UP_DEVICE_STATE_DISCHARGING;
        else if (is_fully_charged)
                state = UP_DEVICE_STATE_FULLY_CHARGED;
        else
                state = UP_DEVICE_STATE_UNKNOWN;

        /* calculate a quick and dirty time remaining value */
        if (energy_rate_total > 0) {
                if (state == UP_DEVICE_STATE_DISCHARGING)
                        time_to_empty = 3600 * (energy_total / energy_rate_total);
                else if (state == UP_DEVICE_STATE_CHARGING)
                        time_to_full = 3600 * ((energy_full_total - energy_total) / energy_rate_total);
        }

        /* okay, we can use the composite device */
        device = manager->priv->device_composite;

        g_debug ("printing composite device");
        g_object_set (device,
                      "energy", energy,
                      "energy-full", energy_full,
                      "energy-rate", energy_rate,
                      "time-to-empty", time_to_empty,
                      "time-to-full", time_to_full,
                      "percentage", percentage,
                      "state", state,
                      NULL);

out:
        /* return composite device or original device */
        return device;
}

static void
engine_device_add (CsdPowerManager *manager, UpDevice *device)
{
        CsdPowerManagerWarning warning;
        UpDeviceState state;
        UpDeviceKind kind;
        UpDevice *composite;

        /* assign warning */
        warning = engine_get_warning (manager, device);
        g_object_set_data (G_OBJECT(device),
                           "engine-warning-old",
                           GUINT_TO_POINTER(warning));

        /* get device properties */
        g_object_get (device,
                      "kind", &kind,
                      "state", &state,
                      NULL);

        /* add old state for transitions */
        g_debug ("adding %s with state %s",
                 up_device_get_object_path (device), up_device_state_to_string (state));
        g_object_set_data (G_OBJECT(device),
                           "engine-state-old",
                           GUINT_TO_POINTER(state));

        if (kind == UP_DEVICE_KIND_BATTERY) {
                g_debug ("updating because we added a device");
                composite = engine_update_composite_device (manager, device);

                /* get the same values for the composite device */
                warning = engine_get_warning (manager, composite);
                g_object_set_data (G_OBJECT(composite),
                                   "engine-warning-old",
                                   GUINT_TO_POINTER(warning));
                g_object_get (composite, "state", &state, NULL);
                g_object_set_data (G_OBJECT(composite),
                                   "engine-state-old",
                                   GUINT_TO_POINTER(state));
        }
}

static gboolean
engine_coldplug (CsdPowerManager *manager)
{
        guint i;
        GPtrArray *array = NULL;
        UpDevice *device;
#if ! UP_CHECK_VERSION(0,99,0)
        gboolean ret;
        GError *error = NULL;
 
        /* get devices from UPower */
        ret = up_client_enumerate_devices_sync (manager->priv->up_client, NULL, &error);
        if (!ret) {
                g_warning ("failed to get device list: %s", error->message);
                g_error_free (error);
                goto out;
        }
#endif

        /* connected mobile phones */
        gpm_phone_coldplug (manager->priv->phone);

        /* add to database */
        array = up_client_get_devices2 (manager->priv->up_client);
        for (i = 0; array != NULL && i < array->len; i++) {
                device = g_ptr_array_index (array, i);
                engine_device_add (manager, device);
        }
#if ! UP_CHECK_VERSION(0,99,0)
out:
#endif
        if (array != NULL)
                g_ptr_array_unref (array);
        /* never repeat */
        return FALSE;
}

static void
engine_device_added_cb (UpClient *client, UpDevice *device, CsdPowerManager *manager)
{
        /* add to list */
        g_ptr_array_add (manager->priv->devices_array, g_object_ref (device));
}

static void
#if UP_CHECK_VERSION(0,99,0)
engine_device_removed_cb (UpClient *client, const char *object_path, CsdPowerManager *manager)
 {
        guint i;

        for (i = 0; i < manager->priv->devices_array->len; i++) {
                UpDevice *device = g_ptr_array_index (manager->priv->devices_array, i);

                if (g_strcmp0 (object_path, up_device_get_object_path (device)) == 0) {
                        g_ptr_array_remove_index (manager->priv->devices_array, i);
                        break;
                }
        }
#else
engine_device_removed_cb (UpClient *client, UpDevice *device, CsdPowerManager *manager)
{
        gboolean ret;
        ret = g_ptr_array_remove (manager->priv->devices_array, device);
        if (!ret)
                return;
        engine_recalculate_state (manager);
#endif
}

static void
on_notification_closed (NotifyNotification *notification, gpointer data)
{
    g_object_unref (notification);
}

static void
create_notification (const char *summary,
                     const char *body,
                     const char *icon,
                     NotifyNotification **weak_pointer_location)
{
        NotifyNotification *notification;

        notification = notify_notification_new (summary, body, icon);
        *weak_pointer_location = notification;
        g_object_add_weak_pointer (G_OBJECT (notification),
                                   (gpointer *) weak_pointer_location);
        g_signal_connect (notification, "closed",
                          G_CALLBACK (on_notification_closed), NULL);
}

static void
engine_ups_discharging (CsdPowerManager *manager, UpDevice *device)
{
        const gchar *title;
        gboolean ret;
        gchar *remaining_text = NULL;
        gdouble percentage;
        GError *error = NULL;
        GIcon *icon = NULL;
        gint64 time_to_empty;
        GString *message;
        UpDeviceKind kind;

        /* get device properties */
        g_object_get (device,
                      "kind", &kind,
                      "percentage", &percentage,
                      "time-to-empty", &time_to_empty,
                      NULL);

        if (kind != UP_DEVICE_KIND_UPS)
                return;

        /* only show text if there is a valid time */
        if (time_to_empty > 0)
                remaining_text = gpm_get_timestring (time_to_empty);

        /* TRANSLATORS: UPS is now discharging */
        title = _("UPS Discharging");

        message = g_string_new ("");
        if (remaining_text != NULL) {
                /* TRANSLATORS: tell the user how much time they have got */
                g_string_append_printf (message, _("%s of UPS backup power remaining"),
                                        remaining_text);
        } else {
                g_string_append (message, gpm_device_to_localised_string (device));
        }
        g_string_append_printf (message, " (%.0f%%)", percentage);

        icon = gpm_upower_get_device_icon (device, TRUE);

        /* close any existing notification of this class */
        notify_close_if_showing (manager->priv->notification_discharging);

        /* create a new notification */
        create_notification (title, message->str,
                             get_first_themed_icon_name (icon),
                             &manager->priv->notification_discharging);
        notify_notification_set_timeout (manager->priv->notification_discharging,
                                         CSD_POWER_MANAGER_NOTIFY_TIMEOUT_LONG);
        notify_notification_set_urgency (manager->priv->notification_discharging,
                                         NOTIFY_URGENCY_NORMAL);
        /* TRANSLATORS: this is the notification application name */
        notify_notification_set_app_name (manager->priv->notification_discharging, _("Power"));
        notify_notification_set_hint (manager->priv->notification_discharging,
                                      "transient", g_variant_new_boolean (TRUE));

        /* try to show */
        ret = notify_notification_show (manager->priv->notification_discharging,
                                        &error);
        if (!ret) {
                g_warning ("failed to show notification: %s", error->message);
                g_error_free (error);
                g_object_unref (manager->priv->notification_discharging);
        }
        g_string_free (message, TRUE);
        if (icon != NULL)
                g_object_unref (icon);
        g_free (remaining_text);
}

static CsdPowerActionType
manager_critical_action_get (CsdPowerManager *manager,
                             gboolean         is_ups)
{
        CsdPowerActionType policy;

        policy = g_settings_get_enum (manager->priv->settings, "critical-battery-action");
        if (policy == CSD_POWER_ACTION_SUSPEND) {
                if (is_ups == FALSE
#if ! UP_CHECK_VERSION(0,99,0)
                    && up_client_get_can_suspend (manager->priv->up_client)
#endif
                )
                        return policy;
                return CSD_POWER_ACTION_SHUTDOWN;
        } else if (policy == CSD_POWER_ACTION_HIBERNATE) {
#if ! UP_CHECK_VERSION(0,99,0)
                if (up_client_get_can_hibernate (manager->priv->up_client))
#endif
                        return policy;
                return CSD_POWER_ACTION_SHUTDOWN;
        }

        return policy;
}

static gboolean
manager_critical_action_do (CsdPowerManager *manager,
                            gboolean         is_ups)
{
        CsdPowerActionType action_type;

        /* stop playing the alert as it's too late to do anything now */
        if (manager->priv->critical_alert_timeout_id > 0)
                play_loop_stop (manager);

        action_type = manager_critical_action_get (manager, is_ups);
        g_message ("Do power critical action: %d(%s), is_ups = %d.", action_type, power_action_type_to_string (action_type), is_ups);
        do_power_action_type (manager, action_type);

        return FALSE;
}

static gboolean
manager_critical_action_do_cb (CsdPowerManager *manager)
{
        manager_critical_action_do (manager, FALSE);
        return FALSE;
}

static gboolean
manager_critical_ups_action_do_cb (CsdPowerManager *manager)
{
        manager_critical_action_do (manager, TRUE);
        return FALSE;
}

static gboolean
engine_just_laptop_battery (CsdPowerManager *manager)
{
        UpDevice *device;
        UpDeviceKind kind;
        GPtrArray *array;
        gboolean ret = TRUE;
        guint i;

        /* find if there are any other device types that mean we have to
         * be more specific in our wording */
        array = manager->priv->devices_array;
        for (i=0; i<array->len; i++) {
                device = g_ptr_array_index (array, i);
                g_object_get (device, "kind", &kind, NULL);
                if (kind != UP_DEVICE_KIND_BATTERY) {
                        ret = FALSE;
                        break;
                }
        }
        return ret;
}

static void
engine_charge_low (CsdPowerManager *manager, UpDevice *device)
{
        const gchar *title = NULL;
        gboolean ret;
        gchar *message = NULL;
        gchar *tmp;
        gchar *remaining_text;
        gdouble percentage;
        GIcon *icon = NULL;
        gint64 time_to_empty;
        UpDeviceKind kind;
        GError *error = NULL;

        /* get device properties */
        g_object_get (device,
                      "kind", &kind,
                      "percentage", &percentage,
                      "time-to-empty", &time_to_empty,
                      NULL);

        /* check to see if the batteries have not noticed we are on AC */
        if (kind == UP_DEVICE_KIND_BATTERY) {
                if (!up_client_get_on_battery (manager->priv->up_client)) {
                        g_warning ("ignoring low message as we are not on battery power");
                        goto out;
                }
        }

        if (kind == UP_DEVICE_KIND_BATTERY) {

                /* if the user has no other batteries, drop the "Laptop" wording */
                ret = engine_just_laptop_battery (manager);
                if (ret) {
                        /* TRANSLATORS: laptop battery low, and we only have one battery */
                        title = _("Battery low");
                } else {
                        /* TRANSLATORS: laptop battery low, and we have more than one kind of battery */
                        title = _("Laptop battery low");
                }
                tmp = gpm_get_timestring (time_to_empty);
                remaining_text = g_strconcat ("<b>", tmp, "</b>", NULL);
                g_free (tmp);

                /* TRANSLATORS: tell the user how much time they have got */
                message = g_strdup_printf (_("Approximately %s remaining (%.0f%%)"), remaining_text, percentage);
                g_free (remaining_text);

        } else if (kind == UP_DEVICE_KIND_UPS) {
                /* TRANSLATORS: UPS is starting to get a little low */
                title = _("UPS low");
                tmp = gpm_get_timestring (time_to_empty);
                remaining_text = g_strconcat ("<b>", tmp, "</b>", NULL);
                g_free (tmp);

                /* TRANSLATORS: tell the user how much time they have got */
                message = g_strdup_printf (_("Approximately %s of remaining UPS backup power (%.0f%%)"),
                                           remaining_text, percentage);
                g_free (remaining_text);
        } else if (kind == UP_DEVICE_KIND_MOUSE) {
                /* TRANSLATORS: mouse is getting a little low */
                title = _("Mouse battery low");

                /* TRANSLATORS: tell user more details */
                message = g_strdup_printf (_("Wireless mouse is low in power (%.0f%%)"), percentage);

        } else if (kind == UP_DEVICE_KIND_KEYBOARD) {
                /* TRANSLATORS: keyboard is getting a little low */
                title = _("Keyboard battery low");

                /* TRANSLATORS: tell user more details */
                message = g_strdup_printf (_("Wireless keyboard is low in power (%.0f%%)"), percentage);

        } else if (kind == UP_DEVICE_KIND_PDA) {
                /* TRANSLATORS: PDA is getting a little low */
                title = _("PDA battery low");

                /* TRANSLATORS: tell user more details */
                message = g_strdup_printf (_("PDA is low in power (%.0f%%)"), percentage);

        } else if (kind == UP_DEVICE_KIND_PHONE) {
                /* TRANSLATORS: cell phone (mobile) is getting a little low */
                title = _("Cell phone battery low");

                /* TRANSLATORS: tell user more details */
                message = g_strdup_printf (_("Cell phone is low in power (%.0f%%)"), percentage);

#if UP_CHECK_VERSION(0,9,5)
        } else if (kind == UP_DEVICE_KIND_MEDIA_PLAYER) {
                /* TRANSLATORS: media player, e.g. mp3 is getting a little low */
                title = _("Media player battery low");

                /* TRANSLATORS: tell user more details */
                message = g_strdup_printf (_("Media player is low in power (%.0f%%)"), percentage);

        } else if (kind == UP_DEVICE_KIND_TABLET) {
                /* TRANSLATORS: graphics tablet, e.g. wacom is getting a little low */
                title = _("Tablet battery low");

                /* TRANSLATORS: tell user more details */
                message = g_strdup_printf (_("Tablet is low in power (%.0f%%)"), percentage);

        } else if (kind == UP_DEVICE_KIND_COMPUTER) {
                /* TRANSLATORS: computer, e.g. ipad is getting a little low */
                title = _("Attached computer battery low");

                /* TRANSLATORS: tell user more details */
                message = g_strdup_printf (_("Attached computer is low in power (%.0f%%)"), percentage);
#endif
        }

        /* get correct icon */
        icon = gpm_upower_get_device_icon (device, TRUE);

        /* close any existing notification of this class */
        notify_close_if_showing (manager->priv->notification_low);

        /* create a new notification */
        create_notification (title, message,
                             get_first_themed_icon_name (icon),
                             &manager->priv->notification_low);
        notify_notification_set_timeout (manager->priv->notification_low,
                                         CSD_POWER_MANAGER_NOTIFY_TIMEOUT_LONG);
        notify_notification_set_urgency (manager->priv->notification_low,
                                         NOTIFY_URGENCY_NORMAL);
        notify_notification_set_app_name (manager->priv->notification_low, _("Power"));
        notify_notification_set_hint (manager->priv->notification_low,
                                      "transient", g_variant_new_boolean (TRUE));

        /* try to show */
        ret = notify_notification_show (manager->priv->notification_low,
                                        &error);
        if (!ret) {
                g_warning ("failed to show notification: %s", error->message);
                g_error_free (error);
                g_object_unref (manager->priv->notification_low);
        }

        /* play the sound, using sounds from the naming spec */
        ca_context_play (manager->priv->canberra_context, 0,
                         CA_PROP_EVENT_ID, "battery-low",
                         /* TRANSLATORS: this is the sound description */
                         CA_PROP_EVENT_DESCRIPTION, _("Battery is low"), NULL);

out:
        if (icon != NULL)
                g_object_unref (icon);
        g_free (message);
}

static void
engine_charge_critical (CsdPowerManager *manager, UpDevice *device)
{
        const gchar *title = NULL;
        gboolean ret;
        gchar *message = NULL;
        gdouble percentage;
        GIcon *icon = NULL;
        gint64 time_to_empty;
        CsdPowerActionType policy;
        UpDeviceKind kind;
        GError *error = NULL;

        /* get device properties */
        g_object_get (device,
                      "kind", &kind,
                      "percentage", &percentage,
                      "time-to-empty", &time_to_empty,
                      NULL);

        /* check to see if the batteries have not noticed we are on AC */
        if (kind == UP_DEVICE_KIND_BATTERY) {
                if (!up_client_get_on_battery (manager->priv->up_client)) {
                        g_warning ("ignoring critically low message as we are not on battery power");
                        goto out;
                }
        }

        if (kind == UP_DEVICE_KIND_BATTERY) {

                /* if the user has no other batteries, drop the "Laptop" wording */
                ret = engine_just_laptop_battery (manager);
                if (ret) {
                        /* TRANSLATORS: laptop battery critically low, and only have one kind of battery */
                        title = _("Battery critically low");
                } else {
                        /* TRANSLATORS: laptop battery critically low, and we have more than one type of battery */
                        title = _("Laptop battery critically low");
                }

                /* we have to do different warnings depending on the policy */
                policy = manager_critical_action_get (manager, FALSE);

                /* use different text for different actions */
                if (policy == CSD_POWER_ACTION_NOTHING) {
                        /* TRANSLATORS: tell the use to insert the plug, as we're not going to do anything */
                        message = g_strdup (_("Plug in your AC adapter to avoid losing data."));

                } else if (policy == CSD_POWER_ACTION_SUSPEND) {
                        /* TRANSLATORS: give the user a ultimatum */
                        message = g_strdup_printf (_("Computer will suspend very soon unless it is plugged in."));

                } else if (policy == CSD_POWER_ACTION_HIBERNATE) {
                        /* TRANSLATORS: give the user a ultimatum */
                        message = g_strdup_printf (_("Computer will hibernate very soon unless it is plugged in."));

                } else if (policy == CSD_POWER_ACTION_SHUTDOWN) {
                        /* TRANSLATORS: give the user a ultimatum */
                        message = g_strdup_printf (_("Computer will shutdown very soon unless it is plugged in."));
                }

        } else if (kind == UP_DEVICE_KIND_UPS) {
                gchar *remaining_text;
                gchar *tmp;

                /* TRANSLATORS: the UPS is very low */
                title = _("UPS critically low");
                tmp = gpm_get_timestring (time_to_empty);
                remaining_text = g_strconcat ("<b>", tmp, "</b>", NULL);
                g_free (tmp);

                /* TRANSLATORS: give the user a ultimatum */
                message = g_strdup_printf (_("Approximately %s of remaining UPS power (%.0f%%). "
                                             "Restore AC power to your computer to avoid losing data."),
                                           remaining_text, percentage);
                g_free (remaining_text);
        } else if (kind == UP_DEVICE_KIND_MOUSE) {
                /* TRANSLATORS: the mouse battery is very low */
                title = _("Mouse battery low");

                /* TRANSLATORS: the device is just going to stop working */
                message = g_strdup_printf (_("Wireless mouse is very low in power (%.0f%%). "
                                             "This device will soon stop functioning if not charged."),
                                           percentage);
        } else if (kind == UP_DEVICE_KIND_KEYBOARD) {
                /* TRANSLATORS: the keyboard battery is very low */
                title = _("Keyboard battery low");

                /* TRANSLATORS: the device is just going to stop working */
                message = g_strdup_printf (_("Wireless keyboard is very low in power (%.0f%%). "
                                             "This device will soon stop functioning if not charged."),
                                           percentage);
        } else if (kind == UP_DEVICE_KIND_PDA) {

                /* TRANSLATORS: the PDA battery is very low */
                title = _("PDA battery low");

                /* TRANSLATORS: the device is just going to stop working */
                message = g_strdup_printf (_("PDA is very low in power (%.0f%%). "
                                             "This device will soon stop functioning if not charged."),
                                           percentage);

        } else if (kind == UP_DEVICE_KIND_PHONE) {

                /* TRANSLATORS: the cell battery is very low */
                title = _("Cell phone battery low");

                /* TRANSLATORS: the device is just going to stop working */
                message = g_strdup_printf (_("Cell phone is very low in power (%.0f%%). "
                                             "This device will soon stop functioning if not charged."),
                                           percentage);

#if UP_CHECK_VERSION(0,9,5)
        } else if (kind == UP_DEVICE_KIND_MEDIA_PLAYER) {

                /* TRANSLATORS: the cell battery is very low */
                title = _("Cell phone battery low");

                /* TRANSLATORS: the device is just going to stop working */
                message = g_strdup_printf (_("Media player is very low in power (%.0f%%). "
                                             "This device will soon stop functioning if not charged."),
                                           percentage);
        } else if (kind == UP_DEVICE_KIND_TABLET) {

                /* TRANSLATORS: the cell battery is very low */
                title = _("Tablet battery low");

                /* TRANSLATORS: the device is just going to stop working */
                message = g_strdup_printf (_("Tablet is very low in power (%.0f%%). "
                                             "This device will soon stop functioning if not charged."),
                                           percentage);
        } else if (kind == UP_DEVICE_KIND_COMPUTER) {

                /* TRANSLATORS: the cell battery is very low */
                title = _("Attached computer battery low");

                /* TRANSLATORS: the device is just going to stop working */
                message = g_strdup_printf (_("Attached computer is very low in power (%.0f%%). "
                                             "The device will soon shutdown if not charged."),
                                           percentage);
#endif
        }

        /* get correct icon */
        icon = gpm_upower_get_device_icon (device, TRUE);

        /* close any existing notification of this class */
        notify_close_if_showing (manager->priv->notification_low);

        /* create a new notification */
        create_notification (title, message,
                             get_first_themed_icon_name (icon),
                             &manager->priv->notification_low);
        notify_notification_set_timeout (manager->priv->notification_low,
                                         CSD_POWER_MANAGER_NOTIFY_TIMEOUT_LONG);
        notify_notification_set_urgency (manager->priv->notification_low,
                                         NOTIFY_URGENCY_CRITICAL);
        notify_notification_set_app_name (manager->priv->notification_low, _("Power"));

        /* try to show */
        ret = notify_notification_show (manager->priv->notification_low,
                                        &error);
        if (!ret) {
                g_warning ("failed to show notification: %s", error->message);
                g_error_free (error);
                g_object_unref (manager->priv->notification_low);
        }

        switch (kind) {

        case UP_DEVICE_KIND_BATTERY:
        case UP_DEVICE_KIND_UPS:
                g_debug ("critical charge level reached, starting sound loop");
                play_loop_start (manager,
                                 "battery-caution",
                                 _("Battery is critically low"),
                                 TRUE,
                                 CSD_POWER_MANAGER_CRITICAL_ALERT_TIMEOUT);
                break;

        default:
                /* play the sound, using sounds from the naming spec */
                ca_context_play (manager->priv->canberra_context, 0,
                                 CA_PROP_EVENT_ID, "battery-caution",
                                 /* TRANSLATORS: this is the sound description */
                                 CA_PROP_EVENT_DESCRIPTION, _("Battery is critically low"), NULL);
                break;
        }
out:
        if (icon != NULL)
                g_object_unref (icon);
        g_free (message);
}

static void
engine_charge_action (CsdPowerManager *manager, UpDevice *device)
{
        const gchar *title = NULL;
        gboolean ret;
        gchar *message = NULL;
        GError *error = NULL;
        GIcon *icon = NULL;
        CsdPowerActionType policy;
        guint timer_id;
        UpDeviceKind kind;

        /* get device properties */
        g_object_get (device,
                      "kind", &kind,
                      NULL);

        /* check to see if the batteries have not noticed we are on AC */
        if (kind == UP_DEVICE_KIND_BATTERY) {
                if (!up_client_get_on_battery (manager->priv->up_client)) {
                        g_warning ("ignoring critically low message as we are not on battery power");
                        goto out;
                }
        }

        if (kind == UP_DEVICE_KIND_BATTERY) {

                /* TRANSLATORS: laptop battery is really, really, low */
                title = _("Laptop battery critically low");

                /* we have to do different warnings depending on the policy */
                policy = manager_critical_action_get (manager, FALSE);

                /* use different text for different actions */
                if (policy == CSD_POWER_ACTION_NOTHING) {
                        /* TRANSLATORS: computer will shutdown without saving data */
                        message = g_strdup (_("The battery is below the critical level and "
                                              "this computer will <b>power-off</b> when the "
                                              "battery becomes completely empty."));

                } else if (policy == CSD_POWER_ACTION_SUSPEND) {
                        /* TRANSLATORS: computer will suspend */
                        message = g_strdup (_("The battery is below the critical level and "
                                              "this computer is about to suspend.\n"
                                              "<b>NOTE:</b> A small amount of power is required "
                                              "to keep your computer in a suspended state."));

                } else if (policy == CSD_POWER_ACTION_HIBERNATE) {
                        /* TRANSLATORS: computer will hibernate */
                        message = g_strdup (_("The battery is below the critical level and "
                                              "this computer is about to hibernate."));

                } else if (policy == CSD_POWER_ACTION_SHUTDOWN) {
                        /* TRANSLATORS: computer will just shutdown */
                        message = g_strdup (_("The battery is below the critical level and "
                                              "this computer is about to shutdown."));
                }

                /* wait 20 seconds for user-panic */
                timer_id = g_timeout_add_seconds (20, (GSourceFunc) manager_critical_action_do_cb, manager);
                g_source_set_name_by_id (timer_id, "[CsdPowerManager] battery critical-action");

        } else if (kind == UP_DEVICE_KIND_UPS) {
                /* TRANSLATORS: UPS is really, really, low */
                title = _("UPS critically low");

                /* we have to do different warnings depending on the policy */
                policy = manager_critical_action_get (manager, TRUE);

                /* use different text for different actions */
                if (policy == CSD_POWER_ACTION_NOTHING) {
                        /* TRANSLATORS: computer will shutdown without saving data */
                        message = g_strdup (_("UPS is below the critical level and "
                                              "this computer will <b>power-off</b> when the "
                                              "UPS becomes completely empty."));

                } else if (policy == CSD_POWER_ACTION_HIBERNATE) {
                        /* TRANSLATORS: computer will hibernate */
                        message = g_strdup (_("UPS is below the critical level and "
                                              "this computer is about to hibernate."));

                } else if (policy == CSD_POWER_ACTION_SHUTDOWN) {
                        /* TRANSLATORS: computer will just shutdown */
                        message = g_strdup (_("UPS is below the critical level and "
                                              "this computer is about to shutdown."));
                }

                /* wait 20 seconds for user-panic */
                timer_id = g_timeout_add_seconds (20, (GSourceFunc) manager_critical_ups_action_do_cb, manager);
                g_source_set_name_by_id (timer_id, "[CsdPowerManager] ups critical-action");
        }
        g_message ("Engine charge action,kind:%d, message: %s", kind, message);
        /* not all types have actions */
        if (title == NULL)
                return;

        /* get correct icon */
        icon = gpm_upower_get_device_icon (device, TRUE);

        /* close any existing notification of this class */
        notify_close_if_showing (manager->priv->notification_low);

        /* create a new notification */
        create_notification (title, message,
                             get_first_themed_icon_name (icon),
                             &manager->priv->notification_low);
        notify_notification_set_timeout (manager->priv->notification_low,
                                         CSD_POWER_MANAGER_NOTIFY_TIMEOUT_LONG);
        notify_notification_set_urgency (manager->priv->notification_low,
                                         NOTIFY_URGENCY_CRITICAL);
        notify_notification_set_app_name (manager->priv->notification_low, _("Power"));

        /* try to show */
        ret = notify_notification_show (manager->priv->notification_low,
                                        &error);
        if (!ret) {
                g_warning ("failed to show notification: %s", error->message);
                g_error_free (error);
                g_object_unref (manager->priv->notification_low);
        }

        /* play the sound, using sounds from the naming spec */
        ca_context_play (manager->priv->canberra_context, 0,
                         CA_PROP_EVENT_ID, "battery-caution",
                         /* TRANSLATORS: this is the sound description */
                         CA_PROP_EVENT_DESCRIPTION, _("Battery is critically low"), NULL);
out:
        if (icon != NULL)
                g_object_unref (icon);
        g_free (message);
}

static void
engine_device_changed_cb (UpClient *client, UpDevice *device, CsdPowerManager *manager)
{
        UpDeviceKind kind;
        UpDeviceState state;
        UpDeviceState state_old;
        CsdPowerManagerWarning warning_old;
        CsdPowerManagerWarning warning;

        /* get device properties */
        g_object_get (device,
                      "kind", &kind,
                      NULL);

        /* if battery then use composite device to cope with multiple batteries */
        if (kind == UP_DEVICE_KIND_BATTERY) {
                g_debug ("updating because %s changed", up_device_get_object_path (device));
                device = engine_update_composite_device (manager, device);
        }

        /* get device properties (may be composite) */
        g_object_get (device,
                      "state", &state,
                      NULL);

        g_debug ("%s state is now %s", up_device_get_object_path (device), up_device_state_to_string (state));

        /* see if any interesting state changes have happened */
        state_old = GPOINTER_TO_INT(g_object_get_data (G_OBJECT(device), "engine-state-old"));
        if (state_old != state) {
                if (state == UP_DEVICE_STATE_DISCHARGING) {
                        g_debug ("discharging");
                        engine_ups_discharging (manager, device);
                } else if (state == UP_DEVICE_STATE_FULLY_CHARGED ||
                           state == UP_DEVICE_STATE_CHARGING) {
                        g_debug ("fully charged or charging, hiding notifications if any");
                        notify_close_if_showing (manager->priv->notification_low);
                        notify_close_if_showing (manager->priv->notification_discharging);
                }

                /* save new state */
                g_object_set_data (G_OBJECT(device), "engine-state-old", GUINT_TO_POINTER(state));
        }

        /* check the warning state has not changed */
        warning_old = GPOINTER_TO_INT(g_object_get_data (G_OBJECT(device), "engine-warning-old"));
        warning = engine_get_warning (manager, device);
        if (warning != warning_old) {
                if (warning == WARNING_LOW) {
                        g_debug ("** EMIT: charge-low");
                        engine_charge_low (manager, device);
                } else if (warning == WARNING_CRITICAL) {
                        g_debug ("** EMIT: charge-critical");
                        engine_charge_critical (manager, device);
                } else if (warning == WARNING_ACTION) {
                        g_debug ("charge-action");
                        engine_charge_action (manager, device);
                }
                /* save new state */
                g_object_set_data (G_OBJECT(device), "engine-warning-old", GUINT_TO_POINTER(warning));
        }
}

static UpDevice *
engine_get_primary_device (CsdPowerManager *manager)
{
        guint i;
        UpDevice *device = NULL;
        UpDevice *device_tmp;
        UpDeviceKind kind;
        UpDeviceState state;
        gboolean is_present;

        for (i=0; i<manager->priv->devices_array->len; i++) {
                device_tmp = g_ptr_array_index (manager->priv->devices_array, i);

                /* get device properties */
                g_object_get (device_tmp,
                              "kind", &kind,
                              "state", &state,
                              "is-present", &is_present,
                              NULL);

                /* not present */
                if (!is_present)
                        continue;

                /* not discharging */
                if (state != UP_DEVICE_STATE_DISCHARGING)
                        continue;

                /* not battery */
                if (kind != UP_DEVICE_KIND_BATTERY)
                        continue;

                /* use composite device to cope with multiple batteries */
                device = g_object_ref (engine_get_composite_device (manager, device_tmp));
                break;
        }
        return device;
}

static void
phone_device_added_cb (GpmPhone *phone, guint idx, CsdPowerManager *manager)
{
        UpDevice *device;
        device = up_device_new ();

        g_debug ("phone added %i", idx);

        /* get device properties */
        g_object_set (device,
                      "kind", UP_DEVICE_KIND_PHONE,
                      "is-rechargeable", TRUE,
                      "native-path", g_strdup_printf ("dummy:phone_%i", idx),
                      "is-present", TRUE,
                      NULL);

        /* state changed */
        engine_device_add (manager, device);
        g_ptr_array_add (manager->priv->devices_array, g_object_ref (device));
}

static void
phone_device_removed_cb (GpmPhone *phone, guint idx, CsdPowerManager *manager)
{
        guint i;
        UpDevice *device;
        UpDeviceKind kind;

        g_debug ("phone removed %i", idx);

        for (i=0; i<manager->priv->devices_array->len; i++) {
                device = g_ptr_array_index (manager->priv->devices_array, i);

                /* get device properties */
                g_object_get (device,
                              "kind", &kind,
                              NULL);

                if (kind == UP_DEVICE_KIND_PHONE) {
                        g_ptr_array_remove_index (manager->priv->devices_array, i);
                        break;
                }
        }
}

static void
phone_device_refresh_cb (GpmPhone *phone, guint idx, CsdPowerManager *manager)
{
        guint i;
        UpDevice *device;
        UpDeviceKind kind;
        UpDeviceState state;
        gboolean is_present;
        gdouble percentage;

        g_debug ("phone refresh %i", idx);

        for (i=0; i<manager->priv->devices_array->len; i++) {
                device = g_ptr_array_index (manager->priv->devices_array, i);

                /* get device properties */
                g_object_get (device,
                              "kind", &kind,
                              "state", &state,
                              "percentage", &percentage,
                              "is-present", &is_present,
                              NULL);

                if (kind == UP_DEVICE_KIND_PHONE) {
                        is_present = gpm_phone_get_present (phone, idx);
                        state = gpm_phone_get_on_ac (phone, idx) ? UP_DEVICE_STATE_CHARGING : UP_DEVICE_STATE_DISCHARGING;
                        percentage = gpm_phone_get_percentage (phone, idx);
                        break;
                }
        }
}

static void
cdos_session_shutdown_cb (GObject *source_object,
                           GAsyncResult *res,
                           gpointer user_data)
{
        GVariant *result;
        GError *error = NULL;

        result = g_dbus_proxy_call_finish (G_DBUS_PROXY (source_object),
                                           res,
                                           &error);
        if (result == NULL) {
                g_warning ("couldn't shutdown using cdos-session: %s",
                           error->message);
                g_error_free (error);
        } else {
                g_variant_unref (result);
        }
}

// 关机
static void
cdos_session_shutdown (void)
{
        GError *error = NULL;
        GDBusProxy *proxy;

        /* ask cdos-session to show the shutdown dialog with a timeout */
        proxy = g_dbus_proxy_new_for_bus_sync (G_BUS_TYPE_SESSION,
                                               G_DBUS_PROXY_FLAGS_DO_NOT_LOAD_PROPERTIES,
                                               NULL,
                                               GNOME_SESSION_DBUS_NAME,
                                               GNOME_SESSION_DBUS_PATH,
                                               GNOME_SESSION_DBUS_INTERFACE,
                                               NULL, &error);
        if (proxy == NULL) {
                g_warning ("cannot connect to cdos-session: %s",
                           error->message);
                g_error_free (error);
                return;
        }
        g_dbus_proxy_call (proxy,
                           "Shutdown",
                           NULL,
                           G_DBUS_CALL_FLAGS_NONE,
                           -1, NULL,
                           cdos_session_shutdown_cb, NULL);
        g_object_unref (proxy);
}

// 响应电源事件
static void
do_power_action_type (CsdPowerManager *manager,
                      CsdPowerActionType action_type)
{
        gboolean ret;
        GError *error = NULL;
        g_message ("Do power action: %d(%s)", action_type, power_action_type_to_string (action_type));

        switch (action_type) {
        case CSD_POWER_ACTION_SUSPEND:
                csd_power_suspend (manager->priv->upower_proxy);
                break;
        case CSD_POWER_ACTION_INTERACTIVE:
                cdos_session_shutdown ();
                break;
        case CSD_POWER_ACTION_HIBERNATE:
                csd_power_hibernate (manager->priv->upower_proxy);
                break;
        case CSD_POWER_ACTION_SHUTDOWN:
                /* this is only used on critically low battery where
                 * hibernate is not available and is marginally better
                 * than just powering down the computer mid-write */
                csd_power_poweroff ();
                break;
        case CSD_POWER_ACTION_BLANK:
                /* //关闭屏幕
                if (manager->priv->x11_screen) {
                        ret = gnome_rr_screen_set_dpms_mode (manager->priv->x11_screen,
                                                        GNOME_RR_DPMS_OFF,
                                                        &error);
                        if (!ret) {
                                g_warning ("failed to turn the panel off for policy action: %s",
                                        error->message);
                                g_error_free (error);
                        }
                } */
                lock_screensaver (manager);
                break;
        case CSD_POWER_ACTION_NOTHING:
                break;
        }
}

static gboolean
upower_kbd_set_brightness (CsdPowerManager *manager, guint value, GError **error)
{
        GVariant *retval;

        /* same as before */
        if (manager->priv->kbd_brightness_now == value)
                return TRUE;

        /* update h/w value */
        retval = g_dbus_proxy_call_sync (manager->priv->upower_kdb_proxy,
                                         "SetBrightness",
                                         g_variant_new ("(i)", (gint) value),
                                         G_DBUS_CALL_FLAGS_NONE,
                                         -1,
                                         NULL,
                                         error);
        if (retval == NULL)
                return FALSE;

        /* save new value */
        manager->priv->kbd_brightness_now = value;
        g_variant_unref (retval);
        return TRUE;
}

static gboolean
upower_kbd_toggle (CsdPowerManager *manager,
                   GError **error)
{
        gboolean ret;

        if (manager->priv->kbd_brightness_old >= 0) {
                g_debug ("keyboard toggle off");
                ret = upower_kbd_set_brightness (manager,
                                                 manager->priv->kbd_brightness_old,
                                                 error);
                if (ret) {
                        /* succeeded, set to -1 since now no old value */
                        manager->priv->kbd_brightness_old = -1;
                }
        } else {
                g_debug ("keyboard toggle on");
                /* save the current value to restore later when untoggling */
                manager->priv->kbd_brightness_old = manager->priv->kbd_brightness_now;
                ret = upower_kbd_set_brightness (manager, 0, error);
                if (!ret) {
                        /* failed, reset back to -1 */
                        manager->priv->kbd_brightness_old = -1;
                }
        }

        return ret;
}

static void
do_lid_open_action (CsdPowerManager *manager)
{
        gboolean ret;
        GError *error = NULL;

        /* play a sound, using sounds from the naming spec */
        ca_context_play (manager->priv->canberra_context, 0,
                         CA_PROP_EVENT_ID, "lid-open",
                         /* TRANSLATORS: this is the sound description */
                         CA_PROP_EVENT_DESCRIPTION, _("Lid has been opened"),
                         NULL);
        g_message ("Lid state changed: opened");
        /* ensure we turn the panel back on after lid open */
        if (manager->priv->x11_screen) {
                ret = gnome_rr_screen_set_dpms_mode (manager->priv->x11_screen,
                                                GNOME_RR_DPMS_ON,
                                                &error);
                if (!ret) {
                        g_warning ("failed to turn the panel on after lid open: %s",
                                error->message);
                        g_clear_error (&error);
                }
        }
        /* only toggle keyboard if present and already toggled off */
        if (manager->priv->upower_kdb_proxy != NULL &&
            manager->priv->kbd_brightness_old != -1) {
                ret = upower_kbd_toggle (manager, &error);
                if (!ret) {
                        g_warning ("failed to turn the kbd backlight on: %s",
                                   error->message);
                        g_error_free (error);
                }
        }

        kill_lid_close_safety_timer (manager);
}

static gboolean
is_on (GnomeRROutput *output)
{
	GnomeRRCrtc *crtc;

	crtc = gnome_rr_output_get_crtc (output);
	if (!crtc)
		return FALSE;
	return gnome_rr_crtc_get_current_mode (crtc) != NULL;
}

static gboolean
non_laptop_outputs_are_all_off (GnomeRRScreen *screen)
{
        GnomeRROutput **outputs;
        int i;
        if (screen)
                return FALSE;

        outputs = gnome_rr_screen_list_outputs (screen);
        for (i = 0; outputs[i] != NULL; i++) {
                if (gnome_rr_output_is_builtin_display (outputs[i]))
                        continue;

                if (is_on (outputs[i]))
                        return FALSE;
        }

        return TRUE;
}

/* Timeout callback used to check conditions when the laptop's lid is closed but
 * the machine is not suspended yet.  We try to suspend again, so that the laptop
 * won't overheat if placed in a backpack.
 */
static gboolean
lid_close_safety_timer_cb (CsdPowerManager *manager)
{
        manager->priv->lid_close_safety_timer_id = 0;

        g_debug ("lid has been closed for a while; trying to suspend again");
        do_lid_closed_action (manager);

        return FALSE;
}

/* Sets up a timer to be triggered some seconds after closing the laptop lid
 * when the laptop is *not* suspended for some reason.  We'll check conditions
 * again in the timeout handler to see if we can suspend then.
 */
static void
setup_lid_close_safety_timer (CsdPowerManager *manager)
{
        if (manager->priv->lid_close_safety_timer_id != 0)
                return;

        manager->priv->lid_close_safety_timer_id = g_timeout_add_seconds (CSD_POWER_MANAGER_LID_CLOSE_SAFETY_TIMEOUT,
                                                                          (GSourceFunc) lid_close_safety_timer_cb,
                                                                          manager);
        g_source_set_name_by_id (manager->priv->lid_close_safety_timer_id, "[CsdPowerManager] lid close safety timer");
}

static void
kill_lid_close_safety_timer (CsdPowerManager *manager)
{
        if (manager->priv->lid_close_safety_timer_id != 0) {
                g_source_remove (manager->priv->lid_close_safety_timer_id);
                manager->priv->lid_close_safety_timer_id = 0;
        }
}

static void
suspend_with_lid_closed (CsdPowerManager *manager)
{
        gboolean ret;
        GError *error = NULL;
        CsdPowerActionType action_type;

        /* we have different settings depending on AC state */
        if (up_client_get_on_battery (manager->priv->up_client)) {
                action_type = g_settings_get_enum (manager->priv->settings,
                                                   "lid-close-battery-action");
        } else {
                action_type = g_settings_get_enum (manager->priv->settings,
                                                   "lid-close-ac-action");
        }
        g_message ("Supend with lid closed, action type: %d", action_type);
        /* check we won't melt when the lid is closed */
        if (action_type != CSD_POWER_ACTION_SUSPEND &&
            action_type != CSD_POWER_ACTION_HIBERNATE) {
#if ! UP_CHECK_VERSION(0,99,0)
                if (up_client_get_lid_force_sleep (manager->priv->up_client)) {
                        g_warning ("to prevent damage, now forcing suspend");
                        do_power_action_type (manager, CSD_POWER_ACTION_SUSPEND);
                        return;
                } else {
                        /* maybe lock the screen if the lid is closed */
                        if (action_type != CSD_POWER_ACTION_NOTHING)
                                lock_screensaver (manager);
                }
#else
                if (action_type != CSD_POWER_ACTION_NOTHING)
                        lock_screensaver (manager);
#endif
        }

        /* ensure we turn the panel back on after resume */
        if (manager->priv->x11_screen) {
                ret = gnome_rr_screen_set_dpms_mode (manager->priv->x11_screen,
                                                GNOME_RR_DPMS_OFF,
                                                &error);
                if (!ret) {
                        g_warning ("failed to turn the panel off after lid close: %s",
                                error->message);
                        g_clear_error (&error);
                }
        }
        /* only toggle keyboard if present and not already toggled */
        if (manager->priv->upower_kdb_proxy &&
            manager->priv->kbd_brightness_old == -1) {
                ret = upower_kbd_toggle (manager, &error);
                if (!ret) {
                        g_warning ("failed to turn the kbd backlight off: %s",
                                   error->message);
                        g_error_free (error);
                }
        }

        do_power_action_type (manager, action_type);
}

static void
do_lid_closed_action (CsdPowerManager *manager)
{
        /* play a sound, using sounds from the naming spec */
        ca_context_play (manager->priv->canberra_context, 0,
                         CA_PROP_EVENT_ID, "lid-close",
                         /* TRANSLATORS: this is the sound description */
                         CA_PROP_EVENT_DESCRIPTION, _("Lid has been closed"),
                         NULL);
        g_message ("Lid state changed: closed");
        if (manager->priv->x11_screen) {
                /* refresh RANDR so we get an accurate view of what monitors are plugged in when the lid is closed */
                gnome_rr_screen_refresh (manager->priv->x11_screen, NULL); /* NULL-GError */

                /* perform policy action */
                if (g_settings_get_boolean (manager->priv->settings, "lid-close-suspend-with-external-monitor")
                || non_laptop_outputs_are_all_off (manager->priv->x11_screen)) {
                        g_message ("lid is closed; suspending or hibernating");
                        suspend_with_lid_closed (manager);
                } else {
                        g_message ("lid is closed; not suspending nor hibernating since some external monitor outputs are still active");
                        setup_lid_close_safety_timer (manager);
                }
        }
}


static void
#if UP_CHECK_VERSION(0,99,0)
lid_state_changed_cb (UpClient *client, GParamSpec *pspec, CsdPowerManager *manager)
#else
up_client_changed_cb (UpClient *client, CsdPowerManager *manager)
#endif
{
        gboolean tmp;

        if (!up_client_get_on_battery (client)) {
            /* if we are playing a critical charge sound loop on AC, stop it */
            if (manager->priv->critical_alert_timeout_id > 0) {
                 g_debug ("stopping alert loop due to ac being present");
                 play_loop_stop (manager);
            }
            notify_close_if_showing (manager->priv->notification_low);
        }

        /* same state */
        tmp = up_client_get_lid_is_closed (manager->priv->up_client);
        if (manager->priv->lid_is_closed == tmp)
                return;
        manager->priv->lid_is_closed = tmp;

        /* fake a keypress */
        if (tmp)
                do_lid_closed_action (manager);
        else
                do_lid_open_action (manager);
}

typedef enum {
        SESSION_STATUS_CODE_AVAILABLE = 0,
        SESSION_STATUS_CODE_INVISIBLE,
        SESSION_STATUS_CODE_BUSY,
        SESSION_STATUS_CODE_IDLE,
        SESSION_STATUS_CODE_UNKNOWN
} SessionStatusCode;

typedef enum {
        SESSION_INHIBIT_MASK_LOGOUT = 1,
        SESSION_INHIBIT_MASK_SWITCH = 2,
        SESSION_INHIBIT_MASK_SUSPEND = 4,
        SESSION_INHIBIT_MASK_IDLE = 8
} SessionInhibitMask;

static const gchar *
idle_mode_to_string (CsdPowerIdleMode mode)
{
        if (mode == CSD_POWER_IDLE_MODE_NORMAL)
                return "normal";
        if (mode == CSD_POWER_IDLE_MODE_DIM)
                return "dim";
        if (mode == CSD_POWER_IDLE_MODE_BLANK)
                return "blank";
        if (mode == CSD_POWER_IDLE_MODE_SLEEP)
                return "sleep";
        if (mode == CSD_POWER_IDLE_MODE_SCREENSAVER)
                return "screensaver";
        return "unknown";
}

static GnomeRROutput *
get_primary_output (CsdPowerManager *manager)
{
        GnomeRROutput *output = NULL;
        GnomeRROutput **outputs;
        guint i;
        if (!manager->priv->x11_screen)
                goto out;

        /* search all X11 outputs for the device id */
        outputs = gnome_rr_screen_list_outputs (manager->priv->x11_screen);
        if (outputs == NULL)
                goto out;

        for (i = 0; outputs[i] != NULL; i++) {
        /*        if (gnome_rr_output_info_is_connected (outputs[i]) &&
                    gnome_rr_output_is_builtin_display (outputs[i]) &&
                    gnome_rr_output_get_backlight_min (outputs[i]) >= 0 &&
                    gnome_rr_output_get_backlight_max (outputs[i]) > 0) {
                        output = outputs[i];*/
		if (gnome_rr_output_get_is_primary (outputs[i])){
                        break;
                }
        }
out:
        return output;
}

static void
backlight_override_settings_refresh (CsdPowerManager *manager)
{
        int i = 0;
        /* update all the stored backlight override properties
         * this is called on startup and by engine_settings_key_changed_cb */
        manager->priv->backlight_helper_force = g_settings_get_boolean
                        (manager->priv->settings, "backlight-helper-force");

        /* concatenate all the search preferences into a single argument string */
        gchar** backlight_preference_order = g_settings_get_strv
                (manager->priv->settings, "backlight-helper-preference-order");

        gchar* tmp1 = NULL;
        gchar* tmp2 = NULL;

        if (backlight_preference_order[0] != NULL) {
                tmp1 = g_strdup_printf("-b %s", backlight_preference_order[0]);
        }

        for (i=1; backlight_preference_order[i] != NULL; i++ )
        {
                tmp2 = tmp1;
                tmp1 = g_strdup_printf("%s -b %s", tmp2,
                                backlight_preference_order[i]);
                g_free(tmp2);
        }

        tmp2 = manager->priv->backlight_helper_preference_args;
        manager->priv->backlight_helper_preference_args = tmp1;
        g_free(tmp2);
        tmp2 = NULL;

        g_free(backlight_preference_order);
        backlight_preference_order = NULL;
}

/**
 * backlight_helper_get_value:
 *
 * Gets a brightness value from the PolicyKit helper.
 *
 * Return value: the signed integer value from the helper, or -1
 * for failure. If -1 then @error is set.
 **/
static gint64
backlight_helper_get_value (const gchar *argument, CsdPowerManager* manager,
                GError **error)
{
        gboolean ret;
        gchar *stdout_data = NULL;
        gint exit_status = 0;
        gint64 value = -1;
        gchar *command = NULL;
        gchar *endptr = NULL;

#ifndef __linux__
        /* non-Linux platforms won't have /sys/class/backlight */
        g_set_error_literal (error,
                             CSD_POWER_MANAGER_ERROR,
                             CSD_POWER_MANAGER_ERROR_FAILED,
                             "The sysfs backlight helper is only for Linux");
        goto out;
#endif

        /* get the data */
        command = g_strdup_printf (LIBEXECDIR "/csd-backlight-helper --%s %s",
                                   argument,
                                   manager->priv->backlight_helper_preference_args);
        ret = g_spawn_command_line_sync (command,
                                         &stdout_data,
                                         NULL,
                                         &exit_status,
                                         error);
        g_debug ("executed %s retval: %i", command, exit_status);

        if (!ret)
                goto out;

        if (WEXITSTATUS (exit_status) != 0) {
                 g_set_error (error,
                             CSD_POWER_MANAGER_ERROR,
                             CSD_POWER_MANAGER_ERROR_FAILED,
                             "csd-backlight-helper failed: %s",
                             stdout_data ? stdout_data : "No reason");
                goto out;
        }

        /* parse */
        value = g_ascii_strtoll (stdout_data, &endptr, 10);

        /* parsing error */
        if (endptr == stdout_data) {
                value = -1;
                g_set_error (error,
                             CSD_POWER_MANAGER_ERROR,
                             CSD_POWER_MANAGER_ERROR_FAILED,
                             "failed to parse value: %s",
                             stdout_data);
                goto out;
        }

        /* out of range */
        if (value > G_MAXINT) {
                value = -1;
                g_set_error (error,
                             CSD_POWER_MANAGER_ERROR,
                             CSD_POWER_MANAGER_ERROR_FAILED,
                             "value out of range: %s",
                             stdout_data);
                goto out;
        }

        /* Fetching the value failed, for some other reason */
        if (value < 0) {
                g_set_error (error,
                             CSD_POWER_MANAGER_ERROR,
                             CSD_POWER_MANAGER_ERROR_FAILED,
                             "value negative, but helper did not fail: %s",
                             stdout_data);
                goto out;
        }

out:
        g_free (command);
        g_free (stdout_data);
        return value;
}

/**
 * backlight_helper_set_value:
 *
 * Sets a brightness value using the PolicyKit helper.
 *
 * Return value: Success. If FALSE then @error is set.
 **/
static gboolean
backlight_helper_set_value (const gchar *argument,
                            gint value,
                            CsdPowerManager* manager,
                            GError **error)
{
        gboolean ret;
        gint exit_status = 0;
        gchar *command = NULL;

#ifndef __linux__
        /* non-Linux platforms won't have /sys/class/backlight */
        g_set_error_literal (error,
                             CSD_POWER_MANAGER_ERROR,
                             CSD_POWER_MANAGER_ERROR_FAILED,
                             "The sysfs backlight helper is only for Linux");
        goto out;
#endif

        /* get the data */
        command = g_strdup_printf ("pkexec " LIBEXECDIR "/csd-backlight-helper --%s %i %s",
                                   argument, value,
                                   manager->priv->backlight_helper_preference_args);
        ret = g_spawn_command_line_sync (command,
                                         NULL,
                                         NULL,
                                         &exit_status,
                                         error);

        g_debug ("executed %s retval: %i", command, exit_status);

        if (!ret || WEXITSTATUS (exit_status) != 0)
                goto out;

out:
        g_free (command);
        return ret;
}

static void
backlight_emit_changed (CsdPowerManager *manager)
{
        gboolean ret;
        GError *error = NULL;

        /* not yet connected to the bus */
        if (manager->priv->connection == NULL)
                return;
        ret = g_dbus_connection_emit_signal (manager->priv->connection,
                                             CSD_DBUS_SERVICE,
                                             CSD_POWER_DBUS_PATH,
                                             CSD_POWER_DBUS_INTERFACE_SCREEN,
                                             "Changed",
                                             NULL,
                                             &error);
        if (!ret) {
                g_warning ("failed to emit Changed: %s", error->message);
                g_error_free (error);
        }
}

static gboolean
display_backlight_dim (CsdPowerManager *manager,
                       gint idle_percentage,
                       GError **error)
{
        gint now;
        gboolean ret = FALSE;

        now = backlight_get_abs (error);
        if (now < 0) {
                goto out;
        }

        /* is the dim brightness actually *dimmer* than the
         * brightness we have now? */
        /* save for undim */
        manager->priv->pre_dim_brightness = now;

out:
        return ret;
}

static gboolean
kbd_backlight_dim (CsdPowerManager *manager,
                   gint idle_percentage,
                   GError **error)
{
        gboolean ret;
        gint idle;
        gint max;
        gint now;

        if (manager->priv->upower_kdb_proxy == NULL)
                return TRUE;

        now = manager->priv->kbd_brightness_now;
        max = manager->priv->kbd_brightness_max;
        idle = PERCENTAGE_TO_ABS (0, max, idle_percentage);
        if (idle > now) {
                g_debug ("kbd brightness already now %i/%i, so "
                         "ignoring dim to %i/%i",
                         now, max, idle, max);
                return TRUE;
        }
        ret = upower_kbd_set_brightness (manager, idle, error);
        if (!ret)
                return FALSE;

        /* save for undim */
        manager->priv->kbd_brightness_pre_dim = now;
        return TRUE;
}

static void
idle_set_mode (CsdPowerManager *manager, CsdPowerIdleMode mode)
{
        gboolean ret = FALSE;
        GError *error = NULL;
        gint idle_percentage;
        CsdPowerActionType action_type;
        CdosSettingsSessionState state;

        if (mode == manager->priv->current_idle_mode)
                return;

        // 除非当前idle mode是CSD_POWER_IDLE_MODE_SCREENSAVER，否则忽略设置更小的idle mode
        if (mode < manager->priv->current_idle_mode &&
            mode != CSD_POWER_IDLE_MODE_NORMAL &&
            manager->priv->current_idle_mode != CSD_POWER_IDLE_MODE_SCREENSAVER)
                return;

        /* ensure we're still on an active console */
        state = cdos_settings_session_get_state (manager->priv->session);
        if (state == CDOS_SETTINGS_SESSION_STATE_INACTIVE) {
                g_debug ("ignoring state transition to %s as inactive",
                         idle_mode_to_string (mode));
                return;
        }

        manager->priv->current_idle_mode = mode;
        g_message ("Idle set new mode: %d(%s)", mode, idle_mode_to_string (mode));

        /* don't do any power saving if we're a VM */
        if (manager->priv->is_virtual_machine
        && mode != CSD_POWER_IDLE_MODE_SCREENSAVER) {
                g_debug ("ignoring state transition to %s as virtual machine",
                         idle_mode_to_string (mode));
                return;
        }
        /* gnome-settings-daemon:
         * 如果正在进入空闲模式，添加一个监听事件，
         * 当用户变为活跃状态，返回正常模式 */
        if (mode != CSD_POWER_IDLE_MODE_NORMAL) {
            gnome_idle_monitor_add_user_active_watch (manager->priv->idle_monitor,
                                                      idle_became_active_cb,
                                                      manager,
                                                      NULL);
        }

        /* save current brightness, and set dim level */
        if (mode == CSD_POWER_IDLE_MODE_DIM) {

                /* have we disabled the action */
                if (up_client_get_on_battery (manager->priv->up_client)) {
                        ret = g_settings_get_boolean (manager->priv->settings,
                                                      "idle-dim-battery");
                } else {
                        ret = g_settings_get_boolean (manager->priv->settings,
                                                      "idle-dim-ac");
                }
                if (!ret) {
                        g_debug ("not dimming due to policy");
                        return;
                }

                /* display backlight */
                idle_percentage = g_settings_get_int (manager->priv->settings,
                                                      "idle-brightness");
                /*
		 * 屏幕背光？ 没有 backlight min 和 max 或者相似的接口 FIXME
		 * ret = display_backlight_dim (manager, idle_percentage, &error);
                if (!ret) {
                        g_warning ("failed to set dim backlight to %i%%: %s",
                                   idle_percentage,
                                   error->message);
                        g_clear_error (&error);
                }*/

                /* keyboard backlight */
                ret = kbd_backlight_dim (manager, idle_percentage, &error);
                if (!ret) {
                        g_warning ("failed to set dim kbd backlight to %i%%: %s",
                                   idle_percentage,
                                   error->message);
                        g_clear_error (&error);
                }

        /* turn off screen and kbd */
        } else if (mode == CSD_POWER_IDLE_MODE_BLANK) {
                if (manager->priv->x11_screen) {
                        ret = gnome_rr_screen_set_dpms_mode (manager->priv->x11_screen,
                                                        GNOME_RR_DPMS_OFF,
                                                        &error);
                        if (!ret) {
                                g_warning ("failed to turn the panel off: %s",
                                        error->message);
                                g_clear_error (&error);
                        }
                }

                /* only toggle keyboard if present and not already toggled */
                if (manager->priv->upower_kdb_proxy &&
                    manager->priv->kbd_brightness_old == -1) {
                        ret = upower_kbd_toggle (manager, &error);
                        if (!ret) {
                                g_warning ("failed to turn the kbd backlight off: %s",
                                           error->message);
                                g_error_free (error);
                        }
                }

        /* sleep */
        } else if (mode == CSD_POWER_IDLE_MODE_SLEEP) {

                if (up_client_get_on_battery (manager->priv->up_client)) {
                        action_type = g_settings_get_enum (manager->priv->settings,
                                                           "sleep-inactive-battery-type");
                } else {
                        action_type = g_settings_get_enum (manager->priv->settings,
                                                           "sleep-inactive-ac-type");
                }
                do_power_action_type (manager, action_type);

        // 屏幕保护模式
        } else if (mode == CSD_POWER_IDLE_MODE_SCREENSAVER) {
                if (manager->priv->screensaver_proxy) {
                        g_dbus_proxy_call_sync (manager->priv->screensaver_proxy,
                                                                "ProtectScreen",
                                                                NULL,
                                                                G_DBUS_CALL_FLAGS_NONE,
                                                                -1,
                                                                NULL, NULL);
                }
                if (g_settings_get_boolean (manager->priv->settings, "lock-on-screensaver"))
                        lock_screensaver (manager);

        /* turn on screen and restore user-selected brightness level */
        } else if (mode == CSD_POWER_IDLE_MODE_NORMAL) {
                if (manager->priv->x11_screen) {
                        ret = gnome_rr_screen_set_dpms_mode (manager->priv->x11_screen,
                                                        GNOME_RR_DPMS_ON,
                                                        &error);
                        if (!ret) {
                                g_warning ("failed to turn the panel on: %s",
                                        error->message);
                                g_clear_error (&error);
                        }
                }
                /* reset brightness if we dimmed */
                /*if (manager->priv->pre_dim_brightness >= 0) {
                        ret = backlight_set_abs (manager,
                                                 manager->priv->pre_dim_brightness,
                                                 FALSE,
                                                 &error);
                        if (!ret) {
                                g_warning ("failed to restore backlight to %i: %s",
                                           manager->priv->pre_dim_brightness,
                                           error->message);
                                g_clear_error (&error);
                        } else {
                                manager->priv->pre_dim_brightness = -1;
                        }
                }*/

                /* only toggle keyboard if present and already toggled off */
                if (manager->priv->upower_kdb_proxy &&
                    manager->priv->kbd_brightness_old != -1) {
                        ret = upower_kbd_toggle (manager, &error);
                        if (!ret) {
                                g_warning ("failed to turn the kbd backlight on: %s",
                                           error->message);
                                g_clear_error (&error);
                        }
                }

                /* reset kbd brightness if we dimmed */
                if (manager->priv->kbd_brightness_pre_dim >= 0) {
                        ret = upower_kbd_set_brightness (manager,
                                                         manager->priv->kbd_brightness_pre_dim,
                                                         &error);
                        if (!ret) {
                                g_warning ("failed to restore kbd backlight to %i: %s",
                                           manager->priv->kbd_brightness_pre_dim,
                                           error->message);
                                g_error_free (error);
                        }
                        manager->priv->kbd_brightness_pre_dim = -1;
                }

        }
}

static gboolean
idle_is_session_idle (CsdPowerManager *manager)
{
        gboolean ret;
        GVariant *result;
        guint status;

        /* not yet connected to cdos-session */
        if (manager->priv->session_presence_proxy == NULL) {
                g_warning ("session idleness not available, cdos-session is not available");
                return FALSE;
        }

        /* get the session status */
        result = g_dbus_proxy_get_cached_property (manager->priv->session_presence_proxy,
                                                   "status");
        if (result == NULL) {
                g_warning ("no readable status property on %s",
                           g_dbus_proxy_get_interface_name (manager->priv->session_presence_proxy));
                return FALSE;
        }

        g_variant_get (result, "u", &status);
        ret = (status == SESSION_STATUS_CODE_IDLE);
        g_variant_unref (result);

        return ret;
}

static void
idle_became_active_cb (GnomeIdleMonitor *idle_monitor,
                       guint             watch_id,
                       gpointer          user_data)
{
    CsdPowerManager *manager = CSD_POWER_MANAGER (user_data);

    set_temporary_unidle_on_ac (manager, FALSE);
    //notify_close_if_showing (manager->priv->notification_sleep_warning);

    idle_set_mode (manager, CSD_POWER_IDLE_MODE_NORMAL);
}

static gboolean
idle_is_session_inhibited (CsdPowerManager *manager, guint mask)
{
        gboolean ret;
        GVariant *retval = NULL;
        GError *error = NULL;

        /* not yet connected to cdos-session */
        if (manager->priv->session_proxy == NULL) {
                g_warning ("session inhibition not available, cdos-session is not available");
                return FALSE;
        }

        retval = g_dbus_proxy_call_sync (manager->priv->session_proxy,
                                         "IsInhibited",
                                         g_variant_new ("(u)",
                                                        mask),
                                         G_DBUS_CALL_FLAGS_NONE,
                                         -1, NULL,
                                         &error);
        if (retval == NULL) {
                /* abort as the DBUS method failed */
                g_warning ("IsInhibited failed: %s", error->message);
                g_error_free (error);
                return FALSE;
        }

        g_variant_get (retval, "(b)", &ret);
        g_variant_unref (retval);

        return ret;
}

/**
 *  idle_adjust_timeout:
 *  @idle_time: Current idle time, in seconds.
 *  @timeout: The new timeout we want to set, in seconds.
 *
 *  On slow machines, or machines that have lots to load duing login,
 *  the current idle time could be bigger than the requested timeout.
 *  In this case the scheduled idle timeout will never fire, unless
 *  some user activity (keyboard, mouse) resets the current idle time.
 *  Instead of relying on user activity to correct this issue, we need
 *  to adjust timeout, as related to current idle time, so the idle
 *  timeout will fire as designed.
 *
 *  Return value: timeout to set, adjusted acccording to current idle time.
 **/
static guint
idle_adjust_timeout (guint idle_time, guint timeout)
{
        /* allow 2 sec margin for messaging delay. */
        idle_time += 2;

        /* Double timeout until it's larger than current idle time.
         * Give up for ultra slow machines. (86400 sec = 24 hours) */
        while (timeout < idle_time &&
               timeout < 86400 &&
               timeout > 0) {
                timeout *= 2;
        }
        return timeout;
}

/**
 * idle_adjust_timeout_blank:
 * @idle_time: current idle time, in seconds.
 * @timeout: the new timeout we want to set, in seconds.
 *
 * Same as idle_adjust_timeout(), but also accounts for the duration
 * of the fading animation in the screensaver (so that blanking happens
 * exactly at the end of it, if configured with the same timeouts)
 */
static guint
idle_adjust_timeout_blank (guint idle_time, guint timeout)
{
        return idle_adjust_timeout (idle_time,
                                    timeout + SCREENSAVER_FADE_TIME);
}

static void
idle_configure (CsdPowerManager *manager)
{
        gboolean is_idle_inhibited;
        guint current_idle_time;
        guint timeout_blank;
        guint timeout_sleep;
        guint timeout_screensaver;
        gboolean on_battery;

        /* are we inhibited from going idle */
        is_idle_inhibited = idle_is_session_inhibited (manager,
                                                       SESSION_INHIBIT_MASK_IDLE);
        if (is_idle_inhibited) {
                g_debug ("inhibited, so using normal state");
                idle_set_mode (manager, CSD_POWER_IDLE_MODE_NORMAL);

                gpm_idletime_alarm_remove (manager->priv->idletime,
                                           CSD_POWER_IDLETIME_BLANK_ID);
                gpm_idletime_alarm_remove (manager->priv->idletime,
                                           CSD_POWER_IDLETIME_SLEEP_ID);
                gpm_idletime_alarm_remove (manager->priv->idletime,
                                           CSD_POWER_IDLETIME_LOCK_ID);
                gpm_idletime_alarm_remove (manager->priv->idletime,
                                           CSD_POWER_IDLETIME_SCREENSAVER_ID);
                return;
        }

        current_idle_time = gpm_idletime_get_time (manager->priv->idletime) / 1000;

        /* set up blank callback even when session is not idle,
         * but only if we actually want to blank. */
        on_battery = up_client_get_on_battery (manager->priv->up_client);
        if (on_battery) {
                // 使用电池空闲超时息屏时间
                timeout_blank = g_settings_get_int (manager->priv->settings,
                                                    "sleep-display-battery");
        } else {
                // 使用电源空闲超时息屏时间
                timeout_blank = g_settings_get_int (manager->priv->settings,
                                                    "sleep-display-ac");
        }
        if (timeout_blank != 0) {
                // 设置超时息屏
                gpm_idletime_alarm_set (manager->priv->idletime,
                                        CSD_POWER_IDLETIME_BLANK_ID,
                                        idle_adjust_timeout_blank (current_idle_time, timeout_blank) * 1000);
        } else {
                gpm_idletime_alarm_remove (manager->priv->idletime,
                                           CSD_POWER_IDLETIME_BLANK_ID);
        }

        /* only do the sleep timeout when the session is idle
         * and we aren't inhibited from sleeping */
        if (on_battery) {
                timeout_sleep = g_settings_get_int (manager->priv->settings,
                                                    "sleep-inactive-battery-timeout");
        } else {
                timeout_sleep = g_settings_get_int (manager->priv->settings,
                                                    "sleep-inactive-ac-timeout");
        }
        if (timeout_sleep != 0) {
                g_debug ("setting up sleep callback %is", timeout_sleep);

                gpm_idletime_alarm_set (manager->priv->idletime,
                                        CSD_POWER_IDLETIME_SLEEP_ID,
                                        idle_adjust_timeout (current_idle_time, timeout_sleep) * 1000);
        } else {
                gpm_idletime_alarm_remove (manager->priv->idletime,
                                           CSD_POWER_IDLETIME_SLEEP_ID);
        }

        // 空闲超时显示屏保
        timeout_screensaver = g_settings_get_uint (manager->priv->settings,
                                        "idle-screensaver-timeout");
        if (timeout_screensaver != 0) {
                g_debug ("setting up screensaver callback %is", timeout_screensaver);

                gpm_idletime_alarm_set (manager->priv->idletime,
                                        CSD_POWER_IDLETIME_SCREENSAVER_ID,
                                        idle_adjust_timeout (current_idle_time, timeout_screensaver) * 1000);
        } else {
                gpm_idletime_alarm_remove (manager->priv->idletime,
                                           CSD_POWER_IDLETIME_SCREENSAVER_ID);
        }
}

/**
 * @timeout: The new timeout we want to set, in seconds
 **/
static gboolean
idle_set_timeout_dim (CsdPowerManager *manager, guint timeout)
{
        guint idle_time;

        idle_time = gpm_idletime_get_time (manager->priv->idletime) / 1000;
        if (idle_time == 0)
                return FALSE;

        g_debug ("Setting dim idle timeout: %ds", timeout);
        if (timeout > 0) {
                gpm_idletime_alarm_set (manager->priv->idletime,
                                        CSD_POWER_IDLETIME_DIM_ID,
                                        idle_adjust_timeout (idle_time, timeout) * 1000);
        } else {
                gpm_idletime_alarm_remove (manager->priv->idletime,
                                           CSD_POWER_IDLETIME_DIM_ID);
        }
        return TRUE;
}

static void
refresh_idle_dim_settings (CsdPowerManager *manager)
{
        gint timeout_dim;
        timeout_dim = g_settings_get_int (manager->priv->settings,
                                          "idle-dim-time");
        g_debug ("idle dim set with timeout %i", timeout_dim);
        idle_set_timeout_dim (manager, timeout_dim);
}

static void
csd_power_manager_class_init (CsdPowerManagerClass *klass)
{
        GObjectClass   *object_class = G_OBJECT_CLASS (klass);

        object_class->finalize = csd_power_manager_finalize;
}

static void
idle_dbus_signal_cb (GDBusProxy *proxy,
                     const gchar *sender_name,
                     const gchar *signal_name,
                     GVariant *parameters,
                     gpointer user_data)
{
        CsdPowerManager *manager = CSD_POWER_MANAGER (user_data);

        if (g_strcmp0 (signal_name, "InhibitorAdded") == 0 ||
            g_strcmp0 (signal_name, "InhibitorRemoved") == 0) {
                g_debug ("Received gnome session inhibitor change");
                idle_configure (manager);
        }
        if (g_strcmp0 (signal_name, "StatusChanged") == 0) {
                guint status;

                g_variant_get (parameters, "(u)", &status);
                g_dbus_proxy_set_cached_property (proxy, "status",
                                                  g_variant_new ("u", status));
                g_debug ("Received gnome session status change");
                idle_configure (manager);
        }
}

static void
session_proxy_ready_cb (GObject *source_object,
                        GAsyncResult *res,
                        gpointer user_data)
{
        GError *error = NULL;
        CsdPowerManager *manager = CSD_POWER_MANAGER (user_data);

        manager->priv->session_proxy = g_dbus_proxy_new_for_bus_finish (res, &error);
        if (manager->priv->session_proxy == NULL) {
                g_warning ("Could not connect to cdos-session: %s",
                           error->message);
                g_error_free (error);
        } else {
                g_signal_connect (manager->priv->session_proxy, "g-signal",
                                  G_CALLBACK (idle_dbus_signal_cb), manager);
        }

        idle_configure (manager);
}

static void
session_presence_proxy_ready_cb (GObject *source_object,
                                 GAsyncResult *res,
                                 gpointer user_data)
{
        GError *error = NULL;
        CsdPowerManager *manager = CSD_POWER_MANAGER (user_data);

        manager->priv->session_presence_proxy = g_dbus_proxy_new_for_bus_finish (res, &error);
        if (manager->priv->session_presence_proxy == NULL) {
                g_warning ("Could not connect to gnome-sesson: %s",
                           error->message);
                g_error_free (error);
                return;
        }
        g_signal_connect (manager->priv->session_presence_proxy, "g-signal",
                          G_CALLBACK (idle_dbus_signal_cb), manager);
}

static gboolean
temporary_unidle_done_cb (CsdPowerManager *manager)
{
    idle_set_mode (manager, manager->priv->previous_idle_mode);
    manager->priv->temporary_unidle_on_ac_id = 0;
    return FALSE;
}

static void
set_temporary_unidle_on_ac (CsdPowerManager *manager, gboolean enable)
{
    if (!enable)
    {
        if (manager->priv->temporary_unidle_on_ac_id != 0)
        {
            g_source_remove (manager->priv->temporary_unidle_on_ac_id);
            manager->priv->temporary_unidle_on_ac_id = 0;
        }
    }
    else
    {
        if (manager->priv->temporary_unidle_on_ac_id != 0)
        {
            g_source_remove (manager->priv->temporary_unidle_on_ac_id);
        }
        else
        {
            manager->priv->previous_idle_mode = manager->priv->current_idle_mode;
            idle_set_mode (manager, CSD_POWER_IDLE_MODE_NORMAL);
        }
        manager->priv->temporary_unidle_on_ac_id =
            g_timeout_add_seconds (15 /*POWER_UP_TIME_ON_AC*/,
                                   (GSourceFunc)temporary_unidle_done_cb,
                                   manager);
        g_source_set_name_by_id (manager->priv->temporary_unidle_on_ac_id, "[found-settings-daemon] temporary_unidle_done_cb");
    }
}

static void
handle_wake_up_screen (CsdPowerManager *manager)
{
    set_temporary_unidle_on_ac (manager, TRUE);
}

static void
screensaver_signal_cb (GDBusProxy *proxy,
                       const gchar *sender_name,
                       const gchar *signal_name,
                       GVariant *parameters,
                       gpointer user_data)
{
    if (g_strcmp0 (signal_name, "WakeUpScreen") == 0)
    {
        handle_wake_up_screen (CSD_POWER_MANAGER (user_data));
    }
}

static void
screensaver_proxy_ready_cb (GObject *source_object,
                            GAsyncResult *res,
                            gpointer user_data)
{
    GError *error = NULL;
    CsdPowerManager *manager = CSD_POWER_MANAGER (user_data);

    manager->priv->screensaver_proxy = g_dbus_proxy_new_for_bus_finish (res, &error);
    if (manager->priv->screensaver_proxy == NULL)
    {
        g_warning ("Could not connect to screensaver: %s",
                   error->message);
        g_error_free (error);
        return;
    }
    g_signal_connect (manager->priv->screensaver_proxy, "g-signal",
                      G_CALLBACK (screensaver_signal_cb), manager);
}
/*
热插行为退出节能模式条件:
        1 当前还是节能模式
        2 低电量自动进入节能模式开关打开
        3 记录自动进入低电量模式前的电源模式
*/
static void 
hot_plugging(CsdPowerManager *manager)
{
        CsdPowerRunMode run_mode = CSD_POWER_RUN_MODE_NONE;
        gboolean low_power_auto_save = FALSE;

        run_mode = g_settings_get_enum(manager->priv->settings, "run-mode");
        low_power_auto_save = g_settings_get_boolean(manager->priv->settings, "percentage-low-save-mode");

        // 处理低电量自动进入节能模式的情况
        if (run_mode ==  CSD_POWER_RUN_MODE_POWERSAVE 
        && low_power_auto_save
        && manager->priv->run_mode_bp != CSD_POWER_RUN_MODE_NONE) 
        {
                g_debug("handle hot plugging event");
                set_run_mode(manager, manager->priv->run_mode_bp, FALSE);
        }
}

/*
拔行为结果:
        1保持当前模式
        2 恢复节能模式
*/
static void 
hot_drawing(CsdPowerManager *manager)
{
        GVariant *percentage;
        gint percent;
        gint low_threshold;

        percentage = g_dbus_proxy_get_cached_property (manager->priv->upower_device_proxy, "Percentage");
        percent = (int)g_variant_get_double (percentage);
        low_threshold = g_settings_get_int(manager->priv->settings, "percentage-low-threshold");

        g_debug("handle hot drawing event: %d", percent > low_threshold);
        if (percent > low_threshold) {
                // 1保持当前模式不变
        } else {
                // 2 恢复节能模式
                set_run_mode(manager, CSD_POWER_RUN_MODE_POWERSAVE, FALSE);
        }
}

 // 状态属性改变
static void
upower_prop_state_changed(CsdPowerManager *manager, const gchar *prop)
{
        GVariant *power_state = g_dbus_proxy_get_cached_property (manager->priv->upower_device_proxy, "State");
        uint state = 1;

        if (!power_state)
                return;

        state = g_variant_get_uint32 (power_state);

        // 热插拔处理
        if (state == 1) {
                hot_plugging(manager);
        } else if (state == 2 || state == 3) {
                hot_drawing(manager);
        }

        guint timeout_blank = 0;
        guint current_idle_time = 0;
        guint timeout_sleep;

        current_idle_time = gpm_idletime_get_time (manager->priv->idletime) / 1000;
	if (state == 1){
		timeout_blank = g_settings_get_int (manager->priv->settings, "sleep-display-ac");
		timeout_sleep = g_settings_get_int (manager->priv->settings,
					    "sleep-inactive-ac-timeout");
	} else {
		timeout_blank = g_settings_get_int (manager->priv->settings, "sleep-display-battery");
		timeout_sleep = g_settings_get_int (manager->priv->settings,
					    "sleep-inactive-battery-timeout");
	}

	 if (timeout_blank != 0){
		gpm_idletime_alarm_set (manager->priv->idletime, 
				CSD_POWER_IDLETIME_BLANK_ID,
				idle_adjust_timeout_blank (current_idle_time, timeout_blank) * 1000);
	 }else{
		gpm_idletime_alarm_remove (manager->priv->idletime,
				   CSD_POWER_IDLETIME_BLANK_ID);
	}

	if(timeout_sleep != 0) {
		gpm_idletime_alarm_set (manager->priv->idletime,
				CSD_POWER_IDLETIME_SLEEP_ID,
				idle_adjust_timeout (current_idle_time, timeout_sleep) * 1000);
	} else {
		gpm_idletime_alarm_remove (manager->priv->idletime,
				   CSD_POWER_IDLETIME_SLEEP_ID);
	}

        if (power_state)
                g_variant_unref(power_state);
}

// 电量百分比变化
static void
upower_prop_percentage_changed(CsdPowerManager *manager, GVariant    *changed_properties)
{
        GVariant *power_state;
        GVariant *percentage;
        uint state;
        gint percent;
        gint low_threshold;
        gboolean need_notify = TRUE;
        gboolean low_power_auto_save = TRUE;
        CsdPowerRunMode run_mode = CSD_POWER_RUN_MODE_NONE;

        power_state = g_dbus_proxy_get_cached_property (manager->priv->upower_device_proxy, "State");                           // power state: 1 充电  2 放电 3 电池电量放空 4 电池电量充满      
        state = g_variant_get_uint32 (power_state);
        run_mode = g_settings_get_enum(manager->priv->settings, "run-mode");
        
        if (state != 2 )
        {
                // 非放电状态，清空通知记录
                for (int i = 0; i< 3; i++) {
                        manager->priv->notify_low_num[i] = 0;
                }
                return;
        }

        percentage = g_dbus_proxy_get_cached_property (manager->priv->upower_device_proxy, "Percentage");
        percent = (int)g_variant_get_double (percentage);
        low_threshold = g_settings_get_int(manager->priv->settings, "percentage-low-threshold");
        if (percent > low_threshold)
        {
                // 电量在阈值之上
                return;
        }

        need_notify = g_settings_get_boolean (manager->priv->settings, "percentage-low-notify");
        low_power_auto_save = g_settings_get_boolean(manager->priv->settings, "percentage-low-save-mode");

        // 低电量自动进入节能模式
        if (low_power_auto_save && run_mode != CSD_POWER_RUN_MODE_POWERSAVE)
        {
		g_debug("auto set powersave run-mode");
                set_run_mode(manager, CSD_POWER_RUN_MODE_POWERSAVE, FALSE);
        }

        if (!need_notify) {
                // 免通知
                return;     
        }

        gchar *msg = NULL;
        gchar *action = NULL;
        gchar *icon_name = NULL;

        if (percent <= low_threshold && percent > 10 && manager->priv->notify_low_num[0] == 0) {
                msg = _("The battery is low, please connect it to the power supply");
                icon_name = "found-battery-twenty";
                manager->priv->notify_low_num[0]++;
        }
        else if (percent <= 10 && percent > 5 && manager->priv->notify_low_num[1] == 0) {
                msg = _("The battery is too low, please connect it to the power supply");;
                icon_name = "found-battery-ten";
                manager->priv->notify_low_num[1]++;
        }
        else if (percent <= manager->priv->critical_percentage && manager->priv->notify_low_num[2] == 0) {
                gint action_type = g_settings_get_enum (manager->priv->settings, "critical-battery-action");
                action = power_action_type_to_string (action_type);

                if (action == NULL)
                        msg = _("The battery is too low, please connect it to the power supply");
                else
                        msg = g_strdup_printf (_("The battery is too low, the computer will go to %s"), action);
                
                icon_name = "found-battery-empty";
                manager->priv->notify_low_num[2]++;
        } 
        else if (percent <= manager->priv->action_percentage && !manager->priv->use_time_primary)
        {
                g_message ("Power is too low (percent: %d), do critical action.", percent);
                manager_critical_action_do (manager, FALSE);
        }

        if (msg && icon_name) {
                g_debug ("notify_low_battery: icon_name: %s, message: %s, percent: %d", icon_name, msg, percent);
                notify_low_battery(icon_name, msg);
        }

        if (action != NULL)
                g_free (msg);
}

// upower电源设置属性改变响应
static void
power_devices_changed (GDBusProxy *proxy,
                        GVariant    *changed_properties,
                        GStrv       invalidated_properties,
                        gpointer    data){
    
        CsdPowerManager *manager = CSD_POWER_MANAGER (data);
        if (g_variant_n_children (changed_properties) > 0){
                GVariantIter *iter;
                const gchar *key;
                GVariant *value = NULL;
                gboolean is_idle_inhibited;

                g_variant_get (changed_properties,
                            "a{sv}",
                            &iter);
                is_idle_inhibited = idle_is_session_inhibited (manager,
                                                       SESSION_INHIBIT_MASK_IDLE);
                if (is_idle_inhibited) {
                	g_debug ("inhibited, so using normal state");
			idle_set_mode (manager, CSD_POWER_IDLE_MODE_NORMAL);

			gpm_idletime_alarm_remove (manager->priv->idletime,
						   CSD_POWER_IDLETIME_BLANK_ID);
			gpm_idletime_alarm_remove (manager->priv->idletime,
						   CSD_POWER_IDLETIME_SLEEP_ID);
			gpm_idletime_alarm_remove (manager->priv->idletime,
						   CSD_POWER_IDLETIME_LOCK_ID);
                        gpm_idletime_alarm_remove (manager->priv->idletime,
						   CSD_POWER_IDLETIME_SCREENSAVER_ID);
			return;
		}

		while (g_variant_iter_loop (iter, "{&sv}", &key, &value))
		{      
			if (g_strcmp0 (key, "State") == 0)
				upower_prop_state_changed (manager, key);
			else if (g_strcmp0(key, "Percentage"))
				upower_prop_percentage_changed (manager, key);
		}
	}
}

static void
power_proxy_ready_cb (GObject             *source_object,
                      GAsyncResult        *res,
                      gpointer             user_data)
{
        GError *error = NULL;
        CsdPowerManager *manager = CSD_POWER_MANAGER (user_data);

        manager->priv->upower_proxy = g_dbus_proxy_new_for_bus_finish (res, &error);
        if (manager->priv->upower_proxy == NULL) {
                g_warning ("Could not connect to UPower: %s",
                           error->message);
                g_error_free (error);
        }else {
        
            g_signal_connect (manager->priv->upower_proxy, "g-properties-changed", G_CALLBACK (power_status_changed), manager);
        
        }
}

static void
power_device_proxy_ready_cb (GObject     *source_object,
                            GAsyncResult    *res,
                            gpointer        data){

        GError *error = NULL;
        CsdPowerManager *manager = CSD_POWER_MANAGER (data);

        manager->priv->upower_device_proxy = g_dbus_proxy_new_for_bus_finish (res, &error);
        if (manager->priv->upower_device_proxy == NULL) {
                g_warning ("Could not connect to UPower: %s",
                           error->message);
                g_error_free (error);
        }else {
                // 如果用户拔掉了电源，sleep-display-ac 与 sleep-diaplay-battery 用的是哪一个？
                // 需要监听电源状态，来更新
                g_signal_connect (manager->priv->upower_device_proxy, 
                                                        "g-properties-changed", 
                                                        G_CALLBACK (power_devices_changed), 
                                                        manager);
        }
}

static void
power_keyboard_proxy_ready_cb (GObject             *source_object,
                               GAsyncResult        *res,
                               gpointer             user_data)
{
        GVariant *k_now = NULL;
        GVariant *k_max = NULL;
        GError *error = NULL;
        CsdPowerManager *manager = CSD_POWER_MANAGER (user_data);

        manager->priv->upower_kdb_proxy = g_dbus_proxy_new_for_bus_finish (res, &error);
        if (manager->priv->upower_kdb_proxy == NULL) {
                g_warning ("Could not connect to UPower: %s",
                           error->message);
                g_error_free (error);
                goto out;
        }

        k_now = g_dbus_proxy_call_sync (manager->priv->upower_kdb_proxy,
                                        "GetBrightness",
                                        NULL,
                                        G_DBUS_CALL_FLAGS_NONE,
                                        -1,
                                        NULL,
                                        &error);
        if (k_now == NULL) {
                if (error->domain != G_DBUS_ERROR ||
                    error->code != G_DBUS_ERROR_UNKNOWN_METHOD) {
                        g_warning ("Failed to get brightness: %s",
                                   error->message);
                }
                g_error_free (error);
                goto out;
        }

        k_max = g_dbus_proxy_call_sync (manager->priv->upower_kdb_proxy,
                                        "GetMaxBrightness",
                                        NULL,
                                        G_DBUS_CALL_FLAGS_NONE,
                                        -1,
                                        NULL,
                                        &error);
        if (k_max == NULL) {
                g_warning ("Failed to get max brightness: %s", error->message);
                g_error_free (error);
                goto out;
        }

        g_variant_get (k_now, "(i)", &manager->priv->kbd_brightness_now);
        g_variant_get (k_max, "(i)", &manager->priv->kbd_brightness_max);

        /* set brightness to max if not currently set so is something
         * sensible */
        if (manager->priv->kbd_brightness_now <= 0) {
                gboolean ret;
                ret = upower_kbd_set_brightness (manager,
                                                 manager->priv->kbd_brightness_max,
                                                 &error);
                if (!ret) {
                        g_warning ("failed to initialize kbd backlight to %i: %s",
                                   manager->priv->kbd_brightness_max,
                                   error->message);
                        g_error_free (error);
                }
        }
out:
        if (k_now != NULL)
                g_variant_unref (k_now);
        if (k_max != NULL)
                g_variant_unref (k_max);
}

static void
lock_screensaver (CsdPowerManager *manager)
{
        gboolean do_lock;

        do_lock = g_settings_get_boolean (manager->priv->settings_screensaver,
                                          "lock-enabled");
        if (!do_lock || !manager->priv->screensaver_proxy)
                return;

        /* connect to the screensaver first */
        g_dbus_proxy_call_sync (manager->priv->screensaver_proxy,
                                "Lock",
                                NULL,
                                G_DBUS_CALL_FLAGS_NONE, 
                                -1,
                                NULL, NULL);
}

static void
idle_send_to_sleep (CsdPowerManager *manager)
{
        gboolean is_inhibited;
        //gboolean is_idle;

        /* check the session is not now inhibited */
        is_inhibited = idle_is_session_inhibited (manager,
                                                  SESSION_INHIBIT_MASK_SUSPEND);
        if (is_inhibited) {
                g_debug ("suspend inhibited");
                return;
        }

        /* check the session is really idle*/
       //is_idle = idle_is_session_idle (manager);
       //if (!is_idle) {
                //g_debug ("session is not idle, cannot SLEEP");
                //return;
        //}

        /* send to sleep, and cancel timeout */
        g_debug ("sending to SLEEP");
        idle_set_mode (manager, CSD_POWER_IDLE_MODE_SLEEP);
}

// 空闲时间超时闹钟处理
static void
idle_idletime_alarm_expired_cb (GpmIdletime *idletime,
                                guint alarm_id,
                                CsdPowerManager *manager)
{
        g_debug ("idletime alarm: %i", alarm_id);

        switch (alarm_id) {
            case CSD_POWER_IDLETIME_DIM_ID:
                idle_set_mode (manager, CSD_POWER_IDLE_MODE_DIM);
                break;
            case CSD_POWER_IDLETIME_BLANK_ID:           // 无操作超时则息屏
                idle_set_mode (manager, CSD_POWER_IDLE_MODE_BLANK);
                break;
            case CSD_POWER_IDLETIME_SLEEP_ID:           // 无操作超时则睡眠
                idle_send_to_sleep (manager);
                break;
            case CSD_POWER_IDLETIME_SCREENSAVER_ID:     // 无操作超时则拉起屏保
                idle_set_mode (manager, CSD_POWER_IDLE_MODE_SCREENSAVER);
            default:
                break;
        }
}

static void
idle_idletime_reset_cb (GpmIdletime *idletime,
                        CsdPowerManager *manager)
{
        g_debug ("idletime reset");

        idle_set_mode (manager, CSD_POWER_IDLE_MODE_NORMAL);
}

// 监听gsettings配置改变信号
static void
engine_settings_key_changed_cb (GSettings *settings,
                                const gchar *key,
                                CsdPowerManager *manager)
{
        /* note: you *have* to check if your key was changed here before
         * doing anything here. this gets invoked on module stop, and
         * will crash c-s-d if you don't. */
        if (g_strcmp0 (key, "use-time-for-policy") == 0)
        {
                manager->priv->use_time_primary = g_settings_get_boolean (settings, key);
                return;
        }

        if (g_strcmp0 (key, "idle-dim-time") == 0)
        {
                refresh_idle_dim_settings (manager);
                return;
        }

        if (g_str_has_prefix (key, "sleep-inactive")
        || g_str_has_prefix (key, "sleep-display")                      // 无操作超时息屏
        || g_str_has_prefix (key, "idle-delay")                           // 无操作超时锁屏
        || g_str_equal (key, "idle-screensaver-timeout"))       // 无操作超时显示屏保
            {
                idle_configure (manager);
                return;
        }

        if (g_str_has_prefix (key, "backlight-helper"))
        {
                backlight_override_settings_refresh (manager);
                return;
        }

        // 改变运行模式
        if (g_strcmp0(key, "run-mode") == 0 )
        {
                CsdPowerRunMode mode;
                mode= g_settings_get_enum (manager->priv->settings, "run-mode");
                set_run_mode(manager, mode, TRUE);
                return;
        }
}

static void
engine_session_active_changed_cb (CdosSettingsSession *session,
                                  GParamSpec *pspec,
                                  CsdPowerManager *manager)
{
        /* when doing the fast-user-switch into a new account,
         * ensure the new account is undimmed and with the backlight on */
        idle_set_mode (manager, CSD_POWER_IDLE_MODE_NORMAL);
}

/* This timer goes off every few minutes, whether the user is idle or not,
   to try and clean up anything that has gone wrong.

   It calls disable_builtin_screensaver() so that if xset has been used,
   or some other program (like xlock) has messed with the XSetScreenSaver()
   settings, they will be set back to sensible values (if a server extension
   is in use, messing with xlock can cause the screensaver to never get a wakeup
   event, and could cause monitor power-saving to occur, and all manner of
   heinousness.)

   This code was originally part of cdos-screensaver, see
   http://git.gnome.org/browse/cdos-screensaver/tree/src/gs-watcher-x11.c?id=fec00b12ec46c86334cfd36b37771cc4632f0d4d#n530
 */
static gboolean
disable_builtin_screensaver (gpointer unused)
{
        int current_server_timeout, current_server_interval;
        int current_prefer_blank,   current_allow_exp;
        int desired_server_timeout, desired_server_interval;
        int desired_prefer_blank,   desired_allow_exp;

        XGetScreenSaver (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()),
                         &current_server_timeout,
                         &current_server_interval,
                         &current_prefer_blank,
                         &current_allow_exp);

        desired_server_timeout  = current_server_timeout;
        desired_server_interval = current_server_interval;
        desired_prefer_blank    = current_prefer_blank;
        desired_allow_exp       = current_allow_exp;

        desired_server_interval = 0;

        /* I suspect (but am not sure) that DontAllowExposures might have
           something to do with powering off the monitor as well, at least
           on some systems that don't support XDPMS?  Who know... */
        desired_allow_exp = AllowExposures;

        /* When we're not using an extension, set the server-side timeout to 0,
           so that the server never gets involved with screen blanking, and we
           do it all ourselves.  (However, when we *are* using an extension,
           we tell the server when to notify us, and rather than blanking the
           screen, the server will send us an X event telling us to blank.)
        */
        desired_server_timeout = 0;

        if (desired_server_timeout     != current_server_timeout
            || desired_server_interval != current_server_interval
            || desired_prefer_blank    != current_prefer_blank
            || desired_allow_exp       != current_allow_exp) {

                g_debug ("disabling server builtin screensaver:"
                         " (xset s %d %d; xset s %s; xset s %s)",
                         desired_server_timeout,
                         desired_server_interval,
                         (desired_prefer_blank ? "blank" : "noblank"),
                         (desired_allow_exp ? "expose" : "noexpose"));

                XSetScreenSaver (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()),
                                 desired_server_timeout,
                                 desired_server_interval,
                                 desired_prefer_blank,
                                 desired_allow_exp);

                XSync (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), FALSE);
        }

        return TRUE;
}

static void
inhibit_lid_switch_done (GObject      *source,
                         GAsyncResult *result,
                         gpointer      user_data)
{
        GDBusProxy *proxy = G_DBUS_PROXY (source);
        CsdPowerManager *manager = CSD_POWER_MANAGER (user_data);
        GError *error = NULL;
        GVariant *res;
        GUnixFDList *fd_list = NULL;
        gint idx;

        res = g_dbus_proxy_call_with_unix_fd_list_finish (proxy, &fd_list, result, &error);
        if (res == NULL) {
                g_warning ("Unable to inhibit lid switch: %s", error->message);
                g_error_free (error);
        } else {
                g_variant_get (res, "(h)", &idx);
                manager->priv->inhibit_lid_switch_fd = g_unix_fd_list_get (fd_list, idx, &error);
                if (manager->priv->inhibit_lid_switch_fd == -1) {
                        g_warning ("Failed to receive system inhibitor fd: %s", error->message);
                        g_error_free (error);
                }
                g_debug ("System inhibitor fd is %d", manager->priv->inhibit_lid_switch_fd);
                g_object_unref (fd_list);
                g_variant_unref (res);
        }
}

static void
inhibit_lid_switch (CsdPowerManager *manager)
{
        GVariant *params;

        if (manager->priv->inhibit_lid_switch_taken) {
                g_debug ("already inhibited lid-switch");
                return;
        }
        g_debug ("Adding lid switch system inhibitor");
        manager->priv->inhibit_lid_switch_taken = TRUE;

        params = g_variant_new ("(ssss)",
                                "handle-lid-switch",
                                g_get_user_name (),
                                "Multiple displays attached",
                                "block");
        g_dbus_proxy_call_with_unix_fd_list (manager->priv->logind_proxy,
                                             "Inhibit",
                                             params,
                                             0,
                                             G_MAXINT,
                                             NULL,
                                             NULL,
                                             inhibit_lid_switch_done,
                                             manager);
}

static void
inhibit_suspend_done (GObject      *source,
                      GAsyncResult *result,
                      gpointer      user_data)
{
        GDBusProxy *proxy = G_DBUS_PROXY (source);
        CsdPowerManager *manager = CSD_POWER_MANAGER (user_data);
        GError *error = NULL;
        GVariant *res;
        GUnixFDList *fd_list = NULL;
        gint idx;

        res = g_dbus_proxy_call_with_unix_fd_list_finish (proxy, &fd_list, result, &error);
        if (res == NULL) {
                g_warning ("Unable to inhibit suspend: %s", error->message);
                g_error_free (error);
        } else {
                g_variant_get (res, "(h)", &idx);
                manager->priv->inhibit_suspend_fd = g_unix_fd_list_get (fd_list, idx, &error);
                if (manager->priv->inhibit_suspend_fd == -1) {
                        g_warning ("Failed to receive system inhibitor fd: %s", error->message);
                        g_error_free (error);
                }
                g_debug ("System inhibitor fd is %d", manager->priv->inhibit_suspend_fd);
                g_object_unref (fd_list);
                g_variant_unref (res);
        }
}

/* We take a delay inhibitor here, which causes logind to send a
 * PrepareToSleep signal, which gives us a chance to lock the screen
 * and do some other preparations.
 */
static void
inhibit_suspend (CsdPowerManager *manager)
{
        if (manager->priv->inhibit_suspend_taken) {
                g_debug ("already inhibited lid-switch");
                return;
        }
        g_debug ("Adding suspend delay inhibitor");
        manager->priv->inhibit_suspend_taken = TRUE;
        g_dbus_proxy_call_with_unix_fd_list (manager->priv->logind_proxy,
                                             "Inhibit",
                                             g_variant_new ("(ssss)",
                                                            "sleep",
                                                            g_get_user_name (),
                                                            "Cinnamon needs to lock the screen",
                                                            "delay"),
                                             0,
                                             G_MAXINT,
                                             NULL,
                                             NULL,
                                             inhibit_suspend_done,
                                             manager);
}

static void
uninhibit_suspend (CsdPowerManager *manager)
{
        if (manager->priv->inhibit_suspend_fd == -1) {
                g_debug ("no suspend delay inhibitor");
                return;
        }
        g_debug ("Removing suspend delay inhibitor");
        close (manager->priv->inhibit_suspend_fd);
        manager->priv->inhibit_suspend_fd = -1;
        manager->priv->inhibit_suspend_taken = FALSE;
}

static gboolean 
filter_epson_resolution (CsdPowerManager *manager)
{
        if (!manager->priv->x11_screen)
                return FALSE;
        GnomeRROutput **outputs = gnome_rr_screen_list_outputs (manager->priv->x11_screen);
        gint i = 0;
        for (i=0; outputs[i] != NULL; ++i)
        {
             gchar *vendor, *product, *serial;
             GnomeRROutput *output = outputs[i];
             gnome_rr_output_get_ids_from_edid (output, &vendor, &product, &serial);

             gchar *lowercase = g_ascii_strdown (product, strlen (product));
             g_debug ("vendor %s, product %s: %s, serial %s\n", vendor, product, lowercase, serial);

             if (g_str_match_string ("epson", lowercase, TRUE)){
                 GnomeRRMode *mode = gnome_rr_output_get_current_mode (output);
                 guint width = gnome_rr_mode_get_width (mode);
                 guint height = gnome_rr_mode_get_height (mode);
                 g_debug ("width and height for epson: %d %d\n", width, height);
                 if (height == 900 && width == 1600)
                     return TRUE;
             }
             g_free (lowercase);
        }
        return FALSE;
}

/**
 * pause_play_switch
 *
 * add by niyl, swithc media player status before suspend
 * 
 * @media_player_name: make log clear when have more than one player software.
 * @media_player_bus_name: MEDIA_PLAYER_RHYTHMBOX or MEDIA_PLAYER_SPOTIFY
 * @wanted_status: if now status is not wanted status, function just quit.
 * @methmod_name: method name called by GDBusProxy
 *
 */
static void
pause_play_switch (const gchar *media_player_name,
                   const gchar *media_player_bus_name,
                   const gchar *wanted_status,
                   const gchar *method_name) {
    GDBusProxy *proxy;
    GError *error = NULL;
    gchar *play_back_status = NULL;
    GVariant *status_variant = NULL;

    proxy = g_dbus_proxy_new_for_bus_sync (G_BUS_TYPE_SESSION,
                                           G_DBUS_PROXY_FLAGS_NONE,
                                           NULL,
                                           media_player_bus_name,
                                           MEDIA_PLAYER_OBJECT_PATH,
                                           MEDIA_PLAYER_INTERFACE,
                                           NULL,
                                           &error);
    if (error) {
        g_debug ("Get MediaPlayer2 Proxy failed, because: %s\n", error->message);
        g_error_free (error);
    } else {
        status_variant = g_dbus_proxy_get_cached_property (proxy, "PlaybackStatus");
        if (status_variant != NULL) {
            play_back_status = (gchar *)g_variant_get_string (status_variant, NULL);

            if (play_back_status == NULL || g_strcmp0 (play_back_status, wanted_status) != 0) {
                g_debug ("%s PlaybackStatus is null or not %s, do nothing\n", media_player_name, wanted_status);
                return;
            }

            g_dbus_proxy_call_sync (proxy,
                                    method_name,
                                    NULL,
                                    G_DBUS_CALL_FLAGS_NONE,
                                    -1,
                                    NULL,
                                    &error);

            if (error) {
                g_debug ("org.mpris.MediaPlayer2.Player method %s called failed\n", method_name);
                g_error_free (error);
            }
            g_variant_unref (status_variant);
        } else {
            g_debug ("%s cannot lookup property PlaybackStatus, maybe player not open\n", media_player_name);
        }
    }
    if (proxy)
        g_object_unref (proxy);
}
/* add end, pause media player before suspend */

static gboolean
handle_suspend_actions (CsdPowerManager *manager)
{
        if (is_session_active (manager) == FALSE) 
        {
                g_warning ("System suspend, but session is not active.");
                return FALSE;
        }
        g_message ("Handle system suspend.");
        if (!filter_epson_resolution (manager)){
                gboolean do_lock;

                do_lock = g_settings_get_boolean (manager->priv->settings,
                                "lock-on-suspend");
                g_message ("Lock on suspend from suspend: %d", do_lock);
                if (do_lock)
                        lock_screensaver (manager);
        } 

        /* Try pause rhythmbox when suspend */
        pause_play_switch ("Rhythmbox", MEDIA_PLAYER_RHYTHMBOX, "Playing", "Pause");
        
         /* lift the delay inhibit, so logind can proceed */
        uninhibit_suspend (manager);
        return FALSE;
}

static gboolean
handle_resume_actions (CsdPowerManager *manager)
{
        gboolean ret;
        GError *error = NULL;
        if (is_session_active (manager) == FALSE) 
        {
                g_warning ("System resume, but session is not active.");
                return FALSE;
        }
        g_message ("Handle system resume.");
        if (!filter_epson_resolution (manager)) 
        {
                gboolean do_lock;
                do_lock = g_settings_get_boolean (manager->priv->settings,
                                "lock-on-suspend");
                g_message ("Lock on suspend from resume:%d.", do_lock);
                if (do_lock)
                        lock_screensaver (manager);
        }
        /*
        *signal map_event from gtk will send after simulateuser
        *delay the send time of signal simulateuse */
        struct timeval delay;
        delay.tv_sec = 2;
        delay.tv_usec = 0;
        int res = 0;
        res = select(0,NULL,NULL,NULL,&delay);
        if(res == -1)
                return FALSE;

        /* this displays the unlock dialogue so the user doesn't have
         * to move the mouse or press any key before the window comes up */
        g_dbus_connection_call (manager->priv->connection,
                                GS_DBUS_NAME,
                                GS_DBUS_PATH,
                                GS_DBUS_INTERFACE,
                                "SimulateUserActivity",
                                NULL, NULL,
                                G_DBUS_CALL_FLAGS_NONE, -1,
                                NULL, NULL, NULL);

        /* close existing notifications on resume, the system power
         * state is probably different now */
        notify_close_if_showing (manager->priv->notification_low);
        notify_close_if_showing (manager->priv->notification_discharging);

        /* ensure we turn the panel back on after resume */
        if (manager->priv->x11_screen) {
                ret = gnome_rr_screen_set_dpms_mode (manager->priv->x11_screen,
                                                GNOME_RR_DPMS_ON,
                                                &error);
                if (!ret) {
                        g_warning ("failed to turn the panel on after resume: %s",
                                error->message);
                        g_error_free (error);
                }
        }
        /* set up the delay again */
        inhibit_suspend (manager);
        return FALSE;
}

#if ! UP_CHECK_VERSION(0,99,0)
static void
upower_notify_sleep_cb (UpClient *client,
                        UpSleepKind sleep_kind,
                        CsdPowerManager *manager)
{
        handle_suspend_actions (manager);
}

static void
upower_notify_resume_cb (UpClient *client,
                         UpSleepKind sleep_kind,
                         CsdPowerManager *manager)
{
        handle_resume_actions (manager);
}
#endif

/**
 * 执行挂起或者休眠(通过 login1，在睡前会发出 PrepareForSleep 信号)
 * settings daemon 监听这个信号，并调用锁屏后再挂起或者休眠
 * 这个信号的发出是无差别的，因此如果多用户登录，其中一个用户挂起，
 * 那么所有的用户都会收到，并且触发锁屏。导致的问题是：
 * 其中一个用户挂起唤醒再切换用户到其他人，还需要再输入锁屏密码
 * add by lit 07.06.17
 */
static gboolean
is_session_active (CsdPowerManager *manager)
{
        /* GVariant  *variant;
        gboolean is_session_active = FALSE;

        variant = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (manager->priv->session_proxy),
                                                        "SessionIsActive");

        //g_debug ("get_session_active: %d\n", is_session_active);
        if (variant) {
                is_session_active = g_variant_get_boolean (variant);
                //g_debug ("get_session_active: %d\n", is_session_active);
                g_variant_unref (variant);
        } */
        return cdos_settings_session_get_state (manager->priv->session);
}

static void
logind_proxy_signal_cb (GDBusProxy  *proxy,
                        const gchar *sender_name,
                        const gchar *signal_name,
                        GVariant    *parameters,
                        gpointer     user_data)
{
        CsdPowerManager *manager = CSD_POWER_MANAGER (user_data);
        gboolean is_about_to_suspend;

        if (g_strcmp0 (signal_name, "PrepareForSleep") != 0)
                return;
        g_variant_get (parameters, "(b)", &is_about_to_suspend);
        g_message ("Logind emit signal: PrepareForSleep: %d", is_about_to_suspend);

        if (is_about_to_suspend) {
                //g_timeout_add (250, (GSourceFunc)handle_suspend_actions, manager);
                handle_suspend_actions (manager);
        } else {
                //休眠、睡眠唤醒后有概率性没有激活session，因此延迟等待session激活。
                g_timeout_add (250, (GSourceFunc)handle_resume_actions, manager);
                //handle_resume_actions (manager);
        }
}

// xunli
static void
power_status_changed (GDBusProxy *proxy,
                        GVariant    *changed_properties,
                        GStrv       invalidated_properties,
                        gpointer    data){
    
        CsdPowerManager *manager = CSD_POWER_MANAGER (data);
        if (g_variant_n_children (changed_properties) > 0){
            GVariantIter *iter;
            const gchar *key;
            GVariant *value;

            g_variant_get (changed_properties,
                            "a{sv}",
                            &iter);

            while (g_variant_iter_loop (iter, "{&sv}", &key, &value)){
                if (g_strcmp0 (key, "OnBattery") == 0){
                    gboolean on_battery = FALSE;
                    on_battery = g_variant_get_boolean (value);

                    guint timeout_blank = 0;
                    guint current_idle_time = 0;

                    current_idle_time = gpm_idletime_get_time (manager->priv->idletime) / 1000;
                    if (on_battery) {
                        timeout_blank = g_settings_get_int (manager->priv->settings,
                                                    "sleep-display-battery");
                    } else {
                        timeout_blank = g_settings_get_int (manager->priv->settings,
                                                    "sleep-display-ac");
                    }
                    if (timeout_blank != 0) {
                        gpm_idletime_alarm_set (manager->priv->idletime,
                                        CSD_POWER_IDLETIME_BLANK_ID,
                                        idle_adjust_timeout_blank (current_idle_time, timeout_blank) * 1000);
                    } else {
                        gpm_idletime_alarm_remove (manager->priv->idletime,
                                           CSD_POWER_IDLETIME_BLANK_ID);
                    }

                
                }
            }
        }
}


static gboolean
is_hardware_a_virtual_machine (void)
{
        const gchar *str;
        gboolean ret = FALSE;
        GError *error = NULL;
        GVariant *inner;
        GVariant *variant = NULL;
        GDBusConnection *connection;

        connection = g_bus_get_sync (G_BUS_TYPE_SYSTEM,
                                     NULL,
                                     &error);
        if (connection == NULL) {
                g_warning ("system bus not available: %s", error->message);
                g_error_free (error);
                goto out;
        }
        variant = g_dbus_connection_call_sync (connection,
                                               "org.freedesktop.systemd1",
                                               "/org/freedesktop/systemd1",
                                               "org.freedesktop.DBus.Properties",
                                               "Get",
                                               g_variant_new ("(ss)",
                                                              "org.freedesktop.systemd1.Manager",
                                                              "Virtualization"),
                                               NULL,
                                               G_DBUS_CALL_FLAGS_NONE,
                                               -1,
                                               NULL,
                                               &error);
        if (variant == NULL) {
                g_debug ("Failed to get property '%s': %s", "Virtualization", error->message);
                g_error_free (error);
                goto out;
        }

        /* on bare-metal hardware this is the empty string,
         * otherwise an identifier such as "kvm", "vmware", etc. */
        g_variant_get (variant, "(v)", &inner);
        str = g_variant_get_string (inner, NULL);
        if (str != NULL && str[0] != '\0')
                ret = TRUE;
out:
        if (connection != NULL)
                g_object_unref (connection);
        if (variant != NULL)
                g_variant_unref (variant);
        return ret;
}

static gboolean
x11_screen_timeout (gpointer user_data)
{
        GError *error = NULL;
        CsdPowerManager *manager = CSD_POWER_MANAGER (user_data);
        if (!is_found_desktop_env ())
        {
                g_warning ("Current window manager is not 'Mutter', something does't support.");
                return FALSE;
        }
        manager->priv->x11_screen = gnome_rr_screen_new (gdk_screen_get_default (), &error);
        if (error)
        {
                g_warning ("Failed init x11 screen: %s", error->message);
                g_clear_error (&error);
                return TRUE;
        }
        gnome_rr_screen_set_dpms_mode (manager->priv->x11_screen,
                                        GNOME_RR_DPMS_ON,
                                        &error);
        if (error) 
        {
                g_warning ("Failed set DPMS mode: %s", error->message);
                g_clear_error (&error);
        }
        return FALSE;
}

gboolean
csd_power_manager_start (CsdPowerManager *manager,
                         GError **error)
{
        gboolean ret;

        g_debug ("Starting power manager");
        cdos_settings_profile_start (NULL);

        /* coldplug the list of screens */
        /* removed to x11_screen_timeout() */
        // manager->priv->x11_screen = gnome_rr_screen_new (gdk_screen_get_default (), error);
        // if (manager->priv->x11_screen == NULL)
        //         return FALSE;

        /* Set up the logind proxy */
        manager->priv->logind_proxy =
                g_dbus_proxy_new_for_bus_sync (G_BUS_TYPE_SYSTEM,
                                               G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START,
                                               NULL,
                                               SYSTEMD_DBUS_NAME,
                                               SYSTEMD_DBUS_PATH,
                                               SYSTEMD_DBUS_INTERFACE,
                                               NULL,
                                               error);
        g_signal_connect (manager->priv->logind_proxy, "g-signal",
                          G_CALLBACK (logind_proxy_signal_cb),
                          manager);

        /* Set up a delay inhibitor to be informed about suspend attempts */
        inhibit_suspend (manager);

        /* Disable logind's lid handling while g-s-d is active */
        inhibit_lid_switch (manager);

        /* track the active session */
        manager->priv->session = cdos_settings_session_new ();
        g_signal_connect (manager->priv->session, "notify::state",
                          G_CALLBACK (engine_session_active_changed_cb),
                          manager);

        manager->priv->kbd_brightness_old = -1;
        manager->priv->kbd_brightness_pre_dim = -1;
        manager->priv->pre_dim_brightness = -1;
        manager->priv->settings = g_settings_new (CSD_POWER_SETTINGS_SCHEMA);
        g_signal_connect (manager->priv->settings, "changed",
                          G_CALLBACK (engine_settings_key_changed_cb), manager);
        manager->priv->settings_screensaver = g_settings_new ("org.cdos.desktop.screensaver");
        manager->priv->up_client = up_client_new ();
        manager->priv->lid_is_closed = up_client_get_lid_is_closed (manager->priv->up_client);

        g_signal_connect (manager->priv->up_client, "device-added",
                          G_CALLBACK (engine_device_added_cb), manager);
        g_signal_connect (manager->priv->up_client, "device-removed",
                          G_CALLBACK (engine_device_removed_cb), manager);
#if UP_CHECK_VERSION(0,99,0)
        g_signal_connect_after (manager->priv->up_client, "notify::lid-is-closed",
                                G_CALLBACK (lid_state_changed_cb), manager);
#else
        g_signal_connect_after (manager->priv->up_client, "changed",
                                G_CALLBACK (up_client_changed_cb), manager);
#endif

        /* connect to UPower for async power operations */
        g_dbus_proxy_new_for_bus (G_BUS_TYPE_SYSTEM,
                                  //G_DBUS_PROXY_FLAGS_DO_NOT_LOAD_PROPERTIES,
                                  G_DBUS_PROXY_FLAGS_NONE,
                                  NULL,
                                  UPOWER_DBUS_NAME,
                                  UPOWER_DBUS_PATH,
                                  UPOWER_DBUS_INTERFACE,
                                  NULL,
                                  power_proxy_ready_cb,
                                  manager);

        g_dbus_proxy_new_for_bus (G_BUS_TYPE_SYSTEM,
                                  G_DBUS_PROXY_FLAGS_NONE,
                                  NULL,
                                  UPOWER_DBUS_NAME,
                                  UPOWER_DBUS_DEVICE_PATH,
                                  UPOWER_DBUS_DEVICE_INTERFACE,
                                  NULL,
                                  power_device_proxy_ready_cb,
                                  manager);

        /* connect to UPower for keyboard backlight control */
        g_dbus_proxy_new_for_bus (G_BUS_TYPE_SYSTEM,
                                  G_DBUS_PROXY_FLAGS_DO_NOT_LOAD_PROPERTIES,
                                  NULL,
                                  UPOWER_DBUS_NAME,
                                  UPOWER_DBUS_PATH_KBDBACKLIGHT,
                                  UPOWER_DBUS_INTERFACE_KBDBACKLIGHT,
                                  NULL,
                                  power_keyboard_proxy_ready_cb,
                                  manager);

        /* connect to the session */
        g_dbus_proxy_new_for_bus (G_BUS_TYPE_SESSION,
                                  G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START,
                                  NULL,
                                  GNOME_SESSION_DBUS_NAME,
                                  GNOME_SESSION_DBUS_PATH,
                                  GNOME_SESSION_DBUS_INTERFACE,
                                  NULL,
                                  session_proxy_ready_cb,
                                  manager);
        g_dbus_proxy_new_for_bus (G_BUS_TYPE_SESSION,
                                  G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START,
                                  NULL,
                                  GNOME_SESSION_DBUS_NAME,
                                  GNOME_SESSION_DBUS_PATH_PRESENCE,
                                  GNOME_SESSION_DBUS_INTERFACE_PRESENCE,
                                  NULL,
                                  session_presence_proxy_ready_cb,
                                  manager);
        g_dbus_proxy_new_for_bus (G_BUS_TYPE_SESSION,
                                  G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START,
                                  NULL,
                                  GS_DBUS_NAME,
                                  GS_DBUS_PATH,
                                  GS_DBUS_INTERFACE,
                                  NULL,
                                  screensaver_proxy_ready_cb,
                                  manager);

        manager->priv->devices_array = g_ptr_array_new_with_free_func (g_object_unref);
        manager->priv->canberra_context = ca_gtk_context_get_for_screen (gdk_screen_get_default ());

        manager->priv->phone = gpm_phone_new ();
        g_signal_connect (manager->priv->phone, "device-added",
                          G_CALLBACK (phone_device_added_cb), manager);
        g_signal_connect (manager->priv->phone, "device-removed",
                          G_CALLBACK (phone_device_removed_cb), manager);
        g_signal_connect (manager->priv->phone, "device-refresh",
                          G_CALLBACK (phone_device_refresh_cb), manager);

        /* create a fake virtual composite battery */
        manager->priv->device_composite = up_device_new ();
        g_object_set (manager->priv->device_composite,
                      "kind", UP_DEVICE_KIND_BATTERY,
                      "is-rechargeable", TRUE,
                      "native-path", "dummy:composite_battery",
                      "power-supply", TRUE,
                      "is-present", TRUE,
                      NULL);

        /* get backlight setting overrides */
        manager->priv->backlight_helper_preference_args = NULL;
        backlight_override_settings_refresh (manager);

        /* get percentage policy */
        manager->priv->low_percentage = g_settings_get_int (manager->priv->settings,
                                                            "percentage-low");
        manager->priv->critical_percentage = g_settings_get_int (manager->priv->settings,
                                                                 "percentage-critical");
        manager->priv->action_percentage = g_settings_get_int (manager->priv->settings,
                                                               "percentage-action");

        /* get time policy */
        manager->priv->low_time = g_settings_get_int (manager->priv->settings,
                                                      "time-low");
        manager->priv->critical_time = g_settings_get_int (manager->priv->settings,
                                                           "time-critical");
        manager->priv->action_time = g_settings_get_int (manager->priv->settings,
                                                         "time-action");

        /* we can disable this if the time remaining is inaccurate or just plain wrong */
        manager->priv->use_time_primary = g_settings_get_boolean (manager->priv->settings,
                                                                  "use-time-for-policy");

        // 初始化电源模式
        manager->priv->hd_mode_timer_id = 0;
        manager->priv->hd_helper_spawned = FALSE;
        manager->priv->is_settings_run_mode = FALSE;
        manager->priv->run_mode_bp = CSD_POWER_RUN_MODE_NONE;
        manager->priv->run_mode = CSD_POWER_RUN_MODE_NONE;
        CsdPowerRunMode run_mode = g_settings_get_enum(manager->priv->settings,
                                                                "run-mode");
        CsdPowerRunMode run_mode_bp = g_settings_get_enum(manager->priv->settings,
                                                                "run-mode-before-psm");
	set_run_mode(manager, run_mode, FALSE);
	manager->priv->run_mode = run_mode;
        manager->priv->run_mode_bp = run_mode_bp;
	g_settings_set_enum(manager->priv->settings,
                                	"run-mode-before-psm",
					run_mode_bp);

        /* create IDLETIME watcher */
        manager->priv->idletime = gpm_idletime_new ();
        g_signal_connect (manager->priv->idletime, "reset",
                          G_CALLBACK (idle_idletime_reset_cb), manager);
        g_signal_connect (manager->priv->idletime, "alarm-expired",
                          G_CALLBACK (idle_idletime_alarm_expired_cb), manager);

        manager->priv->idle_monitor = gnome_idle_monitor_new ();

        /* ensure the default dpms timeouts are cleared */
        /* removed to x11_screen_timeout() */
        // ret = gnome_rr_screen_set_dpms_mode (manager->priv->x11_screen,
        //                                      GNOME_RR_DPMS_ON,
        //                                      error);
        // if (!ret) {
        //         g_warning ("Failed set DPMS mode: %s", (*error)->message);
        //         g_clear_error (error);
        // }

        /* coldplug the engine */
        engine_coldplug (manager);

        /* set the initial dim time that can adapt for the user */
        refresh_idle_dim_settings (manager);

        /* Make sure that Xorg's DPMS extension never gets in our way. The defaults seem to have changed in Xorg 1.14
         * being "0" by default to being "600" by default 
         * https://bugzilla.gnome.org/show_bug.cgi?id=709114
         */
        gdk_x11_display_error_trap_push (gdk_display_get_default ());
        int dummy;
        if (DPMSQueryExtension(GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), &dummy, &dummy)) {
            DPMSSetTimeouts (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), 0, 0, 0);
        }
        gdk_x11_display_error_trap_pop (gdk_display_get_default ());

        manager->priv->xscreensaver_watchdog_timer_id = g_timeout_add_seconds (XSCREENSAVER_WATCHDOG_TIMEOUT,
                                                                               disable_builtin_screensaver,
                                                                               NULL);
                                                                               
        manager->priv->x11_screen_timer_id = g_timeout_add_seconds (X11_SCREEN_DELAY_TIMEOUT, (GSourceFunc) x11_screen_timeout, manager);

        /* don't blank inside a VM */
        manager->priv->is_virtual_machine = is_hardware_a_virtual_machine ();

        // 初始化低电量通知记录
        for (int i = 0; i< 3; i++) {
                manager->priv->notify_low_num[i] = 0;
        }

        cdos_settings_profile_end (NULL);
        return TRUE;
}

void
csd_power_manager_stop (CsdPowerManager *manager)
{
        g_debug ("Stopping power manager");

        if (manager->priv->x11_screen_timer_id > 0) {
                g_source_remove (manager->priv->x11_screen_timer_id);
                manager->priv->x11_screen_timer_id = 0;
        }

        if (manager->priv->bus_cancellable != NULL) {
                g_cancellable_cancel (manager->priv->bus_cancellable);
                g_object_unref (manager->priv->bus_cancellable);
                manager->priv->bus_cancellable = NULL;
        }

        if (manager->priv->introspection_data) {
                g_dbus_node_info_unref (manager->priv->introspection_data);
                manager->priv->introspection_data = NULL;
        }

        kill_lid_close_safety_timer (manager);

        g_signal_handlers_disconnect_by_data (manager->priv->up_client, manager);

        if (manager->priv->connection != NULL) {
                g_object_unref (manager->priv->connection);
                manager->priv->connection = NULL;
        }

        if (manager->priv->session != NULL) {
                g_object_unref (manager->priv->session);
                manager->priv->session = NULL;
        }

        if (manager->priv->settings != NULL) {
                g_object_unref (manager->priv->settings);
                manager->priv->settings = NULL;
        }

        if (manager->priv->settings_screensaver != NULL) {
                g_object_unref (manager->priv->settings_screensaver);
                manager->priv->settings_screensaver = NULL;
        }

        if (manager->priv->up_client != NULL) {
                g_object_unref (manager->priv->up_client);
                manager->priv->up_client = NULL;
        }

        if (manager->priv->inhibit_lid_switch_fd != -1) {
                close (manager->priv->inhibit_lid_switch_fd);
                manager->priv->inhibit_lid_switch_fd = -1;
                manager->priv->inhibit_lid_switch_taken = FALSE;
        }
        if (manager->priv->inhibit_suspend_fd != -1) {
                close (manager->priv->inhibit_suspend_fd);
                manager->priv->inhibit_suspend_fd = -1;
                manager->priv->inhibit_suspend_taken = FALSE;
        }

        if (manager->priv->logind_proxy != NULL) {
                g_object_unref (manager->priv->logind_proxy);
                manager->priv->logind_proxy = NULL;
        }

        g_free (manager->priv->backlight_helper_preference_args);
        manager->priv->backlight_helper_preference_args = NULL;

        if (manager->priv->x11_screen != NULL) {
                g_object_unref (manager->priv->x11_screen);
                manager->priv->x11_screen = NULL;
        }

        g_ptr_array_unref (manager->priv->devices_array);
        manager->priv->devices_array = NULL;

        if (manager->priv->phone != NULL) {
                g_object_unref (manager->priv->phone);
                manager->priv->phone = NULL;
        }

        if (manager->priv->device_composite != NULL) {
                g_object_unref (manager->priv->device_composite);
                manager->priv->device_composite = NULL;
        }

        if (manager->priv->previous_icon != NULL) {
                g_object_unref (manager->priv->previous_icon);
                manager->priv->previous_icon = NULL;
        }

        g_free (manager->priv->previous_summary);
        manager->priv->previous_summary = NULL;

        if (manager->priv->upower_proxy != NULL) {
                g_object_unref (manager->priv->upower_proxy);
                manager->priv->upower_proxy = NULL;
        }

        if (manager->priv->upower_device_proxy != NULL) {
                g_object_unref (manager->priv->upower_device_proxy);
                manager->priv->upower_device_proxy = NULL;
        }

        if (manager->priv->session_proxy != NULL) {
                g_object_unref (manager->priv->session_proxy);
                manager->priv->session_proxy = NULL;
        }

        if (manager->priv->screensaver_proxy != NULL) {
                g_object_unref (manager->priv->screensaver_proxy);
                manager->priv->screensaver_proxy = NULL;
        }

        if (manager->priv->session_presence_proxy != NULL) {
                g_object_unref (manager->priv->session_presence_proxy);
                manager->priv->session_presence_proxy = NULL;
        }

        if (manager->priv->power_mode) {
                g_object_unref (manager->priv->power_mode);
                manager->priv->power_mode = NULL;
        }

        if (manager->priv->power_brightness) {
                g_object_unref (manager->priv->power_brightness);
                manager->priv->power_brightness = NULL;
        }

        if (manager->priv->critical_alert_timeout_id > 0) {
                g_source_remove (manager->priv->critical_alert_timeout_id);
                manager->priv->critical_alert_timeout_id = 0;
        }
        g_signal_handlers_disconnect_by_func (manager->priv->idletime,
                                              idle_idletime_reset_cb,
                                              manager);
        g_signal_handlers_disconnect_by_func (manager->priv->idletime,
                                              idle_idletime_alarm_expired_cb,
                                              manager);

        if (manager->priv->idletime != NULL) {
                g_object_unref (manager->priv->idletime);
                manager->priv->idletime = NULL;
        }

        if (manager->priv->xscreensaver_watchdog_timer_id > 0) {
                g_source_remove (manager->priv->xscreensaver_watchdog_timer_id);
                manager->priv->xscreensaver_watchdog_timer_id = 0;
        }

        if (manager->priv->hd_mode_timer_id > 0) {
                g_source_remove (manager->priv->hd_mode_timer_id);
                manager->priv->hd_mode_timer_id = 0;
        }
}

static void
brightness_on_signal_changed (GDBusProxy *proxy,
                        char       *sender_name,
                        char       *signal_name,
                        GVariant   *parameters,
                        gpointer    user_data)
{
        CsdPowerManager *manager = CSD_POWER_MANAGER (user_data);
        if (g_strcmp0 (signal_name, "Changed"))
                return;
        gint brightness = 0;
        g_variant_get (parameters, "(i)", &brightness);
        g_debug ("Current brightness: %d", brightness);
        backlight_iface_emit_changed (manager, CSD_POWER_DBUS_INTERFACE_SCREEN, brightness);	
}

static void
power_mode_get_enabled (GObject *source_object,
                        GAsyncResult *res,
                        gpointer user_data)
{
        GDBusProxy *proxy = G_DBUS_PROXY (source_object);
        CsdPowerManager *manager = CSD_POWER_MANAGER (user_data);
        GError *error = NULL;
        GVariant *v;
        gboolean enabled = FALSE;
        v = g_dbus_proxy_call_finish (proxy, res, &error);
        if (error)
        {
                g_debug ("Get power mode enabled failed: %s", error->message);
                g_error_free (error);
                return;
        }
        g_variant_get (v, "(b)", &enabled);
        g_debug ("Change power mode enabled:%d", enabled);
        manager->priv->power_mode_enable = enabled;
        if (enabled)
        {
                /* Fix: Fn+Q功能键事件，会导致其它功能键失效（无法接收事件）;
                 * 因此将失效功能键也改为由内核上报事件，并响应相关处理
                 */
                manager->priv->power_brightness = csd_get_power_brightness_proxy ();
                if (manager->priv->power_brightness)
                {
                        g_signal_connect (manager->priv->power_brightness, 
                                  "g-signal", 
                                  G_CALLBACK (brightness_on_signal_changed), 
                                  manager);
                }
        }
        g_variant_unref (v);
}

static void
csd_power_manager_init (CsdPowerManager *manager)
{
        manager->priv = CSD_POWER_MANAGER_GET_PRIVATE (manager);
        manager->priv->inhibit_lid_switch_fd = -1;
        manager->priv->inhibit_suspend_fd = -1;

        /* Fix: 针对Lenovo N7笔记本，添加Fn+Q切换电源模式的功能键;
         * 该功能键事件，由内核通过DBus上报;
         */
        manager->priv->power_mode = csd_get_power_mode_proxy ();
        manager->priv->power_mode_enable = FALSE;
        if (manager->priv->power_mode)
        {
                g_debug ("Support change power mode.");
                g_dbus_proxy_call (manager->priv->power_mode, 
                                   "GetEnabled", 
                                   NULL, 
                                   G_DBUS_CALL_FLAGS_NONE, 
                                   -1, 
                                   NULL, 
                                   power_mode_get_enabled, 
                                   manager);
        }
}

static void
csd_power_manager_finalize (GObject *object)
{
        CsdPowerManager *manager;

        manager = CSD_POWER_MANAGER (object);

        g_return_if_fail (manager->priv != NULL);


        G_OBJECT_CLASS (csd_power_manager_parent_class)->finalize (object);
}

/* returns new level */
static void
handle_method_call_keyboard (CsdPowerManager *manager,
                             const gchar *method_name,
                             GVariant *parameters,
                             GDBusMethodInvocation *invocation)
{
        gint step;
        gint value = -1;
        gboolean ret;
        guint percentage;
        GError *error = NULL;

        if (g_strcmp0 (method_name, "StepUp") == 0) {
                g_debug ("keyboard step up");
                step = BRIGHTNESS_STEP_AMOUNT (manager->priv->kbd_brightness_max);
                value = MIN (manager->priv->kbd_brightness_now + step,
                             manager->priv->kbd_brightness_max);
                ret = upower_kbd_set_brightness (manager, value, &error);

        } else if (g_strcmp0 (method_name, "StepDown") == 0) {
                g_debug ("keyboard step down");
                step = BRIGHTNESS_STEP_AMOUNT (manager->priv->kbd_brightness_max);
                value = MAX (manager->priv->kbd_brightness_now - step, 0);
                ret = upower_kbd_set_brightness (manager, value, &error);

        } else if (g_strcmp0 (method_name, "Toggle") == 0) {
                ret = upower_kbd_toggle (manager, &error);
        } else if (g_strcmp0(method_name, "TurnOn") == 0) {
                 g_debug ("keyboard turn on backlight");
                 g_dbus_method_invocation_return_value (invocation,
                                                        g_variant_new ("(u)",
                                                        turn_on_kbd_backlight()));
                return;
        } else if (g_strcmp0(method_name, "TurnOff") == 0) {
                g_debug ("keyboard turn off backlight");
                g_dbus_method_invocation_return_value (invocation,
                                                        g_variant_new ("(u)",
                                                        turn_off_kbd_backlight()));
                return;
        } else {
                g_assert_not_reached ();
        }

        /* return value */
        if (!ret) {
                g_dbus_method_invocation_return_gerror (invocation,
                                                        error);
                g_error_free (error);
        } else {
                percentage = ABS_TO_PERCENTAGE (0,
                                                manager->priv->kbd_brightness_max,
                                                value);
                g_dbus_method_invocation_return_value (invocation,
                                                       g_variant_new ("(u)",
                                                                      percentage));
        }
}

/**
 * 通过 hotkey 调节屏幕亮度，通过 media-key 模块转运到调用 settings daemon
 * 所以这里可以发出信号，使得 control-center 的 power 模块监听属性变化
 * 实时更新控件值
 * */

static void
backlight_iface_emit_changed (CsdPowerManager *manager,
		              const char      *interface_name,
			      gint32           value)
{
        GVariant *params;

       if (manager->priv->connection == NULL)
                return;

        params = g_variant_new_parsed ("(%s, [{'Brightness', <%i>}], @as [])", interface_name,
                                       value);
        g_dbus_connection_emit_signal (manager->priv->connection,
                                       NULL,
                                       CSD_POWER_DBUS_PATH,
                                       "org.freedesktop.DBus.Properties",
                                       "PropertiesChanged",
                                       params, NULL);
}

// Lenovo N7笔记本 获取亮度
static int
power_mode_get_brightness (CsdPowerManager *manager)
{
        GVariant *v;
        GError  *error = NULL;
        gint percentage = 0;
        v = \
        g_dbus_proxy_call_sync (manager->priv->power_brightness, 
                                "GetBrightness", 
                                NULL, 
                                G_DBUS_CALL_FLAGS_NONE, 
                                -1, 
                                NULL, 
                                &error);
        if (error)
        {
                g_warning ("Get brightness from power mode dbus failed:%s", error->message);
                g_error_free (error);
                return 0;
        }

        g_variant_get (v, "(i)", &percentage);
        g_debug ("Current brightness: %d",percentage);
        g_variant_unref (v);
        return percentage;
}

// Lenovo N7笔记本 设置亮度
static int
power_mode_set_brightness (CsdPowerManager *manager, gint percentage)
{
        GVariant *v;
        GError  *error = NULL;
        gboolean  success = FALSE;
        v = \
        g_dbus_proxy_call_sync (manager->priv->power_brightness, 
                                "SetBrightness", 
                                g_variant_new ("(i)", percentage), 
                                G_DBUS_CALL_FLAGS_NONE, 
                                -1, 
                                NULL, 
                                &error);
        if (error)
        {
                g_warning ("Set brightness to power mode dbus failed:%s", error->message);
                g_error_free (error);
                return 0;
        }

        g_variant_get (v, "(b)", &success);
        g_debug ("Set current brightness: %d, success:%d", percentage, success);
        g_variant_unref (v);

        return success == TRUE ? 1 : 0;
}

/* update color brightness of display monitors */
static void 
update_color_brightness_percentage (gint value)
{
	GSettings *settings;
	if (value > 100)
		value = 100;
	if (value < 0 )
		value = 0;
        g_debug ("Updage color brightness: %d", value);
	settings = g_settings_new ("org.cdos.settings-daemon.plugins.color");
	g_settings_set_int (settings, "brightness-percentage", value);
	g_object_unref (settings);
}

/* get color brightness of display monitors */
static int 
get_color_brightness_percentage (void)
{
	GSettings *settings;
        gint value;
	settings = g_settings_new ("org.cdos.settings-daemon.plugins.color");
	value = g_settings_get_int (settings, "brightness-percentage");
	g_object_unref (settings);
        return value;
}

static void
handle_method_call_screen (CsdPowerManager *manager,
                           const gchar *method_name,
                           GVariant  *parameters,
                           GDBusMethodInvocation *invocation)
{
        gint value = -1;
        GError *error = NULL;

        if (g_strcmp0 (method_name, "StepUp") == 0) {
                g_debug ("screen step up\n");
                value = backlight_step_up ( &error);
                if (backlight_get_status () == 0) {     // 台式机
                        value = get_color_brightness_percentage () + 5;
                        update_color_brightness_percentage (value);
                }
		backlight_iface_emit_changed (manager, CSD_POWER_DBUS_INTERFACE_SCREEN, value);	
        } else if (g_strcmp0 (method_name, "StepDown") == 0) {
                g_debug ("screen step down");
                value = backlight_step_down ( &error);
                if (backlight_get_status () == 0) {     // 台式机
                        value = get_color_brightness_percentage () - 5;
                        update_color_brightness_percentage (value);
                }
		backlight_iface_emit_changed (manager, CSD_POWER_DBUS_INTERFACE_SCREEN, value);	
        } else if (g_strcmp0 (method_name, "GetPercentage") == 0) {
                g_debug ("screen step down");
                value = get_brightness_percentage (manager);
	} else if (g_strcmp0 (method_name, "SetValue") == 0) {
		guint real_value = 0;
		g_variant_get (parameters, "(u)", &real_value);
                value = set_brightness_percentage (manager, &real_value);
                g_debug ("set real value is: %d\n", real_value);
	}else {
                g_assert_not_reached ();
        }

        /* return value */
        if (value < 0) {
                g_dbus_method_invocation_return_gerror (invocation,
                                                        error);
                g_error_free (error);
        } else {
                g_dbus_method_invocation_return_value (invocation,
                                                       g_variant_new ("(u)",
                                                                      value));
        }
}

static GVariant *
device_to_variant_blob (UpDevice *device)
{
        const gchar *object_path;
        gchar *device_icon;
        gdouble percentage;
        GIcon *icon;
        guint64 time_empty, time_full;
        guint64 time_state = 0;
        GVariant *value;
        UpDeviceKind kind;
        UpDeviceState state;

        icon = gpm_upower_get_device_icon (device, TRUE);
        device_icon = g_icon_to_string (icon);
        g_object_get (device,
                      "kind", &kind,
                      "percentage", &percentage,
                      "state", &state,
                      "time-to-empty", &time_empty,
                      "time-to-full", &time_full,
                      NULL);

        /* only return time for these simple states */
        if (state == UP_DEVICE_STATE_DISCHARGING)
                time_state = time_empty;
        else if (state == UP_DEVICE_STATE_CHARGING)
                time_state = time_full;

        /* get an object path, even for the composite device */
        object_path = up_device_get_object_path (device);
        if (object_path == NULL)
                object_path = CSD_DBUS_PATH;

        /* format complex object */
        value = g_variant_new ("(susdut)",
                               object_path,
                               kind,
                               device_icon,
                               percentage,
                               state,
                               time_state);
        g_free (device_icon);
        g_object_unref (icon);
        return value;
}

// 回收子进程
static void
set_mode_died (GPid pid, gint status, gpointer user_data)
{
        CsdPowerManager *manager = CSD_POWER_MANAGER (user_data);
        if (manager) {
                manager->priv->hd_helper_spawned = FALSE;
        }

        g_debug ("set_mode_died pid %d stopped with status %i", pid, status);
        g_spawn_close_pid (pid);
}

// 设置CPU频率
static void 
set_cpu_mode (gchar * mode) 
{
        GError *err = NULL;
        gchar *argv[4] = { 0 };
        GPid pid;

        if (access ("/usr/bin/cpufreq-set", F_OK) != 0)
                return;

        argv[0] = "pkexec";
        argv[1] = LIBEXECDIR "/csd-cpu-handler-helper";
        argv[2] = mode;

        g_debug("ready to set cpu mode : %s", mode);
        g_spawn_async (NULL,
			argv,
		     	get_helper_environ (),
			G_SPAWN_SEARCH_PATH | G_SPAWN_DO_NOT_REAP_CHILD,
		     	NULL,
		     	NULL,
                        &pid,
		     	&err);
	if (err) {
                g_warning("failed to set cpu %s mode, err: %s", mode, err->message);
                g_error_free(err);
        } else {
                g_child_watch_add (pid, set_mode_died, NULL);
        }
}

static gboolean
set_hard_disk_mode_cb (CsdPowerManager *manager)
{
        GError *err = NULL;
        gchar *argv[4] = { 0 };
        GPid pid;

        argv[0] = "pkexec";
        argv[1] = LIBEXECDIR "/csd-hd-helper";
        argv[2] = manager->priv->hd_mode;

        if (manager->priv->hd_helper_spawned)
                return FALSE;

        g_debug ("ready to set hard disk mode : %s",  manager->priv->hd_mode);
        g_spawn_async (NULL,
			argv,
		     	get_helper_environ (),
			G_SPAWN_SEARCH_PATH | G_SPAWN_DO_NOT_REAP_CHILD,
		     	NULL,
		     	NULL,
                        &pid,
		     	&err);
	if (err) {
                g_warning("failed to set hard disk mode: %s, err: %s",  manager->priv->hd_mode, err->message);
                g_error_free(err);
        } else {
                manager->priv->hd_helper_spawned = TRUE;
                g_child_watch_add (pid, set_mode_died, manager);
        }

        return FALSE;
}

// 设置硬盘功耗模式
static void
set_hard_disk_mode(CsdPowerManager *manager, gchar *mode)
{
        if (!manager->priv->power_mode_enable)
                return;

        if (manager->priv->hd_mode_timer_id != 0) {
                g_source_remove (manager->priv->hd_mode_timer_id);
        }

        manager->priv->hd_mode = mode;
        manager->priv->hd_mode_timer_id =
            g_timeout_add_seconds (5,   // 延迟5s执行硬盘设置
                                   (GSourceFunc)set_hard_disk_mode_cb,
                                   manager);
}

// 设置键盘背光开关状态
static void
set_kbd_backlight_status(gboolean On)
{
        GError *err = NULL;
        gchar *argv[4] = { 0 };
        GPid pid;
        gchar *dir_path = NULL;

        dir_path = get_kbd_backlight_dir_path ();
        if (dir_path == NULL)
                return;

        argv[0] = "pkexec";
        argv[1] = LIBEXECDIR "/csd-kbd-backlight-helper";
        argv[2] = On ? "on" : "off";

        g_debug("ready to set kbd backlit status : %s", argv[2]);
        g_spawn_async (NULL,
			argv,
		     	get_helper_environ (),
			G_SPAWN_SEARCH_PATH | G_SPAWN_DO_NOT_REAP_CHILD,
		     	NULL,
		     	NULL,
                        &pid,
		     	&err);
	if (err) {
                g_warning("failed to set kbd backligt status: %s , err: %s", argv[2], err->message);
                g_error_free(err);
        } else {
                g_child_watch_add (pid, set_mode_died, NULL);
        }

       g_free (dir_path);
}

// 打开键盘背光
static guint
turn_on_kbd_backlight()
{
        set_kbd_backlight_status (TRUE);
}

// 关闭键盘背光
static guint
turn_off_kbd_backlight()
{
        set_kbd_backlight_status (FALSE);
}

// 获取电量百分比
static int 
get_brightness_percentage (CsdPowerManager *manager)
{
        //台式机以此为返回值
        if (backlight_get_status () == 0)
                return get_color_brightness_percentage ();
        //笔记本等以此为返回值
        return manager->priv->power_mode_enable ?
                        power_mode_get_brightness (manager) :
                        backlight_get_percentage (NULL);
}

// 设置电量百分比
static int 
set_brightness_percentage (CsdPowerManager *manager, gint *percent)
{
        update_color_brightness_percentage (*percent);
        return manager->priv->power_mode_enable ?
                        power_mode_set_brightness (manager, *percent) :
                        backlight_set_percentage (percent, NULL);
} 

// 恢复屏幕亮度
static void
recover_screen_brightness (CsdPowerManager *manager) 
{
        gint bef_brightness_pecent = 0;
        CsdPowerRunMode cur_mode = CSD_POWER_RUN_MODE_NONE;

        bef_brightness_pecent = g_settings_get_int(manager->priv->settings, "brightness-percentage-before-psm");
        cur_mode =  g_settings_get_enum(manager->priv->settings, "run-mode");

        if (bef_brightness_pecent > 0 && cur_mode == CSD_POWER_RUN_MODE_POWERSAVE) 
        {
                // 恢复屏幕亮度
                set_brightness_percentage(manager, &bef_brightness_pecent);
                g_settings_set_int(manager->priv->settings, "brightness-percentage-before-psm", 0);
        }
}

// 设置平衡模式
static void
set_balance_mode (CsdPowerManager *manager)
{
        // 恢复屏幕亮度
        recover_screen_brightness(manager);

        // 设置CPU频率
        set_cpu_mode("ondemand");

        // 打开键盘背光
        turn_on_kbd_backlight ();

        // 设置硬盘功耗模式
        set_hard_disk_mode(manager, "other");
}

// 设置高性能模式
static void 
set_performance_mode (CsdPowerManager *manager)
{
         // 恢复屏幕亮度
        recover_screen_brightness(manager);

        // 设置CPU频率
        set_cpu_mode("performance");

        // 打开键盘背光
        turn_on_kbd_backlight();

        // 设置硬盘功耗模式
        set_hard_disk_mode(manager, "other");
}

// 设置节能模式
static void 
set_powersave_mode (CsdPowerManager *manager)
{
        // 设置屏幕亮度
        int bef_brightness_pecent = 0;
        int new_brightness_pecent = 0;
        int reduce_percentage = 0;

        bef_brightness_pecent =  get_brightness_percentage(manager);
        reduce_percentage = g_settings_get_int(manager->priv->settings, "percentage-low-reduce-brightness");
        new_brightness_pecent = bef_brightness_pecent * (100 - reduce_percentage) / 100;

        g_settings_set_int(manager->priv->settings, "brightness-percentage-before-psm", bef_brightness_pecent);
        if (is_found_desktop_env ())
                set_brightness_percentage(manager, &new_brightness_pecent);

        // 设置CPU频率
        set_cpu_mode("powersave");

        // 关闭键盘背光
        turn_off_kbd_backlight();

        // 设置硬盘功耗模式
        set_hard_disk_mode(manager, "savepower");
}

// 运行模式改变信号
static void
emit_run_mode_changed (CsdPowerManager *manager)
{
        gboolean ret;
        GError *err = NULL;
        GVariant *param  = NULL;

        if (manager->priv->connection == NULL)
                return;

        param = g_variant_new ("(i)", manager->priv->run_mode);
        ret = g_dbus_connection_emit_signal (manager->priv->connection,
                                             CSD_DBUS_SERVICE,
                                             CSD_POWER_DBUS_PATH,
                                             CSD_POWER_DBUS_INTERFACE,
                                             "ModeChanged",
                                             param,
                                             &err);
        if (!ret) {
                g_warning ("failed to emit Changed: %s", err->message);
                g_error_free (err);
        }
        if (param)
                g_variant_unref(param);
}

// 设置电源运行模式
static int
set_run_mode (CsdPowerManager *manager, CsdPowerRunMode mode, gboolean manual)
{
        int retval = -1;
        if (manager->priv->is_settings_run_mode)
                return retval;

        // 避免对gsettings key run-mode改动后触发engine_settings_key_changed_cb，重复调用set_run_mode接口
        manager->priv->is_settings_run_mode = TRUE;
        if (mode == CSD_POWER_RUN_MODE_NONE)
                goto out;

        // 手动设置电源模式，关闭“低电量自动进入节能模式”开关
        if (manual)
        {
                g_settings_set_boolean(manager->priv->settings, "percentage-low-save-mode", FALSE);
        }

        if (mode == manager->priv->run_mode)
        {
                // 更新模式记录
                manager->priv->run_mode_bp = manual ? CSD_POWER_RUN_MODE_NONE : manager->priv->run_mode_bp;
                goto out;
        }

        if (!manual && mode == CSD_POWER_RUN_MODE_POWERSAVE) { 
                // 记录自动进入节能模式前的运行模式
                manager->priv->run_mode_bp =  manager->priv->run_mode;            
        } else {
                manager->priv->run_mode_bp = CSD_POWER_RUN_MODE_NONE;
        }

        // 模式设置
        if (mode == CSD_POWER_RUN_MODE_BALANCE) {
                // 平衡模式
                set_balance_mode(manager);
        } else if (mode == CSD_POWER_RUN_MODE_PERFORMANCE) {
                // 高性能模式
                set_performance_mode(manager);
        } else if (mode == CSD_POWER_RUN_MODE_POWERSAVE) {
                // 节能模式
                set_powersave_mode(manager);
        }

        manager->priv->run_mode = mode;
        g_settings_set_enum(manager->priv->settings, "run-mode", mode);
        emit_run_mode_changed(manager);
        g_debug ("new mode is: %d\n", mode);
        retval = 1;
out:

        // 记录最后一次手动设置的模式
        g_settings_set_enum(manager->priv->settings, "run-mode-before-psm", manager->priv->run_mode_bp);
        manager->priv->is_settings_run_mode = FALSE;
        return retval;
}

static void
handle_method_call_main (CsdPowerManager *manager,
                         const gchar *method_name,
                         GVariant *parameters,
                         GDBusMethodInvocation *invocation)
{
        GPtrArray *array;
        guint i;
        GVariantBuilder *builder;
        GVariant *tuple = NULL;
        GVariant *value = NULL;
        UpDevice *device;

        /* return object */
        if (g_strcmp0 (method_name, "GetPrimaryDevice") == 0)
        {
                /* get the virtual device */
                device = engine_get_primary_device (manager);
                if (device == NULL) {
                        g_dbus_method_invocation_return_dbus_error (invocation,
                                                                    "org.freedesktop.SettingsDaemon.Power.Failed",
                                                                    "There is no primary device.");
                        goto out;
                }

                /* return the value */
                value = device_to_variant_blob (device);
                tuple = g_variant_new_tuple (&value, 1);
                g_dbus_method_invocation_return_value (invocation, tuple);
                g_object_unref (device);
                goto out;
        }

        /* return array */
        if (g_strcmp0 (method_name, "GetDevices") == 0) 
        {

                /* create builder */
                builder = g_variant_builder_new (G_VARIANT_TYPE("a(susdut)"));

                /* add each tuple to the array */
                array = manager->priv->devices_array;
                for (i=0; i<array->len; i++) {
                        device = g_ptr_array_index (array, i);
                        value = device_to_variant_blob (device);
                        g_variant_builder_add_value (builder, value);
                }

                /* return the value */
                value = g_variant_builder_end (builder);
                tuple = g_variant_new_tuple (&value, 1);
                g_dbus_method_invocation_return_value (invocation, tuple);
                g_variant_builder_unref (builder);
                goto out;
        }
        
        // 获取运行模式
        if (g_strcmp0 ( method_name, "GetMode") == 0)
        {
                g_dbus_method_invocation_return_value (invocation,
                                                        g_variant_new ("(i)",
                                                        manager->priv->run_mode));
                goto out;
        }
         
        // 设置运行模式
        if (g_strcmp0 (method_name, "SetMode") == 0)
        {
                guint new_mode = 0;
		g_variant_get (parameters, "(i)", &new_mode);
                set_run_mode(manager, new_mode, TRUE);
                g_dbus_method_invocation_return_value (invocation, NULL);
                goto out;
        }

out:
        if (tuple)
                g_variant_unref(tuple);
        if (value)
                g_variant_unref(value);

        return;
}

static void
handle_method_call (GDBusConnection       *connection,
                    const gchar           *sender,
                    const gchar           *object_path,
                    const gchar           *interface_name,
                    const gchar           *method_name,
                    GVariant              *parameters,
                    GDBusMethodInvocation *invocation,
                    gpointer               user_data)
{
        CsdPowerManager *manager = CSD_POWER_MANAGER (user_data);

        /* Check session pointer as a proxy for whether the manager is in the
           start or stop state */
        if (manager->priv->session == NULL) {
                return;
        }

        g_debug ("Calling method '%s.%s' for Power",
                 interface_name, method_name);

        if (g_strcmp0 (interface_name, CSD_POWER_DBUS_INTERFACE) == 0) {
                handle_method_call_main (manager,
                                         method_name,
                                         parameters,
                                         invocation);
        } else if (g_strcmp0 (interface_name, CSD_POWER_DBUS_INTERFACE_SCREEN) == 0) {
                handle_method_call_screen (manager,
                                           method_name,
                                           parameters,
                                           invocation);
        } else if (g_strcmp0 (interface_name, CSD_POWER_DBUS_INTERFACE_KEYBOARD) == 0) {
                handle_method_call_keyboard (manager,
                                             method_name,
                                             parameters,
                                             invocation);
        } else {
                g_warning ("not recognised interface: %s", interface_name);
        }
}

static GVariant*
handle_get_property_screen (CsdPowerManager *manager,
				const gchar * interface_name,
				const gchar *property,
				GError **error){
	GVariant *retval = NULL;
	gint32 value = 0;
        if (g_strcmp0 (interface_name, CSD_POWER_DBUS_INTERFACE_SCREEN) != 0){
                return NULL;
        }
	if (g_strcmp0 (property, "Brightness") == 0){
                value = get_brightness_percentage (manager);
                retval = g_variant_new_int32 (value);
        } else if(g_strcmp0 (property, "Enabled") == 0){
                //value =  backlight_get_status ();
                retval = g_variant_new_boolean (TRUE);
        }
	return retval;
}

// 获取电源DBus属性
static GVariant*
handle_get_property_power (CsdPowerManager *manager,
				const gchar * interface_name,
				const gchar *property_name,
				GError **error){
	GVariant *retval = NULL;
	gint32 value = 0;
    if (g_strcmp0 (interface_name, CSD_POWER_DBUS_INTERFACE) != 0){
        return NULL;
    }

    if (g_strcmp0 (property_name, "Icon") == 0) {
	retval = get_icon_property_variant (manager);
    } 
    else if (g_strcmp0 (property_name, "Tooltip") == 0) 
    {
	retval = get_tooltip_property_variant (manager);
    } 
    else if (g_strcmp0(property_name, "LowPowerThreshold") == 0) 
    {
        gint low_threshold = g_settings_get_int(manager->priv->settings, "percentage-low-threshold");
	retval = g_variant_new_int32 (low_threshold);
    } 
    else if (g_strcmp0(property_name, "LowPowerNotify") == 0) 
    {
        gboolean low_power_notify = g_settings_get_boolean (manager->priv->settings, "percentage-low-notify");
	retval = g_variant_new_boolean (low_power_notify);
    } 
    else if (g_strcmp0(property_name, "LowPowerSaveMode") == 0) 
    {
        gboolean low_power_auto_save = g_settings_get_boolean(manager->priv->settings, "percentage-low-save-mode");
	retval = g_variant_new_boolean (low_power_auto_save);
    } 
    else if (g_strcmp0(property_name, "SaveModeReduceBrightness") == 0) 
    {
         gint reduce_percentage = g_settings_get_int(manager->priv->settings, "percentage-low-reduce-brightness");
	retval = g_variant_new_int32(reduce_percentage);
    }

        return retval;
}

static GVariant *
handle_get_property (GDBusConnection *connection,
                     const gchar *sender,
                     const gchar *object_path,
                     const gchar *interface_name,
                     const gchar *property_name,
                     GError **error, gpointer user_data)
{
        CsdPowerManager *manager = CSD_POWER_MANAGER (user_data);
        GVariant *retval = NULL;

        /* Check session pointer as a proxy for whether the manager is in the
           start or stop state */
        if (manager->priv->session == NULL) {
                return NULL;
        }

        do
        {
                 // power dbus interface
                if (g_strcmp0(interface_name, CSD_POWER_DBUS_INTERFACE) == 0) 
                {
                        retval = handle_get_property_power(manager, interface_name, property_name, error);
                        break;
                }

                // screen dbus interface
	        if (g_strcmp0 (interface_name, CSD_POWER_DBUS_INTERFACE_SCREEN) == 0 ) 
                {
	                retval = handle_get_property_screen (manager, interface_name, property_name, error);
                        break;
	        } 

	        g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_FAILED,
	                       "No such interface: %s", interface_name);
        } while (0);

        return retval;
}

// 设置属性
static gboolean
handle_set_property (GDBusConnection *connection,
                     const gchar *sender,
                     const gchar *object_path,
                     const gchar *interface_name,
                     const gchar *property_name,
                     GVariant *value,
                     GError **error, gpointer user_data)
{
        CsdPowerManager *manager = CSD_POWER_MANAGER (user_data);
        gboolean retval = FALSE;

        /* Check session pointer as a proxy for whether the manager is in the
           start or stop state */
        if (manager->priv->session == NULL) {
                return retval;
        }
        
        // power dbus property
        if (g_strcmp0(interface_name, CSD_POWER_DBUS_INTERFACE) == 0) 
        {
                gboolean bvalue;
                gint ivalue;
                if (g_strcmp0(property_name, "LowPowerThreshold") == 0)
                {
                        g_variant_get (value, "(i)", &ivalue);
                        if (ivalue / 10 >= 1 && ivalue / 10 <= 4)       // 范围检验
                                retval = g_settings_set_int (manager->priv->settings, "percentage-low-threshold", ivalue);
                } 
                else if (g_strcmp0(property_name, "LowPowerNotify") == 0)
                {
                        g_variant_get (value, "(b)", &bvalue);
                        retval = g_settings_set_boolean (manager->priv->settings, "percentage-low-notify", bvalue);
                }
                 else if (g_strcmp0(property_name, "LowPowerSaveMode") == 0) 
                 {
                        g_variant_get (value, "(b)", &bvalue);
                        retval = g_settings_set_boolean (manager->priv->settings, "percentage-low-save-mode", bvalue);
                 }
                 else if (g_strcmp0(property_name, "SaveModeReduceBrightness") == 0) 
                 {
                        g_variant_get (value, "(i)", &ivalue);
                        if (ivalue / 10 >= 1 && ivalue / 10 <= 4)       // 范围检验
                                retval = g_settings_set_int (manager->priv->settings, "percentage-low-reduce-brightness", ivalue);
                 }
        }

        return retval;
}

static const GDBusInterfaceVTable interface_vtable =
{
        handle_method_call,
        handle_get_property,
        handle_set_property,
};

static void
on_bus_gotten (GObject             *source_object,
               GAsyncResult        *res,
               CsdPowerManager     *manager)
{
        GDBusConnection *connection;
        GDBusInterfaceInfo **infos;
        GError *error = NULL;
        guint i;

        if (manager->priv->bus_cancellable == NULL ||
            g_cancellable_is_cancelled (manager->priv->bus_cancellable)) {
                g_warning ("Operation has been cancelled, so not retrieving session bus");
                return;
        }

        connection = g_bus_get_finish (res, &error);
        if (connection == NULL) {
                g_warning ("Could not get session bus: %s", error->message);
                g_error_free (error);
                return;
        }
        manager->priv->connection = connection;
        infos = manager->priv->introspection_data->interfaces;
        for (i = 0; infos[i] != NULL; i++) {
                g_dbus_connection_register_object (connection,
                                                   CSD_POWER_DBUS_PATH,
                                                   infos[i],
                                                   &interface_vtable,
                                                   manager,
                                                   NULL,
                                                   NULL);
        }
}

static void
register_manager_dbus (CsdPowerManager *manager)
{
        manager->priv->introspection_data = g_dbus_node_info_new_for_xml (introspection_xml, NULL);
        manager->priv->bus_cancellable = g_cancellable_new ();
        g_assert (manager->priv->introspection_data != NULL);

        g_bus_get (G_BUS_TYPE_SESSION,
                   manager->priv->bus_cancellable,
                   (GAsyncReadyCallback) on_bus_gotten,
                   manager);
}

CsdPowerManager *
csd_power_manager_new (void)
{
        if (manager_object != NULL) {
                g_object_ref (manager_object);
        } else {
                manager_object = g_object_new (CSD_TYPE_POWER_MANAGER, NULL);
                g_object_add_weak_pointer (manager_object,
                                           (gpointer *) &manager_object);
                register_manager_dbus (manager_object);
        }
        return CSD_POWER_MANAGER (manager_object);
}
