/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*-
 *
 * Licensed under the GNU General Public License Version 2
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "config.h"

#include <unistd.h>
#include <glib-object.h>
#include <locale.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <gio/gio.h>
#include <gio/gsettings.h>

#define CSD_TLP_HELPER_EXIT_CODE_SUCCESS			0
#define CSD_TLP_HELPER_EXIT_CODE_FAILED			1
#define CSD_TLP_HELPER_EXIT_CODE_ARGUMENTS_INVALID	3
#define CSD_TLP_HELPER_EXIT_CODE_INVALID_USER		4
#define CSD_TLP_HELPER_EXIT_CODE_NO_DEVICES		5
#define CSD_TLP_HELPER_EXIT_CODE_NO_SUPPORT		6

#define TLP_CONF_FILE "/etc/default/tlp"

typedef enum _TlpMode
{
    POWERSAVE,
    BALANCE,
    PERFORMANCE
}TlpMode;

static gchar *
get_graphics_config (TlpMode mode)
{
    const char *profile = "RADEON_POWER_PROFILE_ON_AC=%s\n"
        "RADEON_POWER_PROFILE_ON_BAT=%s\n"
        "RADEON_DPM_PERF_LEVEL_ON_AC=%s\n"
        "RADEON_DPM_PERF_LEVEL_ON_BAT=%s\n";
    gchar *result = NULL;
    switch (mode)
    {
        case POWERSAVE:
            result = g_strdup_printf (profile,
                                      "default", "low",
                                      "auto", "low");
            break;
        case PERFORMANCE:
            result = g_strdup_printf (profile,
                                      "high", "high",
                                      "high", "high");
            break;
        case BALANCE:
        default:
            result = g_strdup_printf (profile,
                                      "default", "default",
                                      "auto", "auto");
            break;
    }
    return result;
}

static gchar *
get_processor_config (TlpMode mode)
{
    const char *profile = "CPU_SCALING_GOVERNOR_ON_AC=%s\n"
        "CPU_SCALING_GOVERNOR_ON_BAT=%s\n";
    gchar *result = NULL;
    switch (mode)
    {
        case POWERSAVE:
            result = g_strdup_printf (profile,
                                      "powersave", "powersave");
            break;
        case PERFORMANCE:
            result = g_strdup_printf (profile,
                                      "performance", "performance");
            break;
        case BALANCE:
        default:
            result = g_strdup ("#");
            break;
    }
    return result;
}

static gboolean
csd_tlp_helper_write (TlpMode mode, GError **error)
{
	const gchar *filename_path = NULL;
	gchar *text = NULL;
	gint retval;
	gint length;
	gint fd = -1;
	gboolean ret = TRUE;
    GString *conf;

	filename_path = TLP_CONF_FILE;

	fd = open (filename_path, O_CREAT | O_WRONLY | O_TRUNC, S_IWGRP | S_IROTH);
	if (fd < 0) {
		ret = FALSE;
		g_set_error (error, 1, 0, "failed to open filename: %s", TLP_CONF_FILE);
		goto out;
	}

    conf = g_string_new ("# Cdos Settings Daemon TLP config file\n");
    /*
    if (mode == BALANCE)
        conf = g_string_append (conf, "TLP_ENABLE=0");
    */
    conf = g_string_append (conf, "TLP_ENABLE=1\n");

    conf = g_string_append (conf, get_graphics_config (mode));
    conf = g_string_append (conf, get_processor_config (mode));
    conf = g_string_append (conf, "#USB_AUTOSUSPEND=0\n");

	length = conf->len;
	text = g_string_free (conf, FALSE);

	/* write to tlp conf file */
	retval = write (fd, text, length);
	if (retval != length) {
		ret = FALSE;
		g_set_error (error, 1, 0, "writing '%s' to %s failed", text, TLP_CONF_FILE);
		goto out;
	}
out:
	if (fd >= 0)
		close (fd);
	g_free (text);
	return ret;
}

int
main (int argc, char *argv[])
{
	GOptionContext *context;
	guint retval = 0;
	GError *error = NULL;
	gboolean set_powersave = FALSE, set_balance = FALSE, set_performance = FALSE;
	gchar *filename = NULL;
    guint tlp_mode = 0;

	const GOptionEntry options[] = {
		{ "powersave", '\0', 0, G_OPTION_ARG_NONE, &set_powersave,
		   /* command line argument */
		  "Change to powersave mode", NULL },
		{ "balance", '\0', 0, G_OPTION_ARG_NONE, &set_balance,
		   /* command line argument */
		  "Change to balance mode", NULL },
		{ "performance", '\0', 0, G_OPTION_ARG_NONE, &set_performance,
		   /* command line argument */
		  "Change to performance mode", NULL },
        { NULL}
	};

	context = g_option_context_new (NULL);
	g_option_context_set_summary (context, "Cdos Settings Daemon TLP Helper");
	g_option_context_add_main_entries (context, options, NULL);
	g_option_context_parse (context, &argc, &argv, NULL);
	g_option_context_free (context);

	if (set_powersave)
        tlp_mode = POWERSAVE;
    else if (set_performance)
        tlp_mode = PERFORMANCE;
    else if (set_balance)
        tlp_mode = BALANCE;
    else {
        g_warning ("No valid option was specified. Fallback to balance mode");
        tlp_mode = BALANCE;
    }
    gboolean ret = FALSE;
    ret = csd_tlp_helper_write (tlp_mode, &error);
    if (!ret) {
        g_warning ("%s", error->message);
        g_error_free (error);
        retval = CSD_TLP_HELPER_EXIT_CODE_ARGUMENTS_INVALID;
        goto out;
    }

	/* success */
	retval = CSD_TLP_HELPER_EXIT_CODE_SUCCESS;
    g_spawn_command_line_async ("tlp start", &error);       // tlp应用于笔记本节电

    if (error)
    {
        g_warning ("%s", error->message);
        g_clear_error (&error);
    }
out:
	g_free (filename);
	return retval;
}

