/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*-
 *
 * Copyright (C) 2007 William Jon McCann <mccann@jhu.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street - Suite 500, Boston, MA 02110-1335, USA.
 *
 */

#include "config.h"

#include <glib/gi18n-lib.h>
#include <gio/gio.h>

#include "csd-xsettings-gnome.h"

#define PREFIX_GNOME    "org.gnome"
#define PREFIX_FOUND    "org.cdos"

typedef struct _GSettingsEntry GSettignsEntry;
struct _GSettingsEntry {
        const char     *schema;
        const char     *gnome_schema;
        const char     *keys;
};

static const gchar *interface_keys[] = 
{
        "cursor-size",
        "cursor-theme",
        "font-name",
        "monospace-font-name",
        "text-scaling-factor",
        "scaling-factor",
        "menubar-detachable",
        "toolbar-detachable",
        "toolbar-icons-size",
        "cursor-blink",
        "cursor-blink-time",
        "cursor-blink-timeout",
        "icon-theme",
        "gtk-theme",
        "gtk-key-theme",
        "gtk-im-status-style",
        "gtk-im-preedit-style",
        "gtk-im-module",
        "gtk-color-palette",
        "gtk-color-scheme",
        "gtk-timeout-repeat",
        "gtk-timeout-initial",
        NULL
};

static const gchar *sound_keys[] = 
{
        "event-sounds",
        "theme-name",
        NULL
};

static const gchar *touchpad_keys[] = 
{
        "disable-while-typing",
        "tap-to-click",
        "left-handed",
        "natural-scroll",
        NULL
};

static const gchar *wm_keys[] = 
{
        "action-middle-click-titlebar",
        "action-right-click-titlebar",
        "action-double-click-titlebar",
        "mouse-button-modifier",
        "resize-with-right-button",
        "button-layout",
        "focus-mode",
        "focus-new-windows",
        "raise-on-click",
        "action-double-click-titlebar",
        "auto-raise",
        "auto-raise-delay",
        "titlebar-uses-system-font",
        "titlebar-font",
        "audible-bell",
        "visual-bell",
        "visual-bell-type",
        NULL
};

static const gchar *keybindings_keys[] = 
{
        "switch-to-workspace-left",
        "switch-to-workspace-right",
        "switch-to-workspace-up",
        "switch-to-workspace-down",
        "switch-applications",
        "switch-applications-backward",
        "switch-group",
        "cycle-windows",
        "show-desktop",
        "panel-main-menu",
        "panel-run-file-manager",
        "toggle-maximized",
        "maximize",
        "unmaximize",
        "minimize",
        "close",
        "begin-move",
        "begin-resize",
        NULL
};

/* If 'keys' is NULL, update all keys of this schema. 
 * If 'gnome_schema' is NULL, the same 'schema' will be used.
 */
static GSettignsEntry entries[] = 
{
        {"desktop.wm.keybindings", NULL, keybindings_keys},
        {"desktop.wm.preferences", NULL, wm_keys},
        {"settings-daemon.peripherals.touchpad", "desktop.peripherals.touchpad", touchpad_keys},
        {"desktop.interface", NULL, interface_keys},
        {"desktop.sound", NULL, sound_keys}
};

struct CsdXSettingsGnomePrivate {
        GHashTable        *settings;
};

#define CSD_XSETTINGS_GNOME_GET_PRIVATE(object) (G_TYPE_INSTANCE_GET_PRIVATE ((object), CSD_TYPE_XSETTINGS_GNOME, CsdXSettingsGnomePrivate))

G_DEFINE_TYPE_WITH_PRIVATE(CsdXSettingsGnome, csd_xsettings_gnome, G_TYPE_OBJECT)

static gboolean 
check_key (GSettings *settings, const gchar *key)
{
        gchar *id, *entry_id;
        guint i;
        gboolean update = FALSE;
        g_object_get (settings, "schema-id", &id, NULL);
        
        for (i = 0; i < G_N_ELEMENTS (entries); i++)
        {
                entry_id = g_strdup_printf ("%s.%s", PREFIX_FOUND, entries[i].schema, NULL);
                if (g_strcmp0 (id, entry_id) == 0)
                {
                        if (entries[i].keys && g_strv_contains (entries[i].keys, key))
                        {
                                update = TRUE;
                        }
                        //for all keys
                        if (entries[i].keys == NULL)
                                update = TRUE;
                        g_free (entry_id);
                        break;
                }
                g_free (entry_id);
        }
        return update;
}

static void
update_gnome_gsettings_by_key (GSettings *settings, 
                                const gchar *key)
{
        GSettingsSchemaSource *schema_source;
        //GSettingsSchema *schema;
        gchar *gnome_id;
        GSettings *gnome_settings;
        GVariant *val;
        gboolean ret = FALSE;

        schema_source = g_settings_schema_source_get_default ();
        gnome_id = g_object_get_data (settings, "gnome-id");
        if (!gnome_id)
                return;

        /* schema = g_settings_schema_source_lookup (schema_source, gnome_id, FALSE);
        if (!schema || !g_settings_schema_has_key (schema, key))
        {
              goto out;
        } */

        gnome_settings = g_settings_new (gnome_id);
        val = g_settings_get_value (settings, key);
        if (val)
        {
                ret = g_settings_set_value (gnome_settings, key, val);
                if (!ret)
                {
                        g_warning ("Update gnome gsettings(key:%s) failed.", key);
                }
                g_variant_unref (val);
        }
        g_object_unref (gnome_settings);

/* out:
        if (schema)
                g_settings_schema_unref (schema); */
}

static void
gsettings_changed_cb (GSettings *settings, 
                      const gchar *key, 
                      gpointer user_data)
{
        //CsdXSettingsGnome *self = CSD_XSETTINGS_GNOME (user_data);
        //CsdXSettingsGnomePrivate *priv = self->priv;
        if (check_key (settings, key))
                update_gnome_gsettings_by_key (settings, key);
}

static void
update_gnome_gsettings (CsdXSettingsGnome *self, 
                        GSettings *settings, 
                        gchar **entry_keys)
{
        CsdXSettingsGnomePrivate *priv = self->priv;
        gchar **keys;
        gchar *key;
        gint i;
        GSettingsSchema *schema;

        g_object_get (settings, "settings-schema", &schema, NULL);

        if (entry_keys)
                keys = g_strdupv (entry_keys);
        else 
                keys = g_settings_schema_list_keys (schema);

        g_debug ("%s::keys len = %d, keys = %x", __func__, g_strv_length (keys), keys);
        for (i = 0; keys[i] != NULL; i++)
        {
                key = keys[i];
                g_debug ("%s::key = %s", __func__, key);
                update_gnome_gsettings_by_key (settings, key);
        }
        g_strfreev (keys);
        g_settings_schema_unref (schema);
}

static void
add_gsettings_item (CsdXSettingsGnome *self, GSettignsEntry *entry)
{
        CsdXSettingsGnomePrivate *priv = self->priv;
        gchar *schema_id_g, *schema_id_f;
        GSettings *found_settings;
        GSettingsSchemaSource *schema_source;
        //GSettingsSchema *schema_g, *schema_f;
        gchar *id, *gnome_id;

        id = g_strdup (entry->schema);
        if (entry->gnome_schema)
                gnome_id = g_strdup (entry->gnome_schema);
        else 
                gnome_id = g_strdup (entry->schema);
        schema_source = g_settings_schema_source_get_default ();
        schema_id_g = g_strdup_printf ("%s.%s", PREFIX_GNOME, gnome_id, NULL);
        schema_id_f = g_strdup_printf ("%s.%s", PREFIX_FOUND, id, NULL);
        //schema_g = g_settings_schema_source_lookup (schema_source, schema_id_g, FALSE);
        /* schema_f = g_settings_schema_source_lookup (schema_source, schema_id_f, FALSE);
        if (!schema_f)
        {
                goto out;
        } */

        found_settings = g_settings_new (schema_id_f);
        g_object_set_data_full (found_settings, 
                                "gnome-id", g_strdup (schema_id_g), 
                                g_free);
        g_hash_table_insert (priv->settings,
                             g_strdup (id), 
                             found_settings);
        //更新gnome下的值
        /* update_gnome_gsettings (self, 
                                found_settings, 
                                entry->keys); */
        g_signal_connect_object (G_OBJECT (found_settings), 
                                "changed", 
                                G_CALLBACK (gsettings_changed_cb), 
                                self, 0);
out:
        /*  if (schema_g)
                g_settings_schema_unref (schema_g); */
        /* if (schema_f)
                g_settings_schema_unref (schema_f); */
        g_free (schema_id_g);
        g_free (schema_id_f);
        g_free (id);
        g_free (gnome_id);
}

static void
csd_xsettings_gnome_init (CsdXSettingsGnome *self)
{
        self->priv = CSD_XSETTINGS_GNOME_GET_PRIVATE (self);
        guint i;

        self->priv->settings = g_hash_table_new_full (g_str_hash, g_str_equal,
                                                (GDestroyNotify) g_free, 
                                                (GDestroyNotify) g_object_unref);
        for (i = 0; i < G_N_ELEMENTS (entries); i++)
        {
                add_gsettings_item (self, &entries[i]);
        }
}

static void
csd_xsettings_gnome_finalize (GObject *object)
{
        CsdXSettingsGnome *self = CSD_XSETTINGS_GNOME (object);
        CsdXSettingsGnomePrivate *priv = self->priv;
        if (priv->settings != NULL) {
                g_hash_table_destroy (priv->settings);
                priv->settings = NULL;
        }

        G_OBJECT_CLASS (csd_xsettings_gnome_parent_class)->finalize (object);
}

static void
csd_xsettings_gnome_class_init (CsdXSettingsGnomeClass *klass)
{
        GObjectClass *object_class = G_OBJECT_CLASS (klass);

        object_class->finalize = csd_xsettings_gnome_finalize;
}

CsdXSettingsGnome *
csd_xsettings_gnome_new (void)
{
        return CSD_XSETTINGS_GNOME (g_object_new (CSD_TYPE_XSETTINGS_GNOME, NULL));
}
